/*--------------------------------------------------------------------------*/
/* ALBERTA:  an Adaptive multi Level finite element toolbox using           */
/*           Bisectioning refinement and Error control by Residual          */
/*           Techniques for scientific Applications                         */
/*                                                                          */
/* file:     element.c                                                      */
/*                                                                          */
/*                                                                          */
/* description:  routines on elements that depend on the dimension in 2d    */
/*                                                                          */
/*--------------------------------------------------------------------------*/
/*                                                                          */
/*  authors:   Alfred Schmidt                                               */
/*             Zentrum fuer Technomathematik                                */
/*             Fachbereich 3 Mathematik/Informatik                          */
/*             Universitaet Bremen                                          */
/*             Bibliothekstr. 2                                             */
/*             D-28359 Bremen, Germany                                      */
/*                                                                          */
/*             Kunibert G. Siebert                                          */
/*             Institut fuer Mathematik                                     */
/*             Universitaet Augsburg                                        */
/*             Universitaetsstr. 14                                         */
/*             D-86159 Augsburg, Germany                                    */
/*                                                                          */
/*  http://www.alberta-fem.de/                                              */
/*                                                                          */
/*--------------------------------------------------------------------------*/
/*  (c) by A. Schmidt and K.G. Siebert (1996-2004)                          */
/*                                                                          */
/*     This program is free software; you can redistribute it and/or modify */
/*     it under the terms of the GNU General Public License as published by */
/*     the Free Software Foundation; either version 2 of the License, or    */
/*     any later version.                                                   */
/*                                                                          */
/*     This program is distributed in the hope that it will be useful,      */
/*     but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/*     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/*     GNU General Public License for more details.                         */
/*--------------------------------------------------------------------------*/

#include "alberta.h"

/*--------------------------------------------------------------------------*/
/*  world_to_coord():  return -1 if inside, otherwise index of lambda<0     */
/*--------------------------------------------------------------------------*/

int world_to_coord(const EL_INFO *el_info, const REAL *xy, REAL lambda[DIM+1])
{
  FUNCNAME("world_to_coord");
  REAL  edge[DIM][DIM_OF_WORLD], x[DIM_OF_WORLD];
  REAL  x0, det, det0, det1, adet, lmin;
  int   i, j, k;

#if (DIM != DIM_OF_WORLD)
  ERROR_EXIT("not yet for DIM != DIM_OF_WORLD\n");
#endif

  TEST_FLAG(FILL_COORDS, el_info);
  TEST_EXIT(!el_info->mesh->parametric)("not for parametric meshes\n");

/*  wir haben das gleichungssystem zu loesen: */
/*       ( q1x q2x )  (lambda1)     (qx)      */
/*       ( q1y q2y )  (lambda2)  =  (qy)      */
/*      mit qi=pi-p3, q=xy-p3                 */

  for (j=0; j<DIM_OF_WORLD; j++) {
    x0 = el_info->coord[DIM][j];
    x[j] = xy[j] - x0;
    for (i=0; i<DIM; i++)
      edge[i][j] = el_info->coord[i][j] - x0;
  }

  det  = edge[0][0] * edge[1][1] - edge[0][1] * edge[1][0]; 
  det0 =       x[0] * edge[1][1] -       x[1] * edge[1][0]; 
  det1 = edge[0][0] * x[1]       - edge[0][1] * x[0]; 

  adet = ABS(det);

  if (adet < 1.E-20) {
    ERROR_EXIT("det = %le; abort\n", det);
    return(-2);
  }

  lambda[0] = det0 / det;
  lambda[1] = det1 / det;
  lambda[2] = 1.0 - lambda[0] - lambda[1];

  k = -1;
  lmin = 0.0;
  j = 0;
  for (i = 0; i <= DIM; i++) {
    if ((lambda[i]*adet) < -1.E-15) {
      if (lambda[i] < lmin) {
	k = i;
	lmin = lambda[i];
      }
      j++;
    }
  }

  return(k);
}

/*--------------------------------------------------------------------------*/
/*  transform local coordintes l to world coordinates; if w is non nil      */
/*  store them at w otherwise return a pointer to some local static         */
/*  area containing world coordintes                                        */
/*--------------------------------------------------------------------------*/

const REAL *coord_to_world(const EL_INFO *el_info, const REAL *l, REAL_D w)
{
  FUNCNAME("coord_to_world");
  static REAL world[DIM_OF_WORLD];
  REAL        c, *ret;
  const REAL  *v;
  int         i, j;

  TEST_FLAG(FILL_COORDS, el_info);
  TEST_EXIT(!el_info->mesh->parametric)("not for parametric meshes\n");

  ret = w ? w : world;

  v = el_info->coord[0];
  c = l[0];
  for (j = 0; j < DIM_OF_WORLD; j++)
    ret[j] = c*v[j];
  
  for (i = 1; i < N_VERTICES; i++)
  {
    v = el_info->coord[i];
    c = l[i];
    for (j = 0; j < DIM_OF_WORLD; j++)
      ret[j] += c*v[j];
  }
  return((const REAL *) ret);
}

/*--------------------------------------------------------------------------*/
/*  compute volume of an element                                            */
/*--------------------------------------------------------------------------*/

REAL el_det(const EL_INFO *el_info)
{
  FUNCNAME("el_det");
  REAL        e1[DIM_OF_WORLD], e2[DIM_OF_WORLD], det;
  const REAL  *v0;
  int         i;

  TEST_FLAG(FILL_COORDS, el_info);
  TEST_EXIT(!el_info->mesh->parametric)("not for parametric meshes\n");

  v0 = el_info->coord[0];
  for (i = 0; i < DIM_OF_WORLD; i++)
  {
    e1[i] = el_info->coord[1][i] - v0[i];
    e2[i] = el_info->coord[2][i] - v0[i];
  }

#if DIM_OF_WORLD==2
  det = e1[0]*e2[1] - e1[1]*e2[0];
  det = ABS(det);
#else
#if DIM_OF_WORLD==3
  {
    REAL n[3];

    n[0] = e1[1]*e2[2] - e1[2]*e2[1];
    n[1] = e1[2]*e2[0] - e1[0]*e2[2];
    n[2] = e1[0]*e2[1] - e1[1]*e2[0];
    det = sqrt(n[0]*n[0] + n[1]*n[1] + n[2]*n[2]);
  }
#else
  TEST_EXIT(0)("not yet for DIM_OF_WORLD = %d",DIM_OF_WORLD);
#endif
#endif

  return(det);
}

#define VOL_SHAT 0.5

REAL el_volume(const EL_INFO *el_info)
{
  return(VOL_SHAT*el_det(el_info));
}

/*--------------------------------------------------------------------------*/
/*  compute gradients of basis functions on element; return the absulute    */
/*  value of the determinante from the transformation to the reference      */
/*  element                                                                 */
/*--------------------------------------------------------------------------*/

REAL el_grd_lambda(const EL_INFO *el_info, REAL grd_lam[DIM+1][DIM_OF_WORLD])
{
  FUNCNAME("el_grd_lambda");
#if DIM_OF_WORLD == 2
  int         i, j;
  REAL        e1[DIM_OF_WORLD], e2[DIM_OF_WORLD], det, adet;
  const REAL  *v0;
  REAL        a11, a12, a21, a22;

  TEST_FLAG(FILL_COORDS, el_info);
  TEST_EXIT(!el_info->mesh->parametric)("not for parametric meshes\n");

  v0 = el_info->coord[0];
  for (i = 0; i < DIM_OF_WORLD; i++) {
    e1[i] = el_info->coord[1][i] - v0[i];
    e2[i] = el_info->coord[2][i] - v0[i];
  }

  det = e1[0] * e2[1] - e1[1] * e2[0];
  adet = ABS(det);
  if (adet < 1.0E-25) 
  {
    MSG("abs(det) = %lf\n", adet);
    for (i = 0; i <= DIM; i++)
      for (j = 0; j < DIM_OF_WORLD; j++)
	grd_lam[i][j] = 0.0;
  }
  else
  {
    det = 1.0 / det;
    a11 =  e2[1] * det;       /* (a_ij) = A^{-T} */
    a21 = -e2[0] * det;
    a12 = -e1[1] * det;
    a22 =  e1[0] * det;

    grd_lam[1][0] = a11;
    grd_lam[1][1] = a21;
    grd_lam[2][0] = a12;
    grd_lam[2][1] = a22;
    grd_lam[0][0] = - grd_lam[1][0] - grd_lam[2][0];
    grd_lam[0][1] = - grd_lam[1][1] - grd_lam[2][1];
  }

  return(adet);
#else
  ERROR_EXIT("not yet for DIM_OF_WORLD = %d\n", DIM_OF_WORLD);
  return 1.0/0.0; /* HUGE_VAL */
#endif

}

#if DIM_OF_WORLD == 3
/*--------------------------------------------------------------------------*/
/*  vector product in R^3                                                   */
/*--------------------------------------------------------------------------*/

void vector_product(const REAL_D x, const REAL_D y, REAL_D z)
{
  /* FUNCNAME("vector_product"); */

  z[0] = x[1]*y[2] - x[2]*y[1];
  z[1] = x[2]*y[0] - x[0]*y[2];
  z[2] = x[0]*y[1] - x[1]*y[0];

  return;
}
#endif

/*--------------------------------------------------------------------------*/
/*  calculate a facenormal of edge side of a traingle with coordinates      */
/*  coord; return the absulute value of the determinante from the           */
/*  transformation to the reference element                                 */
/*--------------------------------------------------------------------------*/

REAL get_face_normal(const EL_INFO *el_info, int i0, REAL *normal)
{
  FUNCNAME("get_face_normal");
#if DIM_OF_WORLD == 2
  static int   ind[5] = {0, 1, 2, 0, 1};
  REAL         det;
  int          i1 = ind[i0+1], i2 = ind[i0+2];
  const REAL_D *coord = el_info->coord;

  TEST_FLAG(FILL_COORDS, el_info);

  normal[0] = coord[i2][1] - coord[i1][1];
  normal[1] = coord[i1][0] - coord[i2][0];
  det = sqrt(SQR(normal[0])+SQR(normal[1]));
  TEST_EXIT(det > 1.e-30)("det = 0 on face %d\n", i0);
  normal[0] /= det;
  normal[1] /= det;
  return(det);
#else 
  MSG("not implemented for DIM_OF_WORLD = %d in 2d\n", DIM_OF_WORLD);
  for (i0 = 0; i0 < DIM_OF_WORLD; i0++)  normal[i0] = 0.0;
  return 1.0/0.0; /* HUGE_VAL */
#endif

}

/*--------------------------------------------------------------------------*/
/* orient the vertices of edges/faces                                       */
/* used by estimator for the jumps => same quadrature nodes from both sides!*/
/*--------------------------------------------------------------------------*/

const int *sort_face_indices(const EL *el, int face, int *vec)
{
  static int  sorted[3][2][2] = {
    {{1,2}, {2,1}}, {{2,0},{0,2}}, {{0,1}, {1,0}}};
  static int  vertex_of_face[3][2] = {{1,2}, {2,0}, {0,1}};
  int    i, no, *val  = nil;
  int    *vof = vertex_of_face[face];
  DOF    **dof = el->dof;

  if (dof[vof[0]][0] < dof[vof[1]][0])
    no = 0;
  else
    no = 1;


  if (vec)
  {
    val = vec;
    for (i = 0; i < 2; i++)
      val[i] = sorted[face][no][i];
  }
  else
    val = sorted[face][no];
 
  return((const int *) val);
}
