/*--------------------------------------------------------------------------*/
/* ALBERTA:  an Adaptive multi Level finite element toolbox using           */
/*           Bisectioning refinement and Error control by Residual          */
/*           Techniques for scientific Applications                         */
/*                                                                          */
/* file:     lagrange_1.c                                                   */
/*                                                                          */
/* description:  piecewise linear Lagrange elements in 2d                   */
/*                                                                          */
/*--------------------------------------------------------------------------*/
/*                                                                          */
/*  authors:   Alfred Schmidt                                               */
/*             Zentrum fuer Technomathematik                                */
/*             Fachbereich 3 Mathematik/Informatik                          */
/*             Universitaet Bremen                                          */
/*             Bibliothekstr. 2                                             */
/*             D-28359 Bremen, Germany                                      */
/*                                                                          */
/*             Kunibert G. Siebert                                          */
/*             Institut fuer Mathematik                                     */
/*             Universitaet Augsburg                                        */
/*             Universitaetsstr. 14                                         */
/*             D-86159 Augsburg, Germany                                    */
/*                                                                          */
/*  http://www.alberta-fem.de/                                              */
/*                                                                          */
/*--------------------------------------------------------------------------*/
/*  (c) by A. Schmidt and K.G. Siebert (1996-2004)                          */
/*                                                                          */
/*     This program is free software; you can redistribute it and/or modify */
/*     it under the terms of the GNU General Public License as published by */
/*     the Free Software Foundation; either version 2 of the License, or    */
/*     any later version.                                                   */
/*                                                                          */
/*     This program is distributed in the hope that it will be useful,      */
/*     but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/*     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/*     GNU General Public License for more details.                         */
/*--------------------------------------------------------------------------*/

static const REAL   bary1[3][DIM+1] = {{1.0, 0.0, 0.0},
					{0.0, 1.0, 0.0},
					{0.0, 0.0, 1.0}};
#if 0
static const int    fct_at_node1[3] = {0,1,2};
static int    n_dof1[DIM+1] = {1, 0, 0};
#endif
/*--------------------------------------------------------------------------*/
/*  basisfunction at vertex 0                                               */
/*--------------------------------------------------------------------------*/

static REAL phi1v0(const REAL lambda[DIM+1])
{
  return(lambda[0]);
}

static const REAL *grd_phi1v0(const REAL lambda[DIM+1])
{
  static const REAL  grd[DIM+1] = {1.0, 0.0, 0.0};

  return(grd);
}

static const REAL (*D2_phi1v0(const REAL *lambda))[DIM+1]
{
  static const REAL D2[DIM+1][DIM+1] = {{0, 0, 0},{0, 0, 0}, {0, 0, 0}};

  return(D2);
}

/*--------------------------------------------------------------------------*/
/*  basisfunction at vertex 1                                               */
/*--------------------------------------------------------------------------*/

static REAL phi1v1(const REAL lambda[DIM+1])
{
  return(lambda[1]);
}

static const REAL *grd_phi1v1(const REAL lambda[DIM+1])
{
  static const REAL  grd[DIM+1] = {0.0, 1.0, 0.0};

  return(grd);
}

static const REAL (*D2_phi1v1(const REAL *lambda))[DIM+1]
{
  static const REAL D2[DIM+1][DIM+1] = {{0, 0, 0},{0, 0, 0}, {0, 0, 0}};

  return(D2);
}

/*--------------------------------------------------------------------------*/
/*  basisfunction at vertex 2                                               */
/*--------------------------------------------------------------------------*/

static REAL phi1v2(const REAL lambda[DIM+1])
{
  return(lambda[2]);
}

static const REAL *grd_phi1v2(const REAL lambda[DIM+1])
{
  static const REAL  grd[DIM+1] = {0.0, 0.0, 1.0};

  return(grd);
}

static const REAL (*D2_phi1v2(const REAL *lambda))[DIM+1]
{
  static  const REAL D2[DIM+1][DIM+1] = {{0, 0, 0},{0, 0, 0}, {0, 0, 0}};

  return(D2);
}

/*--------------------------------------------------------------------------*/
/*  functions for combining basisfunctions with coefficients                */
/*--------------------------------------------------------------------------*/

static const DOF *get_dof_indices1(const EL *el, const DOF_ADMIN *admin,
				   DOF *idof)
{
  static DOF  index_vec[N_VERTICES];
  DOF         *rvec = idof ? idof : index_vec;
  int         i, n0 = admin->n0_dof[VERTEX];
  DOF         **dof = el->dof;

  for (i = 0; i < N_VERTICES; i++)
    rvec[i] = dof[i][n0];

  return((const DOF *) rvec);
}

static const S_CHAR *get_bound1(const EL_INFO *el_info, S_CHAR *bound)
{
  FUNCNAME("get_bound1");
  static S_CHAR  bound_vec[N_VERTICES];
  S_CHAR         *rvec = bound ? bound : bound_vec;
  int            i;

  TEST_FLAG(FILL_BOUND, el_info);

  for (i = 0; i < N_VERTICES; i++)
    rvec[i] = el_info->bound[i];

  return((const S_CHAR *) rvec);
}

static const int *get_int_vec1(const EL *el, const DOF_INT_VEC *vec,
			       int *ivec)
{
  FUNCNAME("get_int_vec1");
  int            i, n0;
  static int     local_vec[N_VERTICES];
  int            *v = nil, *rvec = ivec ? ivec : local_vec;
  DOF            **dof = el->dof;

  GET_DOF_VEC(v, vec);
  n0 = vec->fe_space->admin->n0_dof[VERTEX];

  for (i = 0; i < N_VERTICES; i++)
    rvec[i] = v[dof[i][n0]];

  return((const int *) rvec);
}

static const REAL *get_real_vec1(const EL *el, const DOF_REAL_VEC *vec,
				 REAL *Rvec)
{
  FUNCNAME("get_real_v1");
  int            i, n0;
  static REAL    local_vec[N_VERTICES];
  REAL           *v = nil, *rvec = Rvec ? Rvec : local_vec;
  DOF            **dof = el->dof;

  GET_DOF_VEC(v, vec);
  n0 = vec->fe_space->admin->n0_dof[VERTEX];

  for (i = 0; i < N_VERTICES; i++)
    rvec[i] = v[dof[i][n0]];

  return((const REAL *) rvec);
}

static const REAL_D *get_real_d_vec1(const EL *el, const DOF_REAL_D_VEC *vec,
				     REAL_D *Rvec)
{
  FUNCNAME("get_real_d_v1");
  int            i, k, n0;
  static REAL_D  local_vec[N_VERTICES];
  REAL_D         *v = nil, *rvec = Rvec ? Rvec : local_vec;
  DOF            **dof = el->dof;

  GET_DOF_VEC(v, vec);
  n0 = vec->fe_space->admin->n0_dof[VERTEX];

  for (i = 0; i < N_VERTICES; i++)
    for (k = 0; k < DIM_OF_WORLD; k++)
      rvec[i][k] = v[dof[i][n0]][k];

  return((const REAL_D *) rvec);
}

static const U_CHAR *get_uchar_vec1(const EL *el, const DOF_UCHAR_VEC *vec,
				    U_CHAR *uvec)
{
  FUNCNAME("get_uchar_vec1");
  int            i, n0;
  static U_CHAR  local_vec[N_VERTICES];
  U_CHAR         *v = nil, *rvec = uvec ? uvec : local_vec;
  DOF            **dof = el->dof;

  GET_DOF_VEC(v, vec);
  n0 = vec->fe_space->admin->n0_dof[VERTEX];

  for (i = 0; i < N_VERTICES; i++)
    rvec[i] = v[dof[i][n0]];

  return((const U_CHAR *) rvec);
}

static const S_CHAR *get_schar_vec1(const EL *el, const DOF_SCHAR_VEC *vec,
				    S_CHAR *svec)
{
  FUNCNAME("get_schar_vec1");
  int            i, n0;
  static S_CHAR  local_vec[N_VERTICES];
  S_CHAR         *v = nil, *rvec = svec ? svec : local_vec;
  DOF            **dof = el->dof;

  GET_DOF_VEC(v, vec);
  n0 = vec->fe_space->admin->n0_dof[VERTEX];

  for (i = 0; i < N_VERTICES; i++)
    rvec[i] = v[dof[i][n0]];

  return((const S_CHAR *) rvec);
}

static const REAL *interpol1(const EL_INFO *el_info, int no, const int *b_no,
			     REAL (*f)(const REAL_D), REAL *vec)
{
  FUNCNAME("interpol1");
  static REAL  inter[N_VERTICES];
  REAL         *rvec = vec ? vec : inter;
  int          i;
  const PARAMETRIC *parametric = el_info->mesh->parametric;

  if (b_no)
  {
    if (no <= 0 || no > N_VERTICES)
    {
      ERROR("something is wrong, doing nothing\n");
      rvec[0] = 0.0;
      return((const REAL *) rvec);
    }
  }
  
  if (parametric  &&  parametric->init_element(el_info, parametric))
  {
    REAL_D   world[3];

    if (b_no)
    {
      REAL  lambda[3][3] = {{0,0,0},{0,0,0},{0,0,0}};

      for (i = 0; i < no; i++)
	lambda[i][b_no[i]] = 1.0;

      parametric->coord_to_world(el_info, nil, no, 
				 (const REAL (*)[DIM+1]) lambda, world);

      for (i = 0; i < no; i++)
	rvec[i] = (*f)(world[i]);
    }
    else
    {
      static const REAL  lambda[3][3] = {{1,0,0},{0,1,0},{0,0,1}};

      parametric->coord_to_world(el_info, nil, 3, lambda, world);

      for (i = 0; i < 3; i++)
	rvec[i] = (*f)(world[i]);
    }
  }
  else
  {
    TEST_FLAG(FILL_COORDS, el_info);

    if (b_no)
    {
      for (i = 0; i < no; i++)
      {
	rvec[i] = (*f)(el_info->coord[b_no[i]]);
      }
    }
    else
    {
      for (i = 0; i < N_VERTICES; i++)
      {
	rvec[i] = (*f)(el_info->coord[i]);
      }
    }
  }

  return((const REAL *) rvec);
}

static const REAL_D *interpol_d1(const EL_INFO *el_info, int no, 
				 const int *b_no,
				 const REAL *(*f)(const REAL_D, REAL_D),
				 REAL_D *vec)
{
  FUNCNAME("interpol_d1");
  static REAL_D  inter[N_VERTICES];
  REAL_D         *rvec = vec ? vec : inter;
  int            i;
  const PARAMETRIC *parametric = el_info->mesh->parametric;

  if (b_no)
  {
    if (no <= 0 || no > N_VERTICES)
    {
      ERROR("something is wrong, doing nothing\n");
      for (i = 0; i < DIM_OF_WORLD; i++)
	rvec[0][i] = 0.0;
      return((const REAL_D *) rvec);
    }
  }
  
  if (parametric  &&  parametric->init_element(el_info, parametric))
  {
    REAL_D   world[3];

    if (b_no)
    {
      REAL  lambda[3][3] = {{0,0,0},{0,0,0},{0,0,0}};

      for (i = 0; i < no; i++)
	lambda[i][b_no[i]] = 1.0;

      parametric->coord_to_world(el_info, nil, no, 
				 (const REAL (*)[DIM+1]) lambda, world);

      for (i = 0; i < no; i++)
	(*f)(world[i], rvec[i]);
    }
    else
    {
      static const REAL  lambda[3][3] = {{1,0,0},{0,1,0},{0,0,1}};

      parametric->coord_to_world(el_info, nil, 3, lambda, world);

      for (i = 0; i < 3; i++)
	(*f)(world[i], rvec[i]);
    }
  }
  else
  {
    TEST_FLAG(FILL_COORDS, el_info);

    if (b_no)
    {
      for (i = 0; i < no; i++)
      {
	(*f)(el_info->coord[b_no[i]], rvec[i]);
      }
    }
    else
    {
      for (i = 0; i < N_VERTICES; i++)
      {
	(*f)(el_info->coord[i], rvec[i]);
      }
    }
  }
  
  return((const REAL_D *) rvec);
}

/*--------------------------------------------------------------------------*/
/*  functions for interpolation/ restriction during refinement/coarsening   */
/*--------------------------------------------------------------------------*/

static void real_refine_inter1(DOF_REAL_VEC *drv, RC_LIST_EL *list, int n)
{
  FUNCNAME("real_refine_inter1");
  EL      *el;
  REAL    *vec = nil;
  DOF     dof_new, dof0, dof1;
  int     n0;

  if (n < 1) return;
  GET_DOF_VEC(vec, drv);
  n0 = drv->fe_space->admin->n0_dof[VERTEX];
  el = list->el;
  dof0 = el->dof[0][n0];           /* 1st endpoint of refinement edge */
  dof1 = el->dof[1][n0];           /* 2nd endpoint of refinement edge */
  dof_new = el->child[0]->dof[DIM][n0];  /*      newest vertex is DIM */
  vec[dof_new] = 0.5*(vec[dof0] + vec[dof1]);

  return;
}

/*--------------------------------------------------------------------------*/
/*  linear interpolation during coarsening: do nothing                      */
/*--------------------------------------------------------------------------*/

static void real_coarse_restr1(DOF_REAL_VEC *drv, RC_LIST_EL *list, int n)
{
  FUNCNAME("real_coarse_restr1");
  EL      *el;
  REAL    *vec = nil;
  DOF     dof_new, dof0, dof1;
  int     n0;

  if (n < 1) return;
  GET_DOF_VEC(vec, drv);
  n0 = drv->fe_space->admin->n0_dof[VERTEX];
  el = list->el;
  dof0 = el->dof[0][n0];           /* 1st endpoint of refinement edge */
  dof1 = el->dof[1][n0];           /* 2nd endpoint of refinement edge */
  dof_new = el->child[0]->dof[DIM][n0];  /*      newest vertex is DIM */
  vec[dof0] += 0.5*vec[dof_new];
  vec[dof1] += 0.5*vec[dof_new];

  return;
}

static void real_d_refine_inter1(DOF_REAL_D_VEC *drdv, RC_LIST_EL *list, int n)
{
  FUNCNAME("real_d_refine_inter1");
  EL      *el;
  REAL_D  *vec = nil;
  DOF     dof_new, dof0, dof1;
  int     n0, j;

  if (n < 1) return;
  GET_DOF_VEC(vec, drdv);
  n0 = drdv->fe_space->admin->n0_dof[VERTEX];
  el = list->el;
  dof0 = el->dof[0][n0];           /* 1st endpoint of refinement edge */
  dof1 = el->dof[1][n0];           /* 2nd endpoint of refinement edge */
  dof_new = el->child[0]->dof[DIM][n0];  /*      newest vertex is DIM */
  for (j = 0; j < DIM_OF_WORLD; j++)
    vec[dof_new][j] = 0.5*(vec[dof0][j] + vec[dof1][j]);

  return;
}

/*--------------------------------------------------------------------------*/
/*  linear interpolation during coarsening: do nothing                      */
/*--------------------------------------------------------------------------*/

static void real_d_coarse_restr1(DOF_REAL_D_VEC *drdv, RC_LIST_EL *list, int n)
{
  FUNCNAME("real_d_coarse_restr1");
  EL      *el;
  REAL_D  *vec = nil;
  DOF     dof_new, dof0, dof1;
  int     n0, j;

  if (n < 1) return;
  GET_DOF_VEC(vec, drdv);
  n0 = drdv->fe_space->admin->n0_dof[VERTEX];
  el = list->el;
  dof0 = el->dof[0][n0];           /* 1st endpoint of refinement edge */
  dof1 = el->dof[1][n0];           /* 2nd endpoint of refinement edge */
  dof_new = el->child[0]->dof[DIM][n0];  /*      newest vertex is DIM */
  for (j = 0; j < DIM_OF_WORLD; j++)
  {
    vec[dof0][j] += 0.5 * vec[dof_new][j];
    vec[dof1][j] += 0.5 * vec[dof_new][j];
  }

  return;
}

static BAS_FCT      *phi1[3]     = {phi1v0, phi1v1, phi1v2};
static GRD_BAS_FCT  *grd_phi1[3] = {grd_phi1v0, grd_phi1v1, grd_phi1v2};
static D2_BAS_FCT   *D2_phi1[3]  = {D2_phi1v0, D2_phi1v1, D2_phi1v2};

static BAS_FCTS  lagrange1 = {"lagrange1", 3, 1, {1, 0, 0},
			      nil, phi1, grd_phi1, D2_phi1, 
			      get_dof_indices1,
			      get_bound1, 
			      interpol1,
			      interpol_d1,
			      get_int_vec1,
			      get_real_vec1,
			      get_real_d_vec1,
			      get_uchar_vec1,
			      get_schar_vec1,
			      real_refine_inter1,
			      nil,
			      real_coarse_restr1,
			      real_d_refine_inter1,
			      nil,
			      real_d_coarse_restr1,
			      (void *) bary1};
