/*--------------------------------------------------------------------------*/
/* ALBERTA:  an Adaptive multi Level finite element toolbox using           */
/*           Bisectioning refinement and Error control by Residual          */
/*           Techniques for scientific Applications                         */
/*                                                                          */
/* file:     macro.c                                                        */
/*                                                                          */
/*                                                                          */
/* description:  dimension dependent part of reading/writing macro          */
/*               triangulations for 3d                                      */
/*                                                                          */
/*--------------------------------------------------------------------------*/
/*                                                                          */
/*  authors:   Alfred Schmidt                                               */
/*             Zentrum fuer Technomathematik                                */
/*             Fachbereich 3 Mathematik/Informatik                          */
/*             Universitaet Bremen                                          */
/*             Bibliothekstr. 2                                             */
/*             D-28359 Bremen, Germany                                      */
/*                                                                          */
/*             Kunibert G. Siebert                                          */
/*             Institut fuer Mathematik                                     */
/*             Universitaet Augsburg                                        */
/*             Universitaetsstr. 14                                         */
/*             D-86159 Augsburg, Germany                                    */
/*                                                                          */
/*             Daniel Koester                                               */
/*             Institut fuer Mathematik                                     */
/*             Albert-Ludwigs-Universitaet Freiburg                         */
/*             Hermann-Herder-Str. 10                                       */
/*             D-79104 Freiburg                                             */
/*                                                                          */
/*  http://www.alberta-fem.de/                                              */
/*                                                                          */
/*--------------------------------------------------------------------------*/
/*  (c) by A. Schmidt and K.G. Siebert (1996-2004)                          */
/*                                                                          */
/*     This program is free software; you can redistribute it and/or modify */
/*     it under the terms of the GNU General Public License as published by */
/*     the Free Software Foundation; either version 2 of the License, or    */
/*     any later version.                                                   */
/*                                                                          */
/*     This program is distributed in the hope that it will be useful,      */
/*     but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/*     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/*     GNU General Public License for more details.                         */
/*--------------------------------------------------------------------------*/

#include <ctype.h>
#include <rpc/types.h>
#include <rpc/xdr.h>
#include "alberta.h"

/*--------------------------------------------------------------------------*/
/* these definitions are necessary at the moment for the macros "MEM_ALLOC" */
/* and "MEM_FREE" to work properly                                          */
/*--------------------------------------------------------------------------*/

typedef int INT_NVERT[N_VERTICES];
typedef int INT_NNEIGH[N_NEIGH];
typedef S_CHAR SCHAR_NNEIGH[N_NEIGH];
typedef U_CHAR UCHAR_NNEIGH[N_NEIGH];


/*--------------------------------------------------------------------------*/
/* default_boundary(): This procedure returns a constant pointer to a       */
/* BOUNDARY structure corresponding to a standard polygonal boundary.       */
/* NOTE: The total number of possible boundary types is unknown during the  */
/* first calls of this procedure. It therefore has to allocate a new        */
/* BOUNDARY structure for each new argument "bound" passed to it; the       */
/* address of this new BOUNDARY is stored in a dynamic array default_bounds.*/
/*--------------------------------------------------------------------------*/

const BOUNDARY *default_boundary(MESH *mesh, int bound)
{
  FUNCNAME("default_boundary");

  typedef BOUNDARY *BOUND_PTR;

  static BOUND_PTR *default_bounds = nil;
  static unsigned int i = 0, size = 1;

  if(!bound) {
    WARNING("tried to assign a BOUNDARY * with bound == 0!\n");
    return(nil);
  }

  if(!default_bounds) {
     default_bounds = MEM_ALLOC(size, BOUND_PTR);
     default_bounds[0] = MEM_ALLOC(1, BOUNDARY);

     default_bounds[0]->param_bound = nil;
     default_bounds[0]->bound = bound;

     return((const BOUNDARY *) default_bounds[0]);
  }

  for(i = 0; i < size; i++)
    if (default_bounds[i]->bound == bound)
      return((const BOUNDARY *) default_bounds[i]);

  if(i == size) {
    default_bounds = MEM_REALLOC(default_bounds, size, size + 1, BOUND_PTR);

    default_bounds[size] = MEM_ALLOC(1, BOUNDARY);

    default_bounds[size]->param_bound = nil;
    default_bounds[size]->bound = bound;
    size++;
  }
  return((const BOUNDARY *) default_bounds[size - 1]); 
}


/*--------------------------------------------------------------------------*/
/* new_bound_val(): determine the value of the subsimplex being checked     */
/* The type of a boundary subsimplex is set to the highest type of all      */
/* boundary faces containg the edge.                                        */
/*--------------------------------------------------------------------------*/

static S_CHAR new_bound_val(const BOUNDARY *boundary, S_CHAR bound)
{
  if (boundary) 
  {
    if (boundary->bound >= DIRICHLET)
      return(MAX(bound, boundary->bound));
    else
      if (bound != INTERIOR)
       	return(MAX(bound, boundary->bound));
      else
	return(boundary->bound);
  }

  return(bound);
}


/*--------------------------------------------------------------------------*/
/*  fill_bound_info(): fill boundary info on faces from MACRO_DATA *data    */
/* into corresponding macro elements, set boundary types of vertices (edges */
/* are handled later in fill_more_boundary_dofs) using new_bound_val().     */
/*--------------------------------------------------------------------------*/

static void fill_bound_info(MESH *mesh, MACRO_DATA *data,
                            const BOUNDARY *(*bdry)(MESH *, int))
{
  MACRO_EL   *mel = mesh->first_macro_el;
  int         i, j, k;
  int         ne = mesh->n_elements, ned = mesh->n_edges;
  int         nv = mesh->n_vertices;

  S_CHAR  *bound = MEM_ALLOC(nv+ned, S_CHAR);

/*--------------------------------------------------------------------------*/
/* The first N_FACES entries of mel[].boundary will contain pointers to     */
/* BOUNDARY structures corresponding to the elements' faces. These are the  */
/* crucial infos provided externally, the other subsimplices (here edges    */
/* and vertices) are set to a boundary type depending on the face containing*/
/* them.                                                                    */
/*--------------------------------------------------------------------------*/

  for(i = 0; i < data->n_macro_elements; i++) {
    for(j = 0; j < N_FACES; j++) {
      if (data->boundary[i][j] != INTERIOR)
        mel[i].boundary[j] = bdry(mesh, data->boundary[i][j]);
      else
        mel[i].boundary[j] = nil;
    }
  }

  for (i = 0; i < nv; i++)
    bound[i] = INTERIOR;

  for (i = 0; i < ne; i++) {
    for (j = 0; j < N_NEIGH; j++) {
       for (k = 1; k < 4; k++)
         bound[data->mel_vertices[i][(j+k)%4]] =
    new_bound_val(mel[i].boundary[j], bound[data->mel_vertices[i][(j+k)%4]]);
    }
  }

  for (i = 0; i < ne; i++)
    for (j = 0; j < N_VERTICES; j++)
      mel[i].bound[j] = bound[data->mel_vertices[i][j]];

  MEM_FREE(bound, nv, S_CHAR);

  return;
}


/*--------------------------------------------------------------------------*/
/*  new_edge(): This procedure counts the number of faces                   */
/* sharing the edge and adds dof's at the midpoint if necessary. Another    */
/* important task is setting the boundary type of the edge depending on the */
/* types of these faces via new_bound_val().                                */
/*--------------------------------------------------------------------------*/

static int new_edge(MESH *mesh, MACRO_EL *mel, int mel_edge_no, int *n_neigh,
		    const BOUNDARY *(*bdry)(MESH *, int))
{
  FUNCNAME("new_edge"); 

  typedef const BOUNDARY **LIST_BOUND;

  MACRO_EL        *neigh;
  DOF             *dof[2], *edge_dof = nil;
  int             j, k, opp_v, mel_index, node = -1, edge_no;
  S_CHAR          bound = INTERIOR;
  int             max_no_list_el = 20;
  const BOUNDARY  *boundary = nil;
  LIST_BOUND      *list_bound = MEM_ALLOC(max_no_list_el, LIST_BOUND);

  static int  next_el[6][2] = {{3,2},{1,3},{1,2},{0,3},{0,2},{0,1}};

  mel_index = mel->index;
  edge_no = mel_edge_no;

  list_bound[0] = &(mel->boundary[N_FACES+edge_no]);
  
  if (mesh->n_dof[EDGE])
  {
    node = mesh->node[EDGE];
    if (mel->el->dof[node+edge_no])
    {
/*--------------------------------------------------------------------------*/
/*  edge was counted by another macro element and dof was added on the      */
/*  complete patch                                                          */
/*--------------------------------------------------------------------------*/
      MEM_FREE(list_bound, max_no_list_el, LIST_BOUND);
      return(0);
    }
    else
    {
      edge_dof = mel->el->dof[node+edge_no] = get_dof(mesh, EDGE);
    }
  }

  for (j = 0; j < 2; j++)
    dof[j] = mel->el->dof[vertex_of_edge[edge_no][j]];

/*--------------------------------------------------------------------------*/
/*  first look for all neighbours in one direction until a boundary is      */
/*  reached :-( or we are back to mel :-)                                   */
/*  if the index of a neighbour is smaller than the element index, the edge */
/*  is counted by this neighbour, return 0.                                 */
/*  If we are back to element, return 1, to count the edge                  */
/*--------------------------------------------------------------------------*/

  neigh = mel->neigh[next_el[edge_no][0]];
  opp_v = mel->opp_vertex[next_el[edge_no][0]];

  bound = new_bound_val(mel->boundary[next_el[edge_no][0]], bound);

  while (neigh  &&  neigh != mel)
  {
    if (mesh->n_dof[EDGE]) {
      TEST_EXIT(neigh->index > mel_index)
        ("neighbour index %d < element index %d\n", neigh->index, mel_index);
    }

    if (!mesh->n_dof[EDGE] &&  neigh->index < mel_index) {
      MEM_FREE(list_bound, max_no_list_el, LIST_BOUND);
      return(0);
    }
    
    for (j = 0; j < N_VERTICES; j++)
      if (neigh->el->dof[j] == dof[0])  break;
    for (k = 0; k < N_VERTICES; k++)
      if (neigh->el->dof[k] == dof[1])  break;
    edge_no = edge_of_dofs[j][k];

    if (*n_neigh == max_no_list_el) {
      list_bound = MEM_REALLOC(list_bound, max_no_list_el, max_no_list_el + 20,
                               LIST_BOUND);
      max_no_list_el += 20;
    } 

    list_bound[(*n_neigh)++] = &(neigh->boundary[N_FACES+edge_no]);

    if (mesh->n_dof[EDGE])
    {
      neigh->el->dof[node+edge_no] = edge_dof;
    }

    if (next_el[edge_no][0] != opp_v)
    {
      bound = new_bound_val(neigh->boundary[next_el[edge_no][0]], bound);

      opp_v = neigh->opp_vertex[next_el[edge_no][0]];
      neigh = neigh->neigh[next_el[edge_no][0]];
    }
    else
    {
      bound = new_bound_val(neigh->boundary[next_el[edge_no][1]], bound);

      opp_v = neigh->opp_vertex[next_el[edge_no][1]];
      neigh = neigh->neigh[next_el[edge_no][1]];
    }
  }
  if (!neigh)
  {
/*--------------------------------------------------------------------------*/
/*  while looping around the edge the domain's boundary was reached. Now,   */
/*  loop in the other direction to the domain's boundary		    */
/*--------------------------------------------------------------------------*/
    edge_no = mel_edge_no;

    neigh = mel->neigh[next_el[edge_no][1]];
    opp_v = mel->opp_vertex[next_el[edge_no][1]];

    bound = new_bound_val(mel->boundary[next_el[edge_no][1]], bound);

    while (neigh)
    {
      if (mesh->n_dof[EDGE]) {
	TEST_EXIT(neigh->index > mel_index)
	  ("neighbour index %d < element index %d\n", neigh->index,
	   mel_index);
      }

      if (neigh->index < mel_index) {
        MEM_FREE(list_bound, max_no_list_el, LIST_BOUND);
        return(0);
      }

      for (j = 0; j < N_VERTICES; j++)
	if (neigh->el->dof[j] == dof[0])  break;
      for (k = 0; k < N_VERTICES; k++)
	if (neigh->el->dof[k] == dof[1])  break;
      edge_no = edge_of_dofs[j][k];

      if (*n_neigh == max_no_list_el) {
        list_bound = MEM_REALLOC(list_bound, max_no_list_el,
                                 max_no_list_el + 20, LIST_BOUND);
        max_no_list_el += 20;
      } 

      list_bound[(*n_neigh)++] = &(neigh->boundary[N_FACES+edge_no]);

      if (mesh->n_dof[EDGE])
      {
	TEST_EXIT(!neigh->el->dof[node+edge_no])
	  ("dof %d on element %d is already set, but not on element %d\n",
	   node + edge_no, neigh->index, mel_index);

	neigh->el->dof[node+edge_no] = edge_dof;
      }

      if (next_el[edge_no][0] != opp_v)
      {
	bound = new_bound_val(neigh->boundary[next_el[edge_no][0]], bound);

	opp_v = neigh->opp_vertex[next_el[edge_no][0]];
	neigh = neigh->neigh[next_el[edge_no][0]];
      }
      else
      {
	bound = new_bound_val(neigh->boundary[next_el[edge_no][1]], bound);

	opp_v = neigh->opp_vertex[next_el[edge_no][1]];
	neigh = neigh->neigh[next_el[edge_no][1]];
      }
    }
  }

  if (bound)
    boundary = bdry(mesh, bound);
  else
    boundary = nil;

  for (j = 0; j < *n_neigh; j++)
    *(list_bound[j]) = boundary;

  MEM_FREE(list_bound, max_no_list_el, LIST_BOUND);
  return(1);
}

/*--------------------------------------------------------------------------*/
/*  fill_more_boundary_dofs(): This function adds dof's at the edges and    */
/* faces of macro elements in mesh using the function new_edge() above.     */
/* It also counts faces and edges.                                          */
/*--------------------------------------------------------------------------*/

static void fill_more_boundary_dofs(MESH *mesh,
                                    const BOUNDARY *(*bdry)(MESH *, int))
{
  FUNCNAME("fill_more_boundary_dofs");
  int        i, k, node, ne = mesh->n_elements;
  int        n_edges = 0, n_faces = 0, max_n_neigh = 0, n_neigh, ov;
  MACRO_EL   *mel, *neigh;

  mesh->max_edge_neigh = 8;
  node = mesh->node[FACE];

  /*  for (mel = mesh->first_macro_el; mel; mel = mel->next) */
  mel = mesh->first_macro_el;
  for (i = 0; i < ne; i++)
  {
    for (k = 0; k < N_EDGES; k++)
    {      
/*--------------------------------------------------------------------------*/
/*  check for not counted edges                                             */
/*--------------------------------------------------------------------------*/
      n_neigh = 1;
      if (new_edge(mesh, mel+i, k, &n_neigh, bdry))
      {
	n_edges++;
	max_n_neigh = MAX(max_n_neigh, n_neigh);
      }
    }

    for (k = 0; k < N_NEIGH; k++)
    {
      neigh = mel[i].neigh[k];
/*--------------------------------------------------------------------------*/
/*  face is counted and dof is added by the element with bigger index       */
/*--------------------------------------------------------------------------*/
      if (neigh  &&  (neigh->index > mel[i].index))  continue;
      
      n_faces++;

      if (mesh->n_dof[FACE])
      {
	TEST_EXIT(!mel[i].el->dof[node+k])
	  ("dof %d on element %d already set\n", node+k, mel[i].index);

	mel[i].el->dof[node+k] = get_dof(mesh, FACE);

	if (neigh)
	{
	  ov = mel[i].opp_vertex[k];
	  TEST_EXIT(!neigh->el->dof[node+ov])
	    ("dof %d on neighbour %d already set\n", node+ov, neigh->index);

	  neigh->el->dof[node+ov] = mel[i].el->dof[node+k];
	}
      }
    }
  }

  mesh->max_edge_neigh = MAX(mesh->max_edge_neigh, 2*max_n_neigh);
  mesh->n_edges = n_edges;
  mesh->n_faces = n_faces;

  return;
}

/*--------------------------------------------------------------------------*/
/* get_orientation(): This function calculates an elements orientation and  */
/* returns it as a signed integer suitable for the MACRO_EL orientation     */
/* field. Also called from read_mesh[_xdr].c.                               */
/*--------------------------------------------------------------------------*/

S_CHAR get_orientation(MACRO_EL *mel)
{
  int i,j;
  double a[3][3], s;

  for (i = 0; i < 3; i++)
    {
      for (j=0; j<3; j++) 
	{
	  a[i][j] = mel->coord[i+1][j] - mel->coord[0][j];
	}
    }

  s = (a[0][1]*a[1][2] - a[0][2]*a[1][1])*a[2][0]
    + (a[0][2]*a[1][0] - a[0][0]*a[1][2])*a[2][1]
    + (a[0][0]*a[1][1] - a[0][1]*a[1][0])*a[2][2];
  
  return (s >= 0) ? 1 : -1;
}

/*--------------------------------------------------------------------------*/
/* macro_test(): Check data for potential cycles during refinement          */
/* At the moment, a correction (and subsequent writing of a correct macro   */
/* data file) can only be done in 2D.                                       */
/*--------------------------------------------------------------------------*/

#if 0 /* not yet */
static int cycles(MACRO_DATA *);
#endif

void macro_test(MACRO_DATA *data, const char *new_name)
{
  FUNCNAME("macro_test");
#if 1
  WARNING("not implemented for 3d yet: no check is performed\n");
#else
  U_CHAR error_found = false;
  int i = -1;

  i = cycles(data);

  if (i >= 0)
  {
    error_found = true;
    WARNING("There is a cycle beginning in macro element %d.\n", i);
    ERROR_EXIT("Correction of cycles not implemented yet in 3D\n");
  }

  if (error_found && new_name) {
    MSG("Attempting to write corrected macro data to file %s...\n", new_name);
    write_alberta_macro(data, new_name);
  }
#endif
  return;
}


#include "macro_common.c"



