/*--------------------------------------------------------------------------*/
/* ALBERTA:  an Adaptive multi Level finite element toolbox using           */
/*           Bisectioning refinement and Error control by Residual          */
/*           Techniques for scientific Applications                         */
/*                                                                          */
/* file:     bas_fct_common.c                                               */
/*                                                                          */
/* description:  collecting information of all Lagrange elements            */
/*                                                                          */
/*--------------------------------------------------------------------------*/
/*               file is included by bas_fct.c's!!!                         */
/*--------------------------------------------------------------------------*/
/*                                                                          */
/*  authors:   Alfred Schmidt                                               */
/*             Zentrum fuer Technomathematik                                */
/*             Fachbereich 3 Mathematik/Informatik                          */
/*             Universitaet Bremen                                          */
/*             Bibliothekstr. 2                                             */
/*             D-28359 Bremen, Germany                                      */
/*                                                                          */
/*             Kunibert G. Siebert                                          */
/*             Institut fuer Mathematik                                     */
/*             Universitaet Augsburg                                        */
/*             Universitaetsstr. 14                                         */
/*             D-86159 Augsburg, Germany                                    */
/*                                                                          */
/*  http://www.alberta-fem.de/                                              */
/*                                                                          */
/*--------------------------------------------------------------------------*/
/*  (c) by A. Schmidt and K.G. Siebert (1996-2004)                          */
/*                                                                          */
/*     This program is free software; you can redistribute it and/or modify */
/*     it under the terms of the GNU General Public License as published by */
/*     the Free Software Foundation; either version 2 of the License, or    */
/*     any later version.                                                   */
/*                                                                          */
/*     This program is distributed in the hope that it will be useful,      */
/*     but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/*     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/*     GNU General Public License for more details.                         */
/*--------------------------------------------------------------------------*/

#if 0
static const REAL (*bary[5])[DIM+1] = {bary0, bary1, bary2, 
				       bary3, bary4};
static const int  *fct_at_node[5] = {fct_at_node0, fct_at_node1,
				     fct_at_node2, fct_at_node3,
				     fct_at_node4};
#endif

#define MAX_DEG 4
static const BAS_FCTS *lagrange[5] = {&lagrange0, &lagrange1, &lagrange2, 
				      &lagrange3, &lagrange4};

const BAS_FCTS *get_lagrange(int degree)
{
  FUNCNAME("get_lagrange");

  if (degree < 0 || degree > MAX_DEG)
  {
    ERROR("no lagrangian basis functions of degree %d\n", degree);
    return(nil);
  }

  return(lagrange[degree]);
}

/*--------------------------------------------------------------------------*/
/*  linked list of all used basis functions: Lagrange basisfunctions are    */
/*  always members of the list                                              */
/*--------------------------------------------------------------------------*/

struct all_bas_fcts
{
  const BAS_FCTS        *bas_fcts;
  struct all_bas_fcts   *next;
};

#if DIM == 1
static struct all_bas_fcts all_lagrange[2] = {
  {&lagrange0, all_lagrange+1}, {&lagrange1, nil}
 };
#else
static struct all_bas_fcts all_lagrange[5] = {
  {&lagrange0, all_lagrange+1}, {&lagrange1, all_lagrange+2},
  {&lagrange2, all_lagrange+3}, {&lagrange3, all_lagrange+4},
  {&lagrange4, nil}
 };
#endif

static  struct all_bas_fcts *first_bas_fcts = all_lagrange;

/*--------------------------------------------------------------------------*/
/*  add a set of new basis functions to the list; return true, if possible  */
/*  else false                                                              */
/*--------------------------------------------------------------------------*/

int new_bas_fcts(const BAS_FCTS * bas_fcts)
{
  FUNCNAME("new_bas_fcts");
  struct all_bas_fcts  *new_first;

  if (!bas_fcts)
  {
    ERROR("no basis functions specified; bas_fcts pointer to nil\n");
    return(0);
  }

  TEST_EXIT(bas_fcts->name)
    ("new basis functions must have name; bas_fcts->name pointer to nil\n");

  TEST_EXIT(strlen(bas_fcts->name))
    ("new basis functions must have a non empty name\n");
  TEST_EXIT(bas_fcts->degree >= 0)
    ("new basis functions must have a positive quadrature degree\n");

  TEST_EXIT(bas_fcts->phi)
    ("new basis functions: phi not set\n");
  TEST_EXIT(bas_fcts->grd_phi)
    ("new basis functions: grd_phi not set\n");
  TEST(bas_fcts->D2_phi) 
    ("Warning: new basis functions: D2_phi not set\n");
  TEST_EXIT(bas_fcts->get_dof_indices)
    ("new basis functions: get_dof_indices not set\n");
  TEST_EXIT(bas_fcts->get_bound)
    ("new basis functions: get_bound not set\n");
  TEST(bas_fcts->interpol) 
    ("Warning: new basis functions:  interpol not set\n");
  TEST(bas_fcts->interpol_d)
    ("Warning: new basis functions: interpol_d not set\n");

  for (new_first = first_bas_fcts; new_first; new_first = new_first->next)
  {
    if (!strcmp(bas_fcts->name, new_first->bas_fcts->name))
    {
      ERROR("basis functions with this name already exist\n");
      TEST_EXIT(bas_fcts == new_first->bas_fcts)
	("pointer to new and existing basis functions differ %x !%x\n",
	 bas_fcts, new_first->bas_fcts);
      ERROR("pointer to new and existing basis functions are the same\n");
      ERROR("ignoring new basis functions; taking old ones\n");
      return(0);
    }
  }

  new_first = MEM_ALLOC(1, struct all_bas_fcts);
  new_first->bas_fcts = bas_fcts;
  new_first->next = first_bas_fcts;
  first_bas_fcts = new_first;
  
  return(1);
}

/*--------------------------------------------------------------------------*/
/*  get a pointer to a set of basis functions from the list; identifier is  */
/*  the name of the basis functions;                                        */
/*  returns a pointer to the BAS_FCTS structure if a corresponding set was  */
/*  found in the list, else pointer to nil                                  */
/*--------------------------------------------------------------------------*/

const BAS_FCTS *get_bas_fcts(const char *name)
{
  FUNCNAME("get_bas_fcts");
  struct all_bas_fcts *bas_fcts;

  if (!name)
  {
    ERROR("no name specified; can not return pointer to basis functions\n");
    return(nil);
  }
  if (!strlen(name))
  {
    ERROR("empty name; can not return pointer to basis functions\n");
    return(nil);
  }
  for (bas_fcts = first_bas_fcts; bas_fcts; bas_fcts = bas_fcts->next)
    if (!strcmp(bas_fcts->bas_fcts->name, name))
      return(bas_fcts->bas_fcts);

  ERROR("basis functions with name %s not found in list of all functions\n");
  return(nil);
}


