/*--------------------------------------------------------------------------*/
/* ALBERTA:  an Adaptive multi Level finite element toolbox using           */
/*           Bisectioning refinement and Error control by Residual          */
/*           Techniques for scientific Applications                         */
/*                                                                          */
/* file:     dof_admin.c                                                    */
/*                                                                          */
/* description:  implementation of the administration of DOFs               */
/*                                                                          */
/*--------------------------------------------------------------------------*/
/*                                                                          */
/*  authors:   Alfred Schmidt                                               */
/*             Zentrum fuer Technomathematik                                */
/*             Fachbereich 3 Mathematik/Informatik                          */
/*             Universitaet Bremen                                          */
/*             Bibliothekstr. 2                                             */
/*             D-28359 Bremen, Germany                                      */
/*                                                                          */
/*             Kunibert G. Siebert                                          */
/*             Institut fuer Mathematik                                     */
/*             Universitaet Augsburg                                        */
/*             Universitaetsstr. 14                                         */
/*             D-86159 Augsburg, Germany                                    */
/*                                                                          */
/*  http://www.alberta-fem.de/                                              */
/*                                                                          */
/*--------------------------------------------------------------------------*/
/*  (c) by A. Schmidt and K.G. Siebert (1996-2004)                          */
/*                                                                          */
/*     This program is free software; you can redistribute it and/or modify */
/*     it under the terms of the GNU General Public License as published by */
/*     the Free Software Foundation; either version 2 of the License, or    */
/*     any later version.                                                   */
/*                                                                          */
/*     This program is distributed in the hope that it will be useful,      */
/*     but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/*     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/*     GNU General Public License for more details.                         */
/*--------------------------------------------------------------------------*/

#include "alberta.h"

/*--------------------------------------------------------------------------*/
/*      use a bit vector to indicate used/unused dofs			    */
/*      storage needed: one bit per dof                     		    */
/*--------------------------------------------------------------------------*/

#if DOF_FREE_SIZE == 32

DOF_FREE_UNIT dof_free_bit[DOF_FREE_SIZE] =
      { 0x00000001,  0x00000002,  0x00000004,  0x00000008,
	0x00000010,  0x00000020,  0x00000040,  0x00000080, 
	0x00000100,  0x00000200,  0x00000400,  0x00000800, 
	0x00001000,  0x00002000,  0x00004000,  0x00008000, 
	0x00010000,  0x00020000,  0x00040000,  0x00080000, 
	0x00100000,  0x00200000,  0x00400000,  0x00800000, 
	0x01000000,  0x02000000,  0x04000000,  0x08000000, 
	0x10000000,  0x20000000,  0x40000000,  0x80000000 };
DOF_FREE_UNIT dof_free_mod[DOF_FREE_SIZE] =
      { 0xFFFFFFFF,  0xFFFFFFFE,  0xFFFFFFFC,  0xFFFFFFF8,
	0xFFFFFFF0,  0xFFFFFFE0,  0xFFFFFFC0,  0xFFFFFF80, 
	0xFFFFFF00,  0xFFFFFE00,  0xFFFFFC00,  0xFFFFF800, 
	0xFFFFF000,  0xFFFFE000,  0xFFFFC000,  0xFFFF8000, 
	0xFFFF0000,  0xFFFE0000,  0xFFFC0000,  0xFFF80000, 
	0xFFF00000,  0xFFE00000,  0xFFC00000,  0xFF800000, 
	0xFF000000,  0xFE000000,  0xFC000000,  0xF8000000, 
	0xF0000000,  0xE0000000,  0xC0000000,  0x80000000 };
#else

DOF_FREE_UNIT dof_free_bit[DOF_FREE_SIZE] =
      { 0x01,  0x02,  0x04,  0x08,
	0x10,  0x20,  0x40,  0x80 };

DOF_FREE_UNIT dof_free_mod[DOF_FREE_SIZE] =
      { 0xFF,  0xFE,  0xFC,  0xF8,
	0xF0,  0xE0,  0xC0,  0x80 };

#endif

/*--------------------------------------------------------------------------*/
/*  SIZE_INCREMENT:                                                         */
/*  default increment for DOF_VECs etc. in enlarge_dof_lists                */
/*--------------------------------------------------------------------------*/
#define SIZE_INCREMENT (DOF_FREE_SIZE * 32)

/*--------------------------------------------------------------------------*/
#if 0
static const int *get_int_vec(const EL *el, const DOF_INT_VEC *vec,
			      int *ivec)
{
  FUNCNAME("get_int_vec");
  int            i, n_bas_fcts;
  const DOF      *index;
  int            *v = nil, *rvec;

  static int     *local_vec = nil;
  static int     max_size = 0;
  
  GET_DOF_VEC(v, vec);
  n_bas_fcts = vec->fe_space->bas_fcts->n_bas_fcts;
  if (ivec)
  {
    rvec = ivec;
  }
  else
  {
    if (max_size < n_bas_fcts)
    {
      local_vec = MEM_REALLOC(local_vec, max_size, n_bas_fcts, int);
      max_size = n_bas_fcts;
    }
    rvec = local_vec;
  }
  index = vec->fe_space->bas_fcts->get_dof_indices(el, vec->fe_space->admin, nil);

  for (i = 0; i < n_bas_fcts; i++)
    rvec[i] = v[index[i]];

  return((const int *) rvec);
}

static const REAL *get_real_vec(const EL *el, const DOF_REAL_VEC *vec, 
				   REAL *Rvec)
{
  FUNCNAME("get_real_vec");
  int            i, n_bas_fcts;
  const DOF      *index;
  REAL           *v = nil, *rvec;

  static REAL    *local_vec = nil;
  static int     max_size = 0;
  
  GET_DOF_VEC(v, vec);
  n_bas_fcts = vec->fe_space->bas_fcts->n_bas_fcts;
  if (Rvec)
  {
    rvec = Rvec;
  }
  else
  {
    if (max_size < n_bas_fcts)
    {
      local_vec = MEM_REALLOC(local_vec, max_size, n_bas_fcts, REAL);
      max_size = n_bas_fcts;
    }
    rvec = local_vec;
  }
  index = vec->fe_space->bas_fcts->get_dof_indices(el, vec->fe_space->admin, nil);

  for (i = 0; i < n_bas_fcts; i++)
    rvec[i] = v[index[i]];

  return((const REAL *) rvec);
}

static const REAL_D *get_real_d_vec(const EL *el, const DOF_REAL_D_VEC *vec,
				    REAL_D *Rvec)
{
  FUNCNAME("get_real_d_vec");
  int            i, j, n_bas_fcts;
  const DOF      *index;
  REAL_D         *v = nil, *rvec;

  static REAL_D  *local_vec = nil;
  static int     max_size = 0;
  
  GET_DOF_VEC(v, vec);
  n_bas_fcts = vec->fe_space->bas_fcts->n_bas_fcts;
  if (Rvec)
  {
    rvec = Rvec;
  }
  else
  {
    if (max_size < n_bas_fcts)
    {
      local_vec = MEM_REALLOC(local_vec, max_size, n_bas_fcts, REAL_D);
      max_size = n_bas_fcts;
    }
    rvec = local_vec;
  }
  index = vec->fe_space->bas_fcts->get_dof_indices(el, vec->fe_space->admin, nil);

  for (i = 0; i < n_bas_fcts; i++)
    for (j = 0; j < DIM_OF_WORLD; j++)
      rvec[i][j] = v[index[i]][j];

  return((const REAL_D *) rvec);
}

static const U_CHAR *get_uchar_vec(const EL *el, const DOF_UCHAR_VEC *vec,
				      U_CHAR *uvec)
{
  FUNCNAME("get_uchar_vec");
  int            i, n_bas_fcts;
  const DOF      *index;
  U_CHAR         *v = nil, *rvec;

  static U_CHAR  *local_vec = nil;
  static int     max_size = 0;
  
  GET_DOF_VEC(v, vec);
  n_bas_fcts = vec->fe_space->bas_fcts->n_bas_fcts;
  if (uvec)
  {
    rvec = uvec;
  }
  else
  {
    if (max_size < n_bas_fcts)
    {
      local_vec = MEM_REALLOC(local_vec, max_size, n_bas_fcts, U_CHAR);
      max_size = n_bas_fcts;
    }
    rvec = local_vec;
  }
  index = vec->fe_space->bas_fcts->get_dof_indices(el, vec->fe_space->admin, nil);

  for (i = 0; i < n_bas_fcts; i++)
    rvec[i] = v[index[i]];

  return((const U_CHAR *) rvec);
}

static const S_CHAR *get_schar_vec(const EL *el, const DOF_SCHAR_VEC *vec,
				   S_CHAR *svec)
{
  FUNCNAME("get_schar_vec");
  int            i, n_bas_fcts;
  const DOF      *index;
  S_CHAR         *v = nil, *rvec;

  static S_CHAR  *local_vec = nil;
  static int     max_size = 0;
  
  GET_DOF_VEC(v, vec);
  n_bas_fcts = vec->fe_space->bas_fcts->n_bas_fcts;
  if (svec)
  {
    rvec = svec;
  }
  else
  {
    if (max_size < n_bas_fcts)
    {
      local_vec = MEM_REALLOC(local_vec, max_size, n_bas_fcts, S_CHAR);
      max_size = n_bas_fcts;
    }
    rvec = local_vec;
  }
  index = vec->fe_space->bas_fcts->get_dof_indices(el, vec->fe_space->admin, nil);

  for (i = 0; i < n_bas_fcts; i++)
    rvec[i] = v[index[i]];

  return((const S_CHAR *) rvec);
}

void add_bas_fcts_to_admin(DOF_ADMIN *admin, const BAS_FCTS *bas_fcts)
{
  FUNCNAME("add_bas_fcts_to_admin");
  BAS_FCTS  *basfcts;

  TEST_EXIT(admin)("no admin specified; admin nil pointer\n");
  TEST_EXIT(bas_fcts)("no bas_fcts specified; bas_fcts nil pointer\n");

  basfcts = (BAS_FCTS *) bas_fcts;

  TEST_EXIT(basfcts->get_dof_indices)
    ("no function 'get_dof_indices' in bas_fcts %s structure\n",
     NAME(bas_fcts));
  TEST_EXIT(basfcts->get_bound)
    ("no function 'get_bound' in admin->bas_fcts %s structure\n",
     NAME(bas_fcts));
  TEST_EXIT(basfcts->interpol)
    ("no function 'interpol' in admin->bas_fcts %s structure\n",
     NAME(bas_fcts));

  if (!basfcts->get_int_vec)
    basfcts->get_int_vec = get_int_vec;

  if (!basfcts->get_real_vec)
    basfcts->get_real_vec = get_real_vec;

  if (!basfcts->get_real_d_vec)
    basfcts->get_real_d_vec = get_real_d_vec;

  if (!basfcts->get_uchar_vec)
    basfcts->get_uchar_vec = get_uchar_vec;

  if (!basfcts->get_schar_vec)
    basfcts->get_schar_vec = get_schar_vec;

  admin->bas_fcts = bas_fcts;

  return;
}
#endif

/*--------------------------------------------------------------------------*/

void add_dof_admin_to_mesh(DOF_ADMIN *admin, MESH *mesh)
{
  FUNCNAME("add_dof_admin_to_mesh");
  int i, n;

  admin->mesh = mesh;
  n = mesh->n_dof_admin;
  if ((n > 0) & (mesh->dof_admin == nil))
    ERROR_EXIT("no mesh->dof_admin but n_dof_admin=%d\n", n);
  if ((n <= 0) & (mesh->dof_admin != nil))
    ERROR_EXIT("found mesh->dof_admin but n_dof_admin=%d\n", n);

  for (i = 0; i < n; i++)
    if (mesh->dof_admin[i] == admin) {
      ERROR("admin %s is already associated to mesh %s\n",
	    NAME(admin), NAME(mesh));
    }

  mesh->dof_admin = MEM_REALLOC(mesh->dof_admin, n, n+1, DOF_ADMIN *);
  n++;

  mesh->dof_admin[n-1] = admin;
  mesh->n_dof_admin = n;

  mesh->n_dof_el = 0;

  admin->n0_dof[VERTEX] = mesh->n_dof[VERTEX];
  mesh->n_dof[VERTEX]  += admin->n_dof[VERTEX];
  mesh->n_dof_el += N_VERTICES*mesh->n_dof[VERTEX];
#if DIM > 1
  admin->n0_dof[EDGE]   = mesh->n_dof[EDGE];
  mesh->n_dof[EDGE]    += admin->n_dof[EDGE];
  mesh->n_dof_el += N_EDGES*mesh->n_dof[EDGE];
#endif
  admin->n0_dof[CENTER] = mesh->n_dof[CENTER];
  mesh->n_dof[CENTER]  += admin->n_dof[CENTER];
  mesh->n_dof_el += mesh->n_dof[CENTER];

  mesh->node[VERTEX]  = 0;
  mesh->n_node_el     = N_VERTICES;

#if DIM > 1
  mesh->node[EDGE]    = mesh->n_node_el;
  if (mesh->n_dof[EDGE] > 0) mesh->n_node_el += N_EDGES;
#endif

#if DIM==3
  admin->n0_dof[FACE] = mesh->n_dof[FACE];
  mesh->n_dof[FACE]  += admin->n_dof[FACE];
  mesh->n_dof_el     += N_FACES * mesh->n_dof[FACE];
  mesh->node[FACE]    = mesh->n_node_el;
  if (mesh->n_dof[FACE] > 0) mesh->n_node_el += N_FACES;
#endif

  mesh->node[CENTER]    = mesh->n_node_el;
  if (mesh->n_dof[CENTER] > 0) mesh->n_node_el += 1;

  return;
}

/*--------------------------------------------------------------------------*/

void free_dof_index(DOF_ADMIN  *admin, int dof)
{
  FUNCNAME("free_dof_index");
  int          i, j, col, col2;
  unsigned int iunit, ibit;
  DOF_MATRIX   *matrix;
  MATRIX_ROW   *row, *row2;

  TEST_EXIT(admin)("no admin\n");
  TEST_EXIT(admin->used_count > 0)("no dofs in use\n");
  TEST_EXIT((dof >= 0)&&(dof < admin->size))("invalid dof index %d\n",dof);

  for (matrix = admin->dof_matrix; matrix; matrix = matrix->next)
  {
    if ((row = matrix->matrix_row[dof]))
    {
      do
      {
	for (i=0; i<ROW_LENGTH; i++)
	{
	  col = row->col[i];
	  if (ENTRY_USED(col))
	  {
	    if (col != dof)   /* remove symmetric entry if exists */
	    {
	      row2 = matrix->matrix_row[col];
	      while (row2)
	      {
		for (j=0; j<ROW_LENGTH; j++)
		{
		  col2 = row2->col[j];
		  if (col2 == dof) 
		  {
		    row2->col[j] = UNUSED_ENTRY;
		  }
		  else if (col2 == NO_MORE_ENTRIES)
		  {
		    break;
		  }
		}
		row2 = row2->next;
	      };
	    }
	  }
	  else if (col == NO_MORE_ENTRIES)
	  {
	    break;
	  }
	}

	row2 = row;
	row = row->next;
	free_matrix_row(matrix->fe_space, row2);
	
      } while(row);

      matrix->matrix_row[dof] = nil;
    }
  }

  iunit = dof / DOF_FREE_SIZE;
  ibit  = dof % DOF_FREE_SIZE;
  admin->dof_free[iunit] |= dof_free_bit[ibit];
  if (admin->first_hole > iunit) admin->first_hole = iunit;

  admin->used_count--;
  admin->hole_count++;
}

/*--------------------------------------------------------------------------*/

int get_dof_index(DOF_ADMIN *admin)
{
  FUNCNAME("get_dof_index");
  int       dof = 0, ibit;
  unsigned int i;
  TEST_EXIT(admin)("no admin\n");

  if (admin->first_hole < admin->dof_free_size) {
    for (ibit = 0; ibit < DOF_FREE_SIZE; ibit++) {
      if (admin->dof_free[admin->first_hole] & dof_free_bit[ibit]) {
	admin->dof_free[admin->first_hole] ^= dof_free_bit[ibit];
	dof = DOF_FREE_SIZE * admin->first_hole + ibit;
	if (admin->dof_free[admin->first_hole] == 0) {
	  for (i = admin->first_hole+1; i < admin->dof_free_size; i++) {
	    if (admin->dof_free[i]) {
	      break;
	    }
	  }
	  admin->first_hole = i;
	}
	break;
      }
    }
    TEST_EXIT (ibit < DOF_FREE_SIZE)("no free bit in first_hole ?\n");
  }
  else {
    enlarge_dof_lists(admin, 0);
    TEST_EXIT(admin->first_hole < admin->dof_free_size)
      ("no free entry after enlarge_dof_lists\n");
    TEST_EXIT(admin->dof_free[admin->first_hole] & dof_free_bit[0])
      ("no free bit 0\n");
    admin->dof_free[admin->first_hole] ^= dof_free_bit[0];
    dof = DOF_FREE_SIZE * admin->first_hole;
  }

  admin->used_count++;
  if (admin->hole_count > 0) admin->hole_count--;
  admin->size_used = MAX(admin->size_used, dof+1);

  return(dof);
}

/*--------------------------------------------------------------------------*/

void enlarge_dof_lists(DOF_ADMIN *admin, int minsize)
{
  FUNCNAME("enlarge_dof_lists");
  int              old, new, i, j, new_free_size;
  DOF_INT_VEC      *iv;
  DOF_DOF_VEC      *dv;
  DOF_UCHAR_VEC    *uv;
  DOF_SCHAR_VEC    *sv;
  DOF_REAL_VEC     *rv;
  DOF_REAL_D_VEC   *rdv;
  DOF_MATRIX       *mat;

  TEST_EXIT(admin)("no admin\n");

  old = admin->size;
  if (minsize > 0) {
    if (old > minsize) return;
  }

  new = MAX(minsize, admin->size + SIZE_INCREMENT);

  new += (DOF_FREE_SIZE - (new % DOF_FREE_SIZE)) % DOF_FREE_SIZE;
  admin->size = new;

  new_free_size = new / DOF_FREE_SIZE;
  admin->dof_free = MEM_REALLOC(admin->dof_free, admin->dof_free_size,
				new_free_size, DOF_FREE_UNIT);
  for (i = admin->dof_free_size; i < new_free_size; i++)
    admin->dof_free[i] = DOF_UNIT_ALL_FREE;
  admin->first_hole = admin->dof_free_size;
  admin->dof_free_size = new_free_size;

  /* enlarge all vektors and matrices          */
  /* but int_dof_vecs don't have to be changed */

  for (iv = admin->dof_int_vec; iv; iv = iv->next) {
    if (iv->size < new) {
      iv->vec = MEM_REALLOC(iv->vec, iv->size, new, int);
      for (i = iv->size; i < new; i++) iv->vec[i] = 0;
      iv->size   = new;
    }
  }

  for (dv = admin->dof_dof_vec; dv; dv = dv->next) {
    if (dv->size < new) {
      dv->vec = MEM_REALLOC(dv->vec, dv->size, new, DOF);
      for (i = dv->size; i < new; i++) dv->vec[i] = -1;
      dv->size = new;
    }
  }

  for (uv = admin->dof_uchar_vec; uv; uv = uv->next) {
    if (uv->size < new) {
      uv->vec = MEM_REALLOC(uv->vec, uv->size, new, U_CHAR);
      for (i = uv->size; i < new; i++) uv->vec[i] = 0;
      uv->size   = new;
    }
  }

  for (sv = admin->dof_schar_vec; sv; sv = sv->next) {
    if (sv->size < new) {
      sv->vec = MEM_REALLOC(sv->vec, old, new, S_CHAR);
      for (i = sv->size; i < new; i++) sv->vec[i] = 0;
      sv->size   = new;
    }
  }

  for (rv = admin->dof_real_vec; rv; rv = rv->next) {
    if (rv->size < new) {
      rv->vec = MEM_REALLOC(rv->vec, rv->size, new, REAL);
      for (i = rv->size; i < new; i++) rv->vec[i] = 0.0;
      rv->size = new;
    }
  }

  for (rdv = admin->dof_real_d_vec; rdv; rdv = rdv->next) {
    if (rdv->size < new) {
      rdv->vec = MEM_REALLOC(rdv->vec, rdv->size, new, REAL_D);
      for (i = rdv->size; i < new; i++)
	for (j=0; j<DIM_OF_WORLD; j++)
	  rdv->vec[i][j] = 0.0;
      rdv->size = new;
    }
  }

  for (mat = admin->dof_matrix; mat; mat = mat->next) {
    if (mat->size < new) {
      mat->matrix_row = MEM_REALLOC(mat->matrix_row, mat->size, new,
				    MATRIX_ROW *);
      for (i = mat->size; i < new; i++)
	mat->matrix_row[i] = nil;
      mat->size = new;
    }
  }
}


/*--------------------------------------------------------------------------*/
/*  dof_compress: remove holes in dof vectors                               */
/*--------------------------------------------------------------------------*/

static int       *new_dof  = nil;
static DOF_ADMIN *compress_admin;
static int       *g_n_dof  = nil;
static int       *g_n0_dof = nil;
static int       *g_node   = nil;

/*--------------------------------------------------------------------------*/
/*  ATTENTION:                                                              */
/*  new_dof_fct() destroys new_dof !!!!!!!!!!                               */
/*  should be used only at the end of dof_compress()!!!!!                   */
/*--------------------------------------------------------------------------*/

/* CHANGE_DOFS_1 changes old dofs to NEGATIVE new dofs */

#define CHANGE_DOFS_1(el) \
      dof = el->dof[n0+i] + nd0; \
      for (j = 0; j < nd; j++) { \
	if ((k = dof[j]) >= 0) { \
	  /* do it only once! (dofs are visited more than once) */ \
	  dof[j] = - new_dof[k]; \
	} }

/* CHANGE_DOFS_2 changes NEGATIVE new dofs to POSITIVE */

#define CHANGE_DOFS_2(el) \
      dof = el->dof[n0+i] + nd0; \
      for (j = 0; j < nd; j++) { \
	if ((k = dof[j]) < 0) { \
	  /* do it only once! (dofs are visited more than once) */ \
	  dof[j] = - k; \
	} }

static void new_dof_fct1(const EL_INFO *el_info)
{
  EL      *el = el_info->el;
  int     i, j, k, n0, nd, nd0;
  DOF     *dof;
  
  if ((nd = g_n_dof[VERTEX]))  {
    nd0 = g_n0_dof[VERTEX];
    n0 = g_node[VERTEX];
    for (i = 0; i < N_VERTICES; i++) {
      CHANGE_DOFS_1(el);
    }
  }

#if DIM > 1
  if ((nd = g_n_dof[EDGE]))
  {
    nd0 = g_n0_dof[EDGE];
    n0 = g_node[EDGE];
    for (i = 0; i < N_EDGES; i++) {
      CHANGE_DOFS_1(el);
    }
  }
#endif

#if DIM==3
  if ((nd = g_n_dof[FACE]))  {
    nd0 = g_n0_dof[FACE];
    n0 = g_node[FACE];
    for (i = 0; i < N_FACES; i++) {
      CHANGE_DOFS_1(el);
    }
  }
#endif

  if ((nd = g_n_dof[CENTER]))  {
    nd0 = g_n0_dof[CENTER];
    n0 = g_node[CENTER];
    i = 0;          /* only one center */
    CHANGE_DOFS_1(el);
  }
}


static void new_dof_fct2(const EL_INFO *el_info)
{
  EL      *el = el_info->el;
  int     i, j, k, n0, nd, nd0;
  DOF     *dof;
  
  if ((nd = g_n_dof[VERTEX]))  {
    nd0 = g_n0_dof[VERTEX];
    n0 = g_node[VERTEX];
    for (i = 0; i < N_VERTICES; i++) {
      CHANGE_DOFS_2(el);
    }
  }

#if DIM > 1
  if ((nd = g_n_dof[EDGE]))
  {
    nd0 = g_n0_dof[EDGE];
    n0 = g_node[EDGE];
    for (i = 0; i < N_EDGES; i++) 
    {
      CHANGE_DOFS_2(el);
    }
  }
#endif

#if DIM==3
  if ((nd = g_n_dof[FACE]))  {
    nd0 = g_n0_dof[FACE];
    n0 = g_node[FACE];
    for (i = 0; i < N_FACES; i++) {
      CHANGE_DOFS_2(el);
    }
  }
#endif

  if ((nd = g_n_dof[CENTER]))  {
    nd0 = g_n0_dof[CENTER];
    n0 = g_node[CENTER];
    i = 0;          /* only one center */
    CHANGE_DOFS_2(el);
  }
}

#undef CHANGE_DOFS_1
#undef CHANGE_DOFS_2

/*--------------------------------------------------------------------------*/

void dof_compress(MESH *mesh)
{
  FUNCNAME("dof_compress");
  int             i,j,k,size,first,last=0,col,n,iadmin;
  FLAGS           fill_flag;
  DOF_INT_VEC     *div;
  DOF_DOF_VEC     *ddv;
  DOF_UCHAR_VEC   *duv;
  DOF_SCHAR_VEC   *dsv;
  DOF_REAL_D_VEC  *drdv;
  DOF_REAL_VEC    *drv;
  DOF_MATRIX      *dm;
  MATRIX_ROW      *row, *row_next;

  TEST_EXIT(mesh)("no Mesh\n");

  g_node = mesh->node;

  for (iadmin = 0; iadmin < mesh->n_dof_admin; iadmin++)
  {
    TEST_EXIT((compress_admin = mesh->dof_admin[iadmin]))
      ("no dof_admin[%d] in mesh\n", iadmin);

    if ((size = compress_admin->size) < 1) continue;
    if (compress_admin->used_count < 1)    continue;
    if (compress_admin->hole_count < 1)    continue;
    
    g_n_dof = compress_admin->n_dof;
    g_n0_dof = compress_admin->n0_dof;

    new_dof = MEM_ALLOC(size, int);
    for (i=0; i<size; i++) {                       /* mark everything unused */
      new_dof[i] = -1;
    }

    FOR_ALL_DOFS(compress_admin, new_dof[dof] = 1);

    n = 0;
    for (i = 0; i < size; i++) {             /* create a MONOTONE compress */
      if (new_dof[i] == 1) {
	new_dof[i] = n++;
	last = i;
      }
    }

    TEST_EXIT(n == compress_admin->used_count)("count %d != used_count %d\n",
					       n, compress_admin->used_count);

    {
      int n1 = n / DOF_FREE_SIZE;

      for (i = 0; i < n1; i++)
	compress_admin->dof_free[i] = 0;
      if (n1 < (int)compress_admin->dof_free_size)
	compress_admin->dof_free[n1] = dof_free_mod[n % DOF_FREE_SIZE];
      for (i = n1+1; i < (int)compress_admin->dof_free_size; i++)
	compress_admin->dof_free[i] = DOF_UNIT_ALL_FREE;
      compress_admin->first_hole = n1;
    }

    compress_admin->hole_count = 0;
    compress_admin->size_used  = n;

    first = 0;
    for (i=0; i<size; i++) {
      if ((new_dof[i] < (int)i) && (new_dof[i] >= 0)) {
	first = i;
	break;
      }
    }
    if (last >= first) {
      last++;   /* for (i = first; i < last; i++) ... */

      /* compress all vectors associated to admin */

      for (div = compress_admin->dof_int_vec; div; div = div->next) {
	for (i=first; i<last; i++) {
	  if ((k = new_dof[i]) >= 0)
	    div->vec[k] = div->vec[i];
	}
      }
  
      for (ddv = compress_admin->dof_dof_vec; ddv; ddv = ddv->next) {
	for (i=first; i<last; i++) {
	  if ((k = new_dof[i]) >= 0)
	    ddv->vec[k] = ddv->vec[i];
	}
	for (i=0; i<n; i++) {
	  if ((k = new_dof[ddv->vec[i]]) >= 0)
	    ddv->vec[i] = k;
	}
      }
  
      for (ddv = compress_admin->int_dof_vec; ddv; ddv = ddv->next) {
	for (i=0; (int)i < ddv->size; i++) {
	  if (ddv->vec[i] >= 0) 
	    if ((k = new_dof[ddv->vec[i]]) >= 0)
	      ddv->vec[i] = k;
	}
      }
  
      for (duv = compress_admin->dof_uchar_vec; duv; duv = duv->next) {
	for (i=first; i<last; i++) {
	  if ((k = new_dof[i]) >= 0)
	    duv->vec[k] = duv->vec[i];
	}
      }
  
      for (dsv = compress_admin->dof_schar_vec; dsv; dsv = dsv->next) {
	for (i=first; i<last; i++) {
	  if ((k = new_dof[i]) >= 0)
	    dsv->vec[k] = dsv->vec[i];
	}
      }
  
      for (drv = compress_admin->dof_real_vec; drv; drv = drv->next) {
	for (i=first; i<last; i++) {
	  if ((k = new_dof[i]) >= 0)
	    drv->vec[k] = drv->vec[i];
	}
      }
  
      for (drdv = compress_admin->dof_real_d_vec; drdv; drdv = drdv->next) {
	for (i=first; i<last; i++) {
	  if ((k = new_dof[i]) >= 0)
	    for (j=0; j<DIM_OF_WORLD; j++)
	      drdv->vec[k][j] = drdv->vec[i][j];
	}
      }
  
      for (dm = compress_admin->dof_matrix; dm; dm = dm->next) {
	for (i=first; i<last; i++) {
	  if ((k = new_dof[i]) >= 0) {
	    /* free dm->matrix_row[k]; */
	    for (row = dm->matrix_row[k]; row; row = row_next) {
	      row_next = row->next;
	      free_matrix_row(dm->fe_space, row);
	    }
	    dm->matrix_row[k] = dm->matrix_row[i];
	    dm->matrix_row[i] = nil;
	  }
	}
	for (i=0; i<n; i++) { /* change columns */
	  for (row = dm->matrix_row[i]; row; row = row->next) {
	    for (j=0; j<ROW_LENGTH; j++) {
	      col = row->col[j];
	      if (ENTRY_USED(col)) row->col[j] = new_dof[col];
	    }
	  }
	}
      }
      
      /* now, change dofs in MESH's nodes */
      /* new_dof_fct destroys new_dof[] !!! */

      if (mesh->preserve_coarse_dofs) {
	fill_flag = CALL_EVERY_EL_PREORDER | FILL_NOTHING;
      }
      else {
	fill_flag = CALL_LEAF_EL | FILL_NOTHING;
      }
      mesh_traverse(mesh, -1, fill_flag, new_dof_fct1);
      mesh_traverse(mesh, -1, fill_flag, new_dof_fct2);

    }

    MEM_FREE(new_dof, size, int);
    new_dof = nil;
  }
  return;
}


/*--------------------------------------------------------------------------*/
/*  matrix administration                            			    */
/*--------------------------------------------------------------------------*/
/*  add_dof_matrix_to_admin():       enter new matrix to DOF_ADMIN list     */
/*                                   (re)alloc matrix_row, el_* if necessary*/
/*  add_dof_int_vec_to_admin():      enter new vec to DOF_ADMIN list        */
/*  add_int_dof_vec_to_admin():      enter new vec to DOF_ADMIN list        */
/*  add_dof_dof_vec_to_admin():      enter new vec to DOF_ADMIN list        */
/*  add_dof_uchar_vec_to_admin():    enter new vec to DOF_ADMIN list        */
/*  add_dof_schar_vec_to_admin():    enter new vec to DOF_ADMIN list        */
/*  add_dof_real_vec_to_admin():     enter new vec to DOF_ADMIN list        */
/*  add_dof_real_d_vec_to_admin():   enter new vec to DOF_ADMIN list        */
/*                                   (re)alloc vector if necessary          */
/*                                                                          */
/*  update_dof_matrix():             add/substract el_mat to/from matrix    */
/*--------------------------------------------------------------------------*/

#define CHECK_IF_PRESENT(admin, TYPE, list, vec) \
  {TYPE *v; for (v=admin->list; v; v = v->next) \
      if (v==vec) {ERROR("dof_vec %s already associated to admin %s\n",\
			 NAME(vec), NAME(admin)); return;} }

void add_dof_int_vec_to_admin(DOF_INT_VEC *vec, DOF_ADMIN *admin)
{
  FUNCNAME("add_dof_int_vec_to_admin");

  if (!vec) {
    MSG("no vec\n");
    return;
  }

  CHECK_IF_PRESENT(admin, DOF_INT_VEC, dof_int_vec, vec);

  if (vec->size < admin->size) {
    vec->vec = MEM_REALLOC(vec->vec, vec->size, admin->size, int);
    vec->size = admin->size;
  }
  
  vec->next = admin->dof_int_vec;
  admin->dof_int_vec = vec;
  /* vec->admin = admin; */
}

void remove_dof_int_vec_from_admin(DOF_INT_VEC *vec)
{
  FUNCNAME("remove_dof_int_vec_from_admin");
  DOF_ADMIN    *admin;
  DOF_INT_VEC  *vec_last;

  if (vec->fe_space && (admin = (DOF_ADMIN *)vec->fe_space->admin))
  {
    if (admin->dof_int_vec == vec)
      admin->dof_int_vec = vec->next;
    else
    {
      vec_last = admin->dof_int_vec;
      while (vec_last &&  vec_last->next != vec)
	vec_last = vec_last->next;
      
      if (!vec_last)
      {
	ERROR("dof_int_vec %s not in list of dof admin %s found\n",
	      NAME(vec), NAME(admin));
      }
      else
	vec_last->next = vec->next;
    }
  }
  return;
}

void add_dof_dof_vec_to_admin(DOF_DOF_VEC *vec, DOF_ADMIN *admin)
{
  FUNCNAME("add_dof_dof_vec_to_admin");
  
  if (!vec) {
    MSG("no vec\n");
    return;
  }

  CHECK_IF_PRESENT(admin, DOF_DOF_VEC, dof_dof_vec, vec);

  if (vec->size < admin->size) {
    vec->vec = MEM_REALLOC(vec->vec, vec->size, admin->size, DOF);
    vec->size = admin->size;
  }
  
  vec->next = admin->dof_dof_vec;
  admin->dof_dof_vec = vec;
  /* vec->admin = admin; */
}

void remove_dof_dof_vec_from_admin(DOF_DOF_VEC *vec)
{
  FUNCNAME("remove_dof_dof_vec_from_admin");
  DOF_ADMIN    *admin;
  DOF_DOF_VEC  *vec_last;

  if (vec->fe_space && (admin = (DOF_ADMIN *)vec->fe_space->admin))
  {
    if (admin->dof_dof_vec == vec)
      admin->dof_dof_vec = vec->next;
    else
    {
      vec_last = admin->dof_dof_vec;
      while (vec_last &&  vec_last->next != vec)
	vec_last = vec_last->next;
      
      if (!vec_last)
      {
	ERROR("dof_dof_vec %s not in list of dof admin %s found\n",
	      NAME(vec), NAME(admin));
      }
      else
	vec_last->next = vec->next;
    }
  }
  return;
}

void add_int_dof_vec_to_admin(DOF_DOF_VEC *vec, DOF_ADMIN *admin)
{
  FUNCNAME("add_int_dof_vec_to_admin");
  
  if (!vec) {
    MSG("no vec\n");
    return;
  }

  CHECK_IF_PRESENT(admin, DOF_DOF_VEC, int_dof_vec, vec);

  /* don't change size */

  vec->next = admin->int_dof_vec;
  admin->int_dof_vec = vec;
  /* vec->admin = admin; */
}

void remove_int_dof_vec_from_admin(DOF_DOF_VEC *vec)
{
  FUNCNAME("remove_int_dof_vec_from_admin");
  DOF_ADMIN    *admin;
  DOF_DOF_VEC  *vec_last;

  if (vec->fe_space && (admin = (DOF_ADMIN *)vec->fe_space->admin))
  {
    if (admin->int_dof_vec == vec)
      admin->int_dof_vec = vec->next;
    else
    {
      vec_last = admin->int_dof_vec;
      while (vec_last &&  vec_last->next != vec)
	vec_last = vec_last->next;
      
      if (!vec_last)
      {
	ERROR("int_dof_vec %s not in list of dof admin %s found\n",
	      NAME(vec), NAME(admin));
      }
      else
	vec_last->next = vec->next;
    }
  }
  return;
}

void add_dof_uchar_vec_to_admin(DOF_UCHAR_VEC *vec, DOF_ADMIN *admin)
{
  FUNCNAME("add_dof_uchar_vec_to_admin");
  
  if (!vec) {
    MSG("no vec\n");
    return;
  }

  CHECK_IF_PRESENT(admin, DOF_UCHAR_VEC, dof_uchar_vec, vec);

  if (vec->size < admin->size) {
    vec->vec = MEM_REALLOC(vec->vec, vec->size, admin->size, U_CHAR);
    vec->size = admin->size;
  }
  
  vec->next = admin->dof_uchar_vec;
  admin->dof_uchar_vec = vec;
  /* vec->admin = admin; */
}

void remove_dof_uchar_vec_from_admin(DOF_UCHAR_VEC *vec)
{
  FUNCNAME("remove_dof_uchar_vec_from_admin");
  DOF_ADMIN      *admin;
  DOF_UCHAR_VEC  *vec_last;

  if (vec->fe_space && (admin = (DOF_ADMIN *)vec->fe_space->admin))
  {
    if (admin->dof_uchar_vec == vec)
      admin->dof_uchar_vec = vec->next;
    else
    {
      vec_last = admin->dof_uchar_vec;
      while (vec_last &&  vec_last->next != vec)
	vec_last = vec_last->next;
      
      if (!vec_last)
      {
	ERROR("dof_uchar_vec %s not in list of dof admin %s found\n",
	      NAME(vec), NAME(admin));
      }
      else
	vec_last->next = vec->next;
    }
  }
  return;
}

void add_dof_schar_vec_to_admin(DOF_SCHAR_VEC *vec, DOF_ADMIN *admin)
{
  FUNCNAME("add_dof_schar_vec_to_admin");
  
  if (!vec) {
    MSG("no vec\n");
    return;
  }

  CHECK_IF_PRESENT(admin, DOF_SCHAR_VEC, dof_schar_vec, vec);

  if (vec->size < admin->size) {
    vec->vec = MEM_REALLOC(vec->vec, vec->size, admin->size, S_CHAR);
    vec->size = admin->size;
  }
  
  vec->next = admin->dof_schar_vec;
  admin->dof_schar_vec = vec;
  /* vec->admin = admin; */
}


void remove_dof_schar_vec_from_admin(DOF_SCHAR_VEC *vec)
{
  FUNCNAME("remove_dof_schar_vec_from_admin");
  DOF_ADMIN      *admin;
  DOF_SCHAR_VEC  *vec_last;

  if (vec->fe_space && (admin = (DOF_ADMIN *)vec->fe_space->admin))
  {
    if (admin->dof_schar_vec == vec)
      admin->dof_schar_vec = vec->next;
    else
    {
      vec_last = admin->dof_schar_vec;
      while (vec_last &&  vec_last->next != vec)
	vec_last = vec_last->next;
      
      if (!vec_last)
      {
	ERROR("dof_schar_vec %s not in list of dof admin %s found\n",
	      NAME(vec), NAME(admin));
      }
      else
	vec_last->next = vec->next;
    }
  }
  return;
}

void add_dof_real_vec_to_admin(DOF_REAL_VEC *vec, DOF_ADMIN *admin)
{
  FUNCNAME("add_dof_real_vec_to_admin");
  
  if (!vec) {
    MSG("no vec\n");
    return;
  }

  CHECK_IF_PRESENT(admin, DOF_REAL_VEC, dof_real_vec, vec);

  if (vec->size < admin->size) {
    vec->vec = MEM_REALLOC(vec->vec, vec->size, admin->size, REAL);
    vec->size = admin->size;
  }
  
  vec->next = admin->dof_real_vec;
  admin->dof_real_vec = vec;
  /* vec->admin = admin; */
}

void remove_dof_real_vec_from_admin(DOF_REAL_VEC *vec)
{
  FUNCNAME("remove_dof_real_vec_from_admin");
  DOF_ADMIN      *admin;
  DOF_REAL_VEC   *vec_last;

  if (vec->fe_space && (admin = (DOF_ADMIN *)vec->fe_space->admin))
  {
    if (admin->dof_real_vec == vec)
      admin->dof_real_vec = vec->next;
    else
    {
      vec_last = admin->dof_real_vec;
      while (vec_last &&  vec_last->next != vec)
	vec_last = vec_last->next;
      
      if (!vec_last)
      {
	ERROR("dof_real_vec %s not in list of dof admin %s found\n",
	      NAME(vec), NAME(admin));
      }
      else
	vec_last->next = vec->next;
    }
  }
  return;
}

void add_dof_real_d_vec_to_admin(DOF_REAL_D_VEC *vec, DOF_ADMIN *admin)
{
  FUNCNAME("add_dof_real_d_vec_to_admin");
  
  if (!vec) {
    MSG("no vec\n");
    return;
  }

  CHECK_IF_PRESENT(admin, DOF_REAL_D_VEC, dof_real_d_vec, vec);

  if (vec->size < admin->size) {
    vec->vec = MEM_REALLOC(vec->vec, vec->size, admin->size, REAL_D);
    vec->size = admin->size;
  }

  vec->next = admin->dof_real_d_vec;
  admin->dof_real_d_vec = vec;
  /* vec->admin = admin; */
}

void remove_dof_real_d_vec_from_admin(DOF_REAL_D_VEC *vec)
{
  FUNCNAME("remove_dof_real_d_vec_from_admin");
  DOF_ADMIN       *admin;
  DOF_REAL_D_VEC  *vec_last;

  if (vec->fe_space && (admin = (DOF_ADMIN *)vec->fe_space->admin))
  {
    if (admin->dof_real_d_vec == vec)
      admin->dof_real_d_vec = vec->next;
    else
    {
      vec_last = admin->dof_real_d_vec;
      while (vec_last &&  vec_last->next != vec)
	vec_last = vec_last->next;
      
      if (!vec_last)
      {
	ERROR("dof_real_d_vec %s not in list of dof admin %s found\n",
	      NAME(vec), NAME(admin));
      }
      else
	vec_last->next = vec->next;
    }
  }
  return;
}

void add_dof_matrix_to_admin(DOF_MATRIX *matrix, DOF_ADMIN *admin)
{
  FUNCNAME("add_dof_matrix_to_admin");
  int      i;

  if (!matrix) {
    MSG("no matrix\n");
    return;
  }

  CHECK_IF_PRESENT(admin, DOF_MATRIX, dof_matrix, matrix);

  if (matrix->size < admin->size)
  {
    matrix->matrix_row = MEM_REALLOC(matrix->matrix_row, matrix->size,
				     admin->size, MATRIX_ROW *);
    for (i = matrix->size; i < admin->size; i++)
      matrix->matrix_row[i] = nil;

    matrix->size = admin->size;
  }

  matrix->next = admin->dof_matrix;
  admin->dof_matrix = matrix;
  /* matrix->admin = admin; */
}

void remove_dof_matrix_from_admin(DOF_MATRIX *mat)
{
  FUNCNAME("remove_dof_real_vec_from_admin");
  DOF_ADMIN      *admin;
  DOF_MATRIX     *mat_last;

  if (mat->fe_space && (admin = (DOF_ADMIN *)mat->fe_space->admin))
  {
    if (admin->dof_matrix == mat)
      admin->dof_matrix = mat->next;
    else
    {
      mat_last = admin->dof_matrix;
      while (mat_last &&  mat_last->next != mat)
	mat_last = mat_last->next;
      
      if (!mat_last)
      {
	ERROR("dof_matrix %s not in list of dof admin %s found\n",
	      NAME(mat), NAME(admin));
      }
      else
	mat_last->next = mat->next;
    }
  }
  return;
}

#undef CHECK_IF_PRESENT

/*--------------------------------------------------------------------------*/

void update_dof_matrix(DOF_MATRIX *matrix, REAL sign, int dim, 
		       const REAL **el_mat, const DOF  *el_dof, 
		       const S_CHAR *el_bound)
{
  FUNCNAME("update_dof_matrix");
  DOF        i, j, k, irow, jcol;
  int        free_col=0;
  MATRIX_ROW *row, *free_row;

  TEST_EXIT(matrix)("no matrix\n");

  for (i = 0; i < dim; i++) 
  {
    irow = el_dof[i];

    if (matrix->matrix_row[irow] == nil)
    {
      row = matrix->matrix_row[irow] = get_matrix_row(matrix->fe_space);
      row->col[0] = irow;           /* first entry is diagonal element */

      if (el_bound && el_bound[i] >= DIRICHLET)
      {
	row->entry[0] = 1.0;
	continue;
      }
      else
	row->entry[0] = 0.0;
    }
    else if (el_bound && el_bound[i] >= DIRICHLET)
      continue;

    for (j = 0; j < dim; j++)
    {
      jcol = el_dof[j];
      row = matrix->matrix_row[irow];
      free_row = nil;
      do 
      {
	for (k=0; k<ROW_LENGTH; k++)
	{
	  if (row->col[k] == jcol)
	  {
	    row->entry[k] += sign*el_mat[i][j];
	    break;
	  }
	  if (ENTRY_NOT_USED(row->col[k])) 
	  {
	    free_col = k;
	    free_row = row;
	    if (row->col[k] == NO_MORE_ENTRIES) 
	    {
	      k = ROW_LENGTH;
	      break;
	    }
	  }
	}
	if (k < ROW_LENGTH) break;               /* done? */
	if (row->next || free_row)
	{
	  row = row->next;
	}
	else 
	{
	  free_row = row->next = get_matrix_row(matrix->fe_space);
	  free_col = 0;
	  row = nil;
	}
      } while (row);
      
      if (k >= ROW_LENGTH)                       /* not done? */
      {                     
	TEST_EXIT(free_row)("no free_row\n");
	TEST_EXIT(sign > 0.0)
	  ("new entry (%d,%d) in matrix %s with sign=%lf\n",
	   irow, jcol, matrix->name, sign);
	free_row->col[free_col]   = jcol;
	free_row->entry[free_col] = sign * el_mat[i][j];
      }
    }
  }
  return;
}

void update_dof_real_vec(DOF_REAL_VEC *drv, REAL sign, int el_dim, 
			 const REAL *el_vec, const DOF  *el_dof, 
			 const S_CHAR *el_bound)
{
  FUNCNAME("update_dof_real_vec");
  REAL   *vec;
  int    i;

  TEST_EXIT(drv)("no dof_real_vec\n");
  vec = drv->vec;

  if (el_bound)
  {
    for (i = 0; i < el_dim; i++)
      if (el_bound[i] < DIRICHLET)
	vec[el_dof[i]] += sign*el_vec[i];
  }
  else
  {
    for (i = 0; i < el_dim; i++)
      vec[el_dof[i]] += sign*el_vec[i];

  }
  return;
}

void update_dof_real_d_vec(DOF_REAL_D_VEC *drdv, REAL sign, int el_dim, 
			   const REAL_D *el_vec, const DOF  *el_dof, 
			   const S_CHAR *el_bound)
{
  FUNCNAME("update_dof_real_d_vec");
  REAL_D *vec;
  int    i, n;

  TEST_EXIT(drdv)("no dof_real_vec\n");
  vec = drdv->vec;

  if (el_bound)
  {
    for (i = 0; i < el_dim; i++)
      if (el_bound[i] < DIRICHLET)
	for (n = 0; n < DIM_OF_WORLD; n++)
	  vec[el_dof[i]][n] += sign*el_vec[i][n];
  }
  else
  {
    for (i = 0; i < el_dim; i++)
      for (n = 0; n < DIM_OF_WORLD; n++)
	vec[el_dof[i]][n] += sign*el_vec[i][n];
  }
  return;
}



/*--------------------------------------------------------------------------*/
/* some BLAS level 1 routines:                                              */
/* ---------------------------                                              */
/*   dof_asum:  asum = ||X||_L1                                             */
/*   dof_nrm2:  nrm2 = ||X||_L2                                             */
/*   dof_set:   X = alpha                                                   */
/*   dof_scal:  X = alpha * X                                               */
/*   dof_copy:  Y = X                                                       */
/*   dof_dot:   dot = X . Y                                                 */
/*   dof_axpy:  Y = Y + alpha * X                                           */
/*                                                                          */
/* some BLAS level 2 routines:                                              */
/* ---------------------------                                              */
/*   dof_gemv:  y = alpha*A*x + beta*y   or   y = alpha*A'*x + beta*y       */
/*                                                                          */
/* some non-BLAS level 2 routines:                                          */
/* ---------------------------                                              */
/*   dof_mv:    y = A*x   or   y = A'*x                                     */
/*   dof_min:   min of x                                                    */
/*   dof_max:   max of x                                                    */
/*--------------------------------------------------------------------------*/

REAL dof_nrm2(const DOF_REAL_VEC *x)
{
  FUNCNAME("dof_nrm2");
  REAL nrm;
  const DOF_ADMIN *admin = nil;

  TEST_EXIT(x && x->fe_space && (admin = x->fe_space->admin))
    ("pointer is nil: %8X, %8X\n", x, admin);
  TEST_EXIT(x->size >= admin->size_used)
    ("x->size = %d too small: admin->size_used = %d\n", x->size,
     admin->size_used);

  nrm = 0.0;
  FOR_ALL_DOFS(admin, nrm += x->vec[dof] * x->vec[dof] );

  return(sqrt(nrm));
}


REAL dof_asum(const DOF_REAL_VEC *x)
{
  FUNCNAME("dof_asum");
  REAL    nrm;
  const DOF_ADMIN *admin = nil;

  TEST_EXIT(x && x->fe_space && (admin = x->fe_space->admin))
    ("pointer is nil: %8X, %8X\n", x, admin);
  TEST_EXIT(x->size >= admin->size_used)
    ("x->size = %d too small: admin->size_used = %d\n", x->size, 
     admin->size_used);

  nrm = 0.0;
  FOR_ALL_DOFS(admin, nrm += ABS(x->vec[dof]) );

  return(nrm);
}


void dof_set(REAL alpha, DOF_REAL_VEC *x)
{
  FUNCNAME("dof_set");
  const DOF_ADMIN *admin = nil;

  TEST_EXIT(x && x->fe_space && (admin = x->fe_space->admin))
    ("pointer is nil: %8X, %8X\n", x, admin);
  TEST_EXIT(x->size >= admin->size_used)
    ("x->size = %d too small: admin->size_used = %d\n", x->size,
     admin->size_used);

  FOR_ALL_DOFS(admin, x->vec[dof] = alpha );
}


void dof_scal(REAL alpha, DOF_REAL_VEC *x)
{
  FUNCNAME("dof_scal");
  const DOF_ADMIN *admin = nil;

  TEST_EXIT(x && x->fe_space && (admin = x->fe_space->admin))
    ("pointer is nil: %8X, %8X\n", x, admin);
  TEST_EXIT(x->size >= admin->size_used)
    ("x->size = %d too small: admin->size_used = %d\n", x->size, 
     admin->size_used);

  FOR_ALL_DOFS(admin, x->vec[dof] *= alpha );
}


REAL dof_dot(const DOF_REAL_VEC *x, const DOF_REAL_VEC *y)
{
  FUNCNAME("dof_dot");
  REAL      dot;
  const DOF_ADMIN *admin = nil;

  TEST_EXIT(x && y)("pointer is nil: %8X, %8X\n", x,y);
  TEST_EXIT(x->fe_space && y->fe_space)
    ("fe_space is nil: %8X, %8X\n", x->fe_space,y->fe_space);
  TEST_EXIT((admin = x->fe_space->admin) && (admin == y->fe_space->admin))
    ("no admin or different admins: %8X, %8X\n",
     x->fe_space->admin, y->fe_space->admin);
  TEST_EXIT(x->size >= admin->size_used)
    ("x->size = %d too small: admin->size_used = %d\n", x->size, 
     admin->size_used);
  TEST_EXIT(y->size >= admin->size_used)
    ("y->size = %d too small: admin->size_used = %d\n", y->size, 
     admin->size_used);

  dot = 0.0;
  FOR_ALL_DOFS(admin, dot += x->vec[dof] * y->vec[dof] );

  return(dot);
}


void dof_copy(const DOF_REAL_VEC *x, DOF_REAL_VEC *y)
{
  FUNCNAME("dof_copy");
  REAL      *xvec, *yvec;
  const DOF_ADMIN *admin = nil;

  TEST_EXIT(x && y)("pointer is nil: %8X, %8X\n", x,y);
  TEST_EXIT(x->fe_space && y->fe_space)
    ("fe_space is nil: %8X, %8X\n", x->fe_space,y->fe_space);
  TEST_EXIT((admin = x->fe_space->admin) && (admin == y->fe_space->admin))
    ("no admin or different admins: %8X, %8X\n",
     x->fe_space->admin, y->fe_space->admin);
  TEST_EXIT(x->size >= admin->size_used)
    ("x->size = %d too small: admin->size_used = %d\n", x->size, 
     admin->size_used);
  TEST_EXIT(y->size >= admin->size_used)
    ("y->size = %d too small: admin->size_used = %d\n", y->size, 
     admin->size_used);
  xvec = x->vec;
  yvec = y->vec;

  FOR_ALL_DOFS(admin, yvec[dof] = xvec[dof] );
}


void dof_axpy(REAL alpha, const DOF_REAL_VEC *x, DOF_REAL_VEC *y)
{
  FUNCNAME("dof_axpy");
  REAL      *xvec, *yvec;
  const DOF_ADMIN *admin;

  TEST_EXIT(x && y)("pointer is nil: %8X, %8X\n", x,y);
  TEST_EXIT(x->fe_space && y->fe_space)
    ("fe_space is nil: %8X, %8X\n", x->fe_space,y->fe_space);
  TEST_EXIT((admin = x->fe_space->admin) && (admin == y->fe_space->admin))
    ("no admin or different admins: %8X, %8X\n",
     x->fe_space->admin, y->fe_space->admin);
  TEST_EXIT(x->size >= admin->size_used)
    ("x->size = %d too small: admin->size = %d\n", x->size, admin->size_used);
  TEST_EXIT(y->size >= admin->size_used)
    ("y->size = %d too small: admin->size = %d\n", y->size, admin->size_used);

  xvec = x->vec;
  yvec = y->vec;

  FOR_ALL_DOFS(admin, yvec[dof] += alpha * xvec[dof] );
}

void dof_xpay(REAL alpha, const DOF_REAL_VEC *x, DOF_REAL_VEC *y)
{
  FUNCNAME("dof_axpy");
  REAL      *xvec, *yvec;
  const DOF_ADMIN *admin;

  TEST_EXIT(x && y)("pointer is nil: %8X, %8X\n", x,y);
  TEST_EXIT(x->fe_space && y->fe_space)
    ("fe_space is nil: %8X, %8X\n", x->fe_space,y->fe_space);
  TEST_EXIT((admin = x->fe_space->admin) && (admin == y->fe_space->admin))
    ("no admin or different admins: %8X, %8X\n",
     x->fe_space->admin, y->fe_space->admin);
  TEST_EXIT(x->size >= admin->size_used)
    ("x->size = %d too small: admin->size_used = %d\n",
     x->size, admin->size_used);
  TEST_EXIT(y->size >= admin->size_used)
    ("y->size = %d too small: admin->size_used = %d\n", 
     y->size, admin->size_used);

  xvec = x->vec;
  yvec = y->vec;

  FOR_ALL_DOFS(admin, yvec[dof] = xvec[dof] + alpha*yvec[dof]);
}


REAL dof_min(const DOF_REAL_VEC *x)
{
  FUNCNAME("dof_min");
  REAL m;
  const DOF_ADMIN *admin = nil;

  TEST_EXIT(x && x->fe_space && (admin = x->fe_space->admin))
    ("pointer is nil: %8X, %8X\n", x, admin);
  TEST_EXIT(x->size >= admin->size_used)
    ("x->size = %d too small: admin->size_used = %d\n", x->size, 
     admin->size_used);

  m = 1.0E30;
  FOR_ALL_DOFS(admin, m = MIN(m, x->vec[dof]) );

  return(m);
}

REAL dof_max(const DOF_REAL_VEC *x)
{
  FUNCNAME("dof_max");
  REAL m;
  const DOF_ADMIN *admin = nil;

  TEST_EXIT(x && x->fe_space && (admin = x->fe_space->admin))
    ("pointer is nil: %8X, %8X\n", x, admin);
  TEST_EXIT(x->size >= admin->size_used)
    ("x->size = %d too small: admin->size_used = %d\n", x->size, 
     admin->size_used);

  m = -1.0E30;
  FOR_ALL_DOFS(admin, m = MAX(m, x->vec[dof]) );

  return(m);
}

/*--------------------------------------------------------------------------*/
/*   now the same routines for REAL_D vectors                               */
/*--------------------------------------------------------------------------*/

void dof_set_d(REAL alpha, DOF_REAL_D_VEC *x)
{
  FUNCNAME("dof_set_d");
  const DOF_ADMIN *admin = nil;
  int             n;

  TEST_EXIT(x && x->fe_space  &&  (admin = x->fe_space->admin))
    ("pointer is nil: x: %8X, x->fe_space: %8X, x->fe_space->admin :%8X\n",
     x, admin);
  TEST_EXIT(x->size >= admin->size_used)
    ("x->size = %d too small: admin->size_used = %d\n", 
     x->size, admin->size_used);

  FOR_ALL_DOFS(admin, 
	       for (n = 0; n < DIM_OF_WORLD; n++)
	          x->vec[dof][n] = alpha);
}

void dof_scal_d(REAL alpha, DOF_REAL_D_VEC *x)
{
  FUNCNAME("dof_scal_d");
  const DOF_ADMIN *admin = nil;
  int              n;

  TEST_EXIT(x && x->fe_space  &&  (admin = x->fe_space->admin))
    ("pointer is nil: x: %8X, x->fe_space: %8X, x->fe_space->admin :%8X\n",
     x, admin);
  TEST_EXIT(x->size >= admin->size_used)
    ("x->size = %d too small: admin->size_used = %d\n", 
     x->size, admin->size_used);

  FOR_ALL_DOFS(admin,
	       for (n = 0; n < DIM_OF_WORLD; n++)
	          x->vec[dof][n] *= alpha);
}

REAL dof_dot_d(const DOF_REAL_D_VEC *x, const DOF_REAL_D_VEC *y)
{
  FUNCNAME("dof_dot_d");
  REAL            dot;
  const DOF_ADMIN *admin = nil;
  int             n;

  TEST_EXIT(x && y)("pointer to DOF_REAL_D_VEC is nil: x: %8X, y: %8X\n",
		    x, y);
  TEST_EXIT(x->fe_space && y->fe_space)
    ("pointer to FE_SPACE is nil: x->fe_space: %8X, y->fe_space: %8X\n",
     x->fe_space, y->fe_space);
  
  TEST_EXIT((admin = x->fe_space->admin) && (admin == y->fe_space->admin))
    ("no admin or admins: x->fe_space->admin: %8X, y->fe_space->admin: %8X\n",
     x->fe_space->admin, y->fe_space->admin);

  TEST_EXIT(x->size >= admin->size_used)
    ("x->size = %d too small: admin->size_used = %d\n",
     x->size, admin->size_used);
  TEST_EXIT(y->size >= admin->size_used)
    ("y->size = %d too small: admin->size_used = %d\n",
     y->size, admin->size_used);

  dot = 0.0;
  FOR_ALL_DOFS(admin, for (n = 0; n < DIM_OF_WORLD; n++)
	                 dot += x->vec[dof][n] * y->vec[dof][n] );

  return(dot);
}

void dof_copy_d(const DOF_REAL_D_VEC *x, DOF_REAL_D_VEC *y)
{
  FUNCNAME("dof_copy_d");
  const DOF_ADMIN *admin = nil;
  int             n;

  TEST_EXIT(x && y)("pointer to DOF_REAL_D_VEC is nil: x: %8X, y: %8X\n",
		    x, y);
  TEST_EXIT(x->fe_space && y->fe_space)
    ("pointer to FE_SPACE is nil: x->fe_space: %8X, y->fe_space: %8X\n",
     x->fe_space, y->fe_space);
  
  TEST_EXIT((admin = x->fe_space->admin) && (admin == y->fe_space->admin))
    ("no admin or admins: x->fe_space->admin: %8X, y->fe_space->admin: %8X\n",
     x->fe_space->admin, y->fe_space->admin);
  TEST_EXIT(x->size >= admin->size_used)
    ("x->size = %d too small: admin->size_used = %d\n",
     x->size, admin->size_used);
  TEST_EXIT(y->size >= admin->size_used)
    ("y->size = %d too small: admin->size_used = %d\n",
     y->size, admin->size_used);

  FOR_ALL_DOFS(admin, for (n = 0; n < DIM_OF_WORLD; n++)
	                y->vec[dof][n] = x->vec[dof][n]);

}

void dof_axpy_d(REAL alpha, const DOF_REAL_D_VEC *x, DOF_REAL_D_VEC *y)
{
  FUNCNAME("dof_axpy_d");
  const DOF_ADMIN *admin;
  int             n;

  TEST_EXIT(x && y)("pointer to DOF_REAL_D_VEC is nil: x: %8X, y: %8X\n",
		    x, y);
  TEST_EXIT(x->fe_space && y->fe_space)
    ("pointer to FE_SPACE is nil: x->fe_space: %8X, y->fe_space: %8X\n",
     x->fe_space, y->fe_space);
  
  TEST_EXIT((admin = x->fe_space->admin) && (admin == y->fe_space->admin))
    ("no admin or admins: x->fe_space->admin: %8X, y->fe_space->admin: %8X\n",
     x->fe_space->admin, y->fe_space->admin);
  TEST_EXIT(x->size >= admin->size_used)
    ("x->size = %d too small: admin->size_used = %d\n",
     x->size, admin->size_used);
  TEST_EXIT(y->size >= admin->size_used)
    ("y->size = %d too small: admin->size_used = %d\n",
     y->size, admin->size_used);

  FOR_ALL_DOFS(admin, for (n = 0; n < DIM_OF_WORLD; n++)
                         y->vec[dof][n] += alpha * x->vec[dof][n]);
}

REAL dof_nrm2_d(const DOF_REAL_D_VEC *x)
{
  FUNCNAME("dof_nrm2_d");
  REAL            nrm;
  int             n;
  const DOF_ADMIN *admin = nil;

  TEST_EXIT(x && x->fe_space && (admin = x->fe_space->admin))
    ("pointer is nil: %8X, %8X\n", x, admin);
  TEST_EXIT(x->size >= admin->size_used)
    ("x->size = %d too small: admin->size_used = %d\n", x->size, 
     admin->size_used);

  nrm = 0.0;
  FOR_ALL_DOFS(admin, for (n = 0; n < DIM_OF_WORLD; n++)
	                nrm += x->vec[dof][n] * x->vec[dof][n]);

  return(sqrt(nrm));
}

void dof_xpay_d(REAL alpha, const DOF_REAL_D_VEC *x, DOF_REAL_D_VEC *y)
{
  FUNCNAME("dof_xpay_d");
  const DOF_ADMIN *admin;
  int             n;

  TEST_EXIT(x && y)("pointer to DOF_REAL_D_VEC is nil: x: %8X, y: %8X\n",
		    x, y);
  TEST_EXIT(x->fe_space && y->fe_space)
    ("pointer to FE_SPACE is nil: x->fe_space: %8X, y->fe_space: %8X\n",
     x->fe_space, y->fe_space);
  
  TEST_EXIT((admin = x->fe_space->admin) && (admin == y->fe_space->admin))
    ("no admin or admins: x->fe_space->admin: %8X, y->fe_space->admin: %8X\n",
     x->fe_space->admin, y->fe_space->admin);
  TEST_EXIT(x->size >= admin->size_used)
    ("x->size = %d too small: admin->size_used = %d\n",
     x->size, admin->size_used);
  TEST_EXIT(y->size >= admin->size_used)
    ("y->size = %d too small: admin->size_used = %d\n",
     y->size, admin->size_used);

  FOR_ALL_DOFS(admin, 
	       for (n = 0; n < DIM_OF_WORLD; n++)
	          y->vec[dof][n] = x->vec[dof][n] + alpha*y->vec[dof][n]);
}

REAL dof_min_d(const DOF_REAL_D_VEC *x)
{
  FUNCNAME("dof_min_d");
  REAL m, norm;
  const DOF_ADMIN *admin = nil;

  TEST_EXIT(x && x->fe_space && (admin = x->fe_space->admin))
    ("pointer is nil: %8X, %8X\n", x, admin);
  TEST_EXIT(x->size >= admin->size_used)
    ("x->size = %d too small: admin->size_used = %d\n", x->size, 
     admin->size_used);

  m = 1.0E30;
  FOR_ALL_DOFS(admin, 
	       norm = NORM_DOW(x->vec[dof]);
	       m = MIN(m, norm) );

  return(m);
}

REAL dof_max_d(const DOF_REAL_D_VEC *x)
{
  FUNCNAME("dof_max_d");
  REAL m, norm;
  const DOF_ADMIN *admin = nil;

  TEST_EXIT(x && x->fe_space && (admin = x->fe_space->admin))
    ("pointer is nil: %8X, %8X\n", x, admin);
  TEST_EXIT(x->size >= admin->size_used)
    ("x->size = %d too small: admin->size_used = %d\n", x->size, 
     admin->size_used);

  m = 0.0;
  FOR_ALL_DOFS(admin,  
	       norm = NORM_DOW(x->vec[dof]);
	       m = MAX(m, norm) );

  return(m);
}



void dof_gemv(MatrixTranspose transpose, REAL alpha,
	      const DOF_MATRIX *a, const DOF_REAL_VEC *x,
	      REAL beta, DOF_REAL_VEC *y)
{
  FUNCNAME("dof_gemv");
  int        j, jcol, ysize;
  REAL       sum, ax;
  REAL       *xvec, *yvec;
  MATRIX_ROW *row;
  const DOF_ADMIN *admin;

  TEST_EXIT(a && x && y)("pointer is nil: %8X, %8X, %8X\n", a,x,y);
  TEST_EXIT(a->fe_space && x->fe_space && y->fe_space)
    ("fe_space is nil: %8X, %8X, %8X\n", a->fe_space,x->fe_space,y->fe_space);
  TEST_EXIT((admin = a->fe_space->admin)
	    && (admin == x->fe_space->admin) && (admin == y->fe_space->admin))
    ("no admin or different admins: %8X, %8X, %8X\n",
     a->fe_space->admin, x->fe_space->admin, y->fe_space->admin);

  TEST_EXIT(x->size >= admin->size_used)
    ("x->size = %d too small: admin->size_used = %d\n",
     x->size, admin->size_used);
  TEST_EXIT(y->size >= admin->size_used)
    ("y->size = %d too small: admin->size_used = %d\n",
     y->size, admin->size_used);
  TEST_EXIT(a->size >= admin->size_used)
    ("a->size = %d too small: admin->size_used = %d\n",
     a->size, admin->size_used);

  xvec = x->vec;
  yvec = y->vec;

  ysize = y->size;

  FOR_ALL_FREE_DOFS(admin, if (dof < ysize) yvec[dof] = 0.0);

  if (transpose == NoTranspose) {
    FOR_ALL_DOFS(admin, 
	      sum = 0.0;
	      for (row = a->matrix_row[dof]; row; row = row->next) {
		for (j=0; j<ROW_LENGTH; j++) {
		  jcol = row->col[j];
		  if (ENTRY_USED(jcol)) {
		    sum += row->entry[j] * xvec[jcol];
		  }
		  else {
		    if (jcol == NO_MORE_ENTRIES)
		      break;
		  }
		}
	      }
	      yvec[dof] = beta * yvec[dof] + alpha * sum;
      );
  }
  else if (transpose == Transpose) {

    FOR_ALL_DOFS(admin, yvec[dof]  *= beta );
    FOR_ALL_DOFS(admin, 
		 ax = alpha * xvec[dof];
		 for (row = a->matrix_row[dof]; row; row = row->next) {
		   for (j=0; j<ROW_LENGTH; j++) {
		     jcol = row->col[j];
		     if (ENTRY_USED(jcol))
		       yvec[jcol] += ax * row->entry[j];
		     else 
		       if (jcol == NO_MORE_ENTRIES) break;
		   }
		 }
      );
  }
  else {
    ERROR_EXIT("transpose=%d\n", transpose);
  }
}


/*--------------------------------------------------------------------------*/

void dof_mv(MatrixTranspose transpose,
	    const DOF_MATRIX *a, const DOF_REAL_VEC *x, DOF_REAL_VEC *y)
{
  FUNCNAME("dof_mv");
  int           j, jcol;
  REAL          sum, ax, *xvec, *yvec;
  MATRIX_ROW    *row;
  const DOF_ADMIN     *admin;

  TEST_EXIT(a && x && y)("pointer is nil: %8X, %8X, %8X\n", a,x,y);
  TEST_EXIT(a->fe_space && x->fe_space && y->fe_space)
    ("fe_space is nil: %8X, %8X, %8X\n", a->fe_space,x->fe_space,y->fe_space);
  TEST_EXIT((admin = a->fe_space->admin)
	    && (admin == x->fe_space->admin) && (admin == y->fe_space->admin))
    ("no admin or different admins: %8X, %8X, %8X\n",
     a->fe_space->admin, x->fe_space->admin, y->fe_space->admin);

  TEST_EXIT(x->size >= admin->size_used)
    ("x->size = %d too small: admin->size_used = %d\n",
     x->size, admin->size_used);
  TEST_EXIT(y->size >= admin->size_used)
    ("y->size = %d too small: admin->size_used = %d\n",
     y->size, admin->size_used);
  TEST_EXIT(a->size >= admin->size_used)
    ("a->size = %d too small: admin->size_used = %d\n",
     a->size, admin->size_used);

  xvec = x->vec;
  yvec = y->vec;

  if (transpose == NoTranspose) {
    int ysize = y->size;
    FOR_ALL_FREE_DOFS(admin, if (dof < ysize) yvec[dof] = 0.0);

    FOR_ALL_DOFS(admin,
		 sum = 0.0;
		 for (row = a->matrix_row[dof]; row; row = row->next) {
		   for (j=0; j<ROW_LENGTH; j++) {
		     jcol = row->col[j];
		     if (ENTRY_USED(jcol)) {
		       sum += row->entry[j] * xvec[jcol];
		     }
		     else {
		       if (jcol == NO_MORE_ENTRIES)
			 break;
		     }
		   }
		 }
		 yvec[dof] = sum;
      );
  }
  else if (transpose == Transpose) {
    
    for (j = 0; j < admin->size_used; j++)
      yvec[j] = 0.0;

    FOR_ALL_DOFS(admin,
		 ax = xvec[dof];
		 for (row = a->matrix_row[dof]; row; row = row->next) {
		   for (j=0; j<ROW_LENGTH; j++) {
		     jcol = row->col[j];
		     if (ENTRY_USED(jcol)) {
		       yvec[jcol] += ax * row->entry[j];
		     }
		     else {
		       if (jcol == NO_MORE_ENTRIES)
			 break;
		     }
		   }
		 }
      );
  }
  else {
    ERROR_EXIT("transpose=%d\n", transpose);
  }
}


void dof_gemv_d(MatrixTranspose transpose, REAL alpha,
		const DOF_MATRIX *a, const DOF_REAL_D_VEC *x,
		REAL beta, DOF_REAL_D_VEC *y)
{
  FUNCNAME("dof_gemv_d");
  int               n, j, jcol;
  REAL_D            sum, ax;
  REAL_D            *xvec, *yvec;
  MATRIX_ROW        *row;
  const DOF_ADMIN   *admin;

  TEST_EXIT(a && x && y)("pointer is nil: %8X, %8X, %8X\n", a,x,y);
  TEST_EXIT(a->fe_space  &&  x->fe_space  &&  y->fe_space)
    ("fe_space pointer is nil: %8X, %8X, %8X\n", a->fe_space,
     x->fe_space, y->fe_space);
  TEST_EXIT((admin = a->fe_space->admin) && (admin == x->fe_space->admin) 
	     && (admin == y->fe_space->admin))
    ("no admin or different admins: %8X, %8X, %8X\n",
     a->fe_space->admin, x->fe_space->admin, y->fe_space->admin);
  
  TEST_EXIT(x->size >= admin->size_used)
    ("x->size = %d too small: admin->size_used = %d\n",
     x->size, admin->size_used);
  TEST_EXIT(y->size >= admin->size_used)
    ("y->size = %d too small: admin->size_used = %d\n",
     y->size, admin->size_used);
  TEST_EXIT(a->size >= admin->size_used)
    ("a->size = %d too small: admin->size_used = %d\n",
     a->size, admin->size_used);

  xvec = x->vec;
  yvec = y->vec;

  if (transpose == NoTranspose) {
    int ysize = y->size;
    FOR_ALL_FREE_DOFS(admin, if (dof < ysize) 
                               for (n = 0; n < DIM_OF_WORLD; n++)
		                  yvec[dof][n] = 0.0);

    FOR_ALL_DOFS(admin,	
		 for (n = 0; n < DIM_OF_WORLD; n++)
		    sum[n] = 0.0;
		 for (row = a->matrix_row[dof]; row; row = row->next) {
		   for (j=0; j<ROW_LENGTH; j++) {
		     jcol = row->col[j];
		     if (ENTRY_USED(jcol)) 
		     {
		       for (n = 0; n < DIM_OF_WORLD; n++)
			 sum[n] += row->entry[j] * xvec[jcol][n];
		     }
		     else 
		     {
		       if (jcol == NO_MORE_ENTRIES)
			 break;
		     }
		   }
		 }
		 for (n = 0; n < DIM_OF_WORLD; n++)
		    yvec[dof][n] = beta * yvec[dof][n] + alpha * sum[n];
      );
  }
  else if (transpose == Transpose) {
    
    FOR_ALL_DOFS(admin,
		 for (n = 0; n < DIM_OF_WORLD; n++) {
		     yvec[dof][n]  *= beta;
		 } );
   FOR_ALL_DOFS(admin,
		 for (n = 0; n < DIM_OF_WORLD; n++)
		    ax[n] = alpha * xvec[dof][n];
		 for (row = a->matrix_row[dof]; row; row = row->next) {
		   for (j=0; j<ROW_LENGTH; j++) {
		     jcol = row->col[j];
		     if (ENTRY_USED(jcol))
		     {
		       for (n = 0; n < DIM_OF_WORLD; n++)
			 yvec[jcol][n] += ax[n] * row->entry[j];
		     }
		     else
		     {
		       if (jcol == NO_MORE_ENTRIES)
			 break;
		     }
		   }
		 }
      );
  }
  else {
    ERROR_EXIT("transpose=%d\n", transpose);
  }
}

void dof_mv_d(MatrixTranspose transpose, const DOF_MATRIX *a, 
	      const DOF_REAL_D_VEC *x, DOF_REAL_D_VEC *y)
{
  FUNCNAME("dof_mv_d");
  int               n, j, jcol;
  REAL_D            sum, ax;
  REAL_D            *xvec, *yvec;
  MATRIX_ROW        *row;
  const DOF_ADMIN   *admin;

  TEST_EXIT(a && x && y)("pointer is nil: %8X, %8X, %8X\n", a,x,y);
  TEST_EXIT(a->fe_space  &&  x->fe_space  &&  y->fe_space)
    ("fe_space pointer is nil: %8X, %8X, %8X\n", a->fe_space,
     x->fe_space, y->fe_space);
  TEST_EXIT((admin = a->fe_space->admin) && (admin == x->fe_space->admin) 
	     && (admin == y->fe_space->admin))
    ("no admin or different admins: %8X, %8X, %8X\n",
     a->fe_space->admin, x->fe_space->admin, y->fe_space->admin);
  
  TEST_EXIT(x->size >= admin->size_used)
    ("x->size = %d too small: admin->size_used = %d\n",
     x->size, admin->size_used);
  TEST_EXIT(y->size >= admin->size_used)
    ("y->size = %d too small: admin->size_used = %d\n",
     y->size, admin->size_used);
  TEST_EXIT(a->size >= admin->size_used)
    ("a->size = %d too small: admin->size_used = %d\n",
     a->size, admin->size_used);

  xvec = x->vec;
  yvec = y->vec;

  if (transpose == NoTranspose) {
    int ysize = y->size;
    FOR_ALL_FREE_DOFS(admin, if (dof < ysize) 
                               for (n = 0; n < DIM_OF_WORLD; n++)
		                  yvec[dof][n] = 0.0);

    FOR_ALL_DOFS(admin,	
		 for (n = 0; n < DIM_OF_WORLD; n++)
		    sum[n] = 0.0;
		 for (row = a->matrix_row[dof]; row; row = row->next) {
		   for (j=0; j<ROW_LENGTH; j++) {
		     jcol = row->col[j];
		     if (ENTRY_USED(jcol)) 
		     {
		       for (n = 0; n < DIM_OF_WORLD; n++)
			 sum[n] += row->entry[j] * xvec[jcol][n];
		     }
		     else 
		     {
		       if (jcol == NO_MORE_ENTRIES)
			 break;
		     }
		   }
		 }
		 for (n = 0; n < DIM_OF_WORLD; n++)
		    yvec[dof][n] = sum[n];
      );
  }
  else if (transpose == Transpose) {
    
    for (j = 0; j < admin->size_used; j++)
      for (n = 0; n < DIM_OF_WORLD; n++)
	yvec[j][n] = 0.0;

    FOR_ALL_DOFS(admin,
		 for (n = 0; n < DIM_OF_WORLD; n++)
		    ax[n] = xvec[dof][n];
		 for (row = a->matrix_row[dof]; row; row = row->next) {
		   for (j=0; j<ROW_LENGTH; j++) {
		     jcol = row->col[j];
		     if (ENTRY_USED(jcol))
		     {
		       for (n = 0; n < DIM_OF_WORLD; n++)
			 yvec[jcol][n] += ax[n] * row->entry[j];
		     }
		     else
		     {
		       if (jcol == NO_MORE_ENTRIES)
			 break;
		     }
		   }
		 }
      );
  }
  else {
    ERROR_EXIT("transpose=%d\n", transpose);
  }
}

/*--------------------------------------------------------------------------*/
/*  print_dof_matrix: remove all entries from dof_matrix                    */
/*--------------------------------------------------------------------------*/

void print_dof_matrix(const DOF_MATRIX *matrix)
{
  FUNCNAME("print_dof_matrix");
  int  i, j, jcol;
  MATRIX_ROW *row;

  for (i=0; i<matrix->size; i++) { 
    for (row = matrix->matrix_row[i]; row; row = row->next) {
      MSG("row %3d:",i);
      for (j=0; j<ROW_LENGTH; j++) {
	jcol = row->col[j];
	if (ENTRY_USED(jcol)) {
	  print_msg(" (%3d,%10.7lf)", jcol, row->entry[j]);
	}
      }
      print_msg("\n");
    }
  }
  return;
}

void print_dof_real_vec(const DOF_REAL_VEC *drv)
{
  FUNCNAME("print_dof_real_vec");
  int  i, j;
  const DOF_ADMIN *admin = nil;
  char      *format;

  if (drv->fe_space) admin = drv->fe_space->admin;

  MSG("Vec `%s':\n", drv->name);
  j = 0;
  if (admin)
  {
    if (admin->size_used > 100)
      format = "%s(%3d,%10.5le)";
    else if (admin->size_used > 10)
      format = "%s(%2d,%10.5le)";
    else
      format = "%s(%1d,%10.5le)";

    FOR_ALL_DOFS(admin,
		 if ((j % 3) == 0) {
                   if (j) print_msg("\n");
		   MSG(format, "", dof, drv->vec[dof]);
                 }
		 else 
		   print_msg(format, " ", dof, drv->vec[dof]);
		 j++;
      );
    print_msg("\n");
  }
  else
  {
    MSG("no DOF_ADMIN, print whole vector.\n");
    
    for (i = 0; i < drv->size; i++) {
      if ((j % 3) == 0)
      {
	if (j) print_msg("\n");
	MSG("(%d,%10.5le)",i,drv->vec[i]);
      }
      else 
	print_msg(" (%d,%10.5le)",i,drv->vec[i]);
      j++;
    }
    print_msg("\n");
  }
  return;
}


void print_dof_real_d_vec(const DOF_REAL_D_VEC *drdv)
{
  FUNCNAME("print_dof_real_d_vec");
  int         i, j, k;
  const DOF_ADMIN   *admin = nil;
#if DIM_OF_WORLD < 2
  static int  per_line = 4;
#else
#if DIM_OF_WORLD < 4
  static int  per_line = 2;
#else
  static int  per_line = 1;
#endif
#endif

  if (drdv->fe_space) admin = drdv->fe_space->admin;

  MSG("Vec `%s':\n", drdv->name);
  j = 0;
  if (admin)
  {
    FOR_ALL_DOFS(admin,
		 if ((j % per_line) == 0) {
                    if (j) print_msg("\n");
		    MSG("(%3d:",dof);
                 }
		 else 
		    print_msg(" (%3d:", dof);
		 for (k=0; k<DIM_OF_WORLD; k++)
		 print_msg("%c%10.5le", (k > 0 ? ',':' '), drdv->vec[dof][k]);
		 print_msg(")");
		 j++;
		 );	 
    print_msg("\n");
  }
  else
  {
    MSG("no DOF_ADMIN, print whole vector.\n");
    
    for (i = 0; i < drdv->size; i++) {
      if ((j % per_line) == 0)
      {
	if (j) print_msg("\n");
	MSG("(%3d:",i);
      }
      else 
	print_msg(" (%3d:", i);
      for (k=0; k<DIM_OF_WORLD; k++)
	print_msg("%c%10.5le", (k > 0 ? ',':' '), drdv->vec[i][k]);
      print_msg(")");
    }
    print_msg("\n");
  }
  return;
}


void print_dof_int_vec(const DOF_INT_VEC *div)
{
  FUNCNAME("print_dof_int_vec");
  int       i, j;
  const DOF_ADMIN *admin = nil;
  char      *format;

  if (div->fe_space) admin = div->fe_space->admin;

  MSG("Vector `%s':\n", div->name);
  j = 0;
  if (admin) {
    if (admin->size_used > 100)
      format = "%s(%3d,%3d)";
    else if (admin->size_used > 10)
      format = "%s(%2d,%3d)";
    else
      format = "%s(%1d,%3d)";

    FOR_ALL_DOFS(admin,
		 if ((j % 5) == 0)
		 {
		   if (j) print_msg("\n");
		   MSG(format, "", dof,div->vec[dof]);
		 }
	         else 
		   print_msg(format, " ", dof, div->vec[dof]);
	         j++;
      );
    print_msg("\n");
  }
  else {
    if (div->size > 100)
      format = "%s(%3d,%3d)";
    else if (div->size > 10)
      format = "%s(%2d,%3d)";
    else
      format = "%s(%1d,%3d)";

    for (i = 0; i < div->size; i++) {
      if ((j % 5) == 0)
	{
	  if (j) print_msg("\n");
	  MSG(format, "", i, div->vec[i]);
	}
      else 
	print_msg(format, " ", i, div->vec[i]);
      j++;
    }
    print_msg("\n");
  }
  return;
}

void print_dof_uchar_vec(const DOF_UCHAR_VEC *duv)
{
  FUNCNAME("print_dof_uchar_vec");
  int  i, j;
  const DOF_ADMIN *admin = nil;
  char      *format;

  if (duv->fe_space) admin = duv->fe_space->admin;

  MSG("Vector `%s':\n", duv->name);
  j = 0;

  if (admin) {
    if (admin->size_used > 100)
      format = "%s(%3d,0x%2X)";
    else if (admin->size_used > 10)
      format = "%s(%2d,0x%2X)";
    else
      format = "%s(%1d,0x%2X)";

    FOR_ALL_DOFS(admin,
	      if ((j % 5) == 0) {
		if (j) print_msg("\n");
		MSG(format, "", dof, duv->vec[dof]);
	      }
	      else 
		print_msg(format, " ", dof, duv->vec[dof]);
	      j++;
      );
    print_msg("\n");
  }
  else {
    if (duv->size > 100)
      format = "%s(%3d,0x%2X)";
    else if (duv->size > 10)
      format = "%s(%2d,0x%2X)";
    else
      format = "%s(%1d,0x%2X)";

    for (i = 0; i < duv->size; i++) {
      if ((j % 5) == 0)
	{
	  if (j) print_msg("\n");
	  MSG(format, "", i, duv->vec[i]);
	}
      else 
	print_msg(format, " ", i, duv->vec[i]);
      j++;
    }
    print_msg("\n");
  }
  return;
}

void print_dof_schar_vec(const DOF_SCHAR_VEC *dsv)
{
  FUNCNAME("print_dof_schar_vec");
  int       i, j;
  const DOF_ADMIN *admin = nil;
  char      *format;

  if (dsv->fe_space) admin = dsv->fe_space->admin;

  MSG("Vector `%s':\n", dsv->name);
  j = 0;


  if (admin) {
    if (admin->size_used > 100)
      format = "%s(%3d,0x%2X)";
    else if (admin->size_used > 10)
      format = "%s(%2d,0x%2X)";
    else
      format = "%s(%1d,0x%2X)";

    FOR_ALL_DOFS(admin,
	      if ((j % 5) == 0)
		{
		  if (j) print_msg("\n");
		  MSG("(%d,0x%2X)",dof,dsv->vec[dof]);
		}
	      else 
		print_msg(" (%d,0x%2X)",dof,dsv->vec[dof]);
	      j++;
		 );
    print_msg("\n");
  }
  else {
    if (dsv->size > 100)
      format = "%s(%3d,0x%2X)";
    else if (dsv->size > 10)
      format = "%s(%2d,0x%2X)";
    else
      format = "%s(%1d,0x%2X)";

    for (i = 0; i < dsv->size; i++) {
      if ((j % 5) == 0)
	{
	  if (j) print_msg("\n");
	  MSG(format, "", i, dsv->vec[i]);
	}
      else 
	print_msg(format, " ", i, dsv->vec[i]);
      j++;
    }
    print_msg("\n");
  }
  return;
}

/*--------------------------------------------------------------------------*/
/*  clear_dof_matrix: remove all entries from dof_matrix                    */
/*--------------------------------------------------------------------------*/

void clear_dof_matrix(DOF_MATRIX *matrix)
{
  int        i;
  MATRIX_ROW *row, *next;

  for (i=0; i<matrix->size; i++) {
    for (row = matrix->matrix_row[i]; row; row = next) {
      next = row->next;
      free_matrix_row(matrix->fe_space, row);
    }
    matrix->matrix_row[i] = nil;
  }
  return;
}

/*--------------------------------------------------------------------------*/
/*  test_dof_matrix                                                         */
/*--------------------------------------------------------------------------*/

void test_dof_matrix(DOF_MATRIX *matrix)
{
  FUNCNAME("test_dof_matrix");
  int  i, j, jcol, k,kcol;
  MATRIX_ROW *row, *row2;
  int        non_symmetric, found;
  double     sum;

/* test symmetry */
  non_symmetric = 0;
  for (i=0; i<matrix->size; i++) { 
    sum = 0.0;
    for (row = matrix->matrix_row[i]; row; row = row->next) {
      for (j=0; j<ROW_LENGTH; j++) {
	jcol = row->col[j];
	if (ENTRY_USED(jcol)) {
	  found = 0;
	  if (row->entry[j] != row->entry[j]) {
	    MSG("mat[%d,%d]=%10.5le ???\n", i,jcol,row->entry[j]);
	    WAIT;
	  }
	  for (row2 = matrix->matrix_row[jcol]; row2;
	       row2 = row2 ? row2->next : nil) {
	    for (k=0; k<ROW_LENGTH; k++) {
	      kcol = row2->col[k];
	      if (ENTRY_USED(kcol)) {
		if (kcol==i) {
		  found = 1;
		  if (ABS(row2->entry[k] - row->entry[j]) > 1.E-5) {
		    non_symmetric = 1;
		    MSG("mat[%d,%d]=%10.5le != mat[%d,%d]=%10.5le\n",
			i,jcol,row->entry[j], jcol,i,row2->entry[k]);
		  }
		  row2 = nil;
		  break;
		}
	      }
	    }
	  }
	  if (!found) {
	    non_symmetric = 1;
	    MSG("mat[%d,%d] not found\n",jcol,i);
	  }
	}
      }
    }
    if (ABS(sum) > 1.E-5) {
      MSG("Zeilensumme[%d] = %10.5le\n", i, sum);
    }
  }
  if (non_symmetric) {
    MSG("matrix `%s' not symmetric.\n",matrix->name);
    WAIT;
  }
  else {
    MSG("matrix `%s' is symmetric.\n",matrix->name);
  }
  return;
}

#if 0
/*--------------------------------------------------------------------------*/
/*  fill_dof_matrix: add all el_stiff matrices from mesh leaf elements      */
/*--------------------------------------------------------------------------*/

static DOF_MATRIX *fill_mat;
static const void *fill_info;
static int        n_bas_fcts;
static const DOF_ADMIN  *fill_admin;
static const DOF    *(*get_dof_indices)(const EL *,const DOF_ADMIN *, DOF *);
static const REAL   **(*el_stiff_fct)(const EL_INFO *, const void *);
static const S_CHAR *(*get_bound)(const EL_INFO *, S_CHAR *);

static void fill_matrix_fct(const EL_INFO *el_info)
{
  const REAL   **el_mat  = (*el_stiff_fct)(el_info, fill_info);
  const DOF    *el_dof   = (*get_dof_indices)(el_info->el, fill_admin, nil);
  const S_CHAR *el_bound = (*get_bound)(el_info, nil);

  update_dof_matrix(fill_mat, 1.0, n_bas_fcts, el_mat, el_dof, el_bound);
  return;
}

void fill_dof_matrix(DOF_MATRIX *mat, const EL_STIFF_INFO *info)
{
  FUNCNAME("fill_dof_matrix");
  FLAGS    fill_flag = CALL_LEAF_EL|FILL_BOUND;

  TEST_EXIT(mat)("matrix pointer to nil\n");
  TEST_EXIT(mat->fe_space  &&  mat->fe_space->bas_fcts)
    ("no `fe_space' or no `fe_space->bas_fcts' in matrix %s\n", NAME(mat));

  TEST_EXIT(info  &&  info->el_stiff_fct)
    ("no `el_stiff_info' or `el_stiff_info->el_stiff_fct'\n");

  fill_mat        = mat;
  fill_admin      = mat->fe_space->admin;
  get_dof_indices = mat->fe_space->bas_fcts->get_dof_indices;
  get_bound       = mat->fe_space->bas_fcts->get_bound;
  n_bas_fcts      = mat->fe_space->bas_fcts->n_bas_fcts;
  fill_flag      |= info->fill_flag;
  el_stiff_fct    = info->el_stiff_fct;
  fill_info       = info->fill_info;

  mesh_traverse(mat->fe_space->mesh, -1, fill_flag, fill_matrix_fct);
  return;
}
#endif

/*--------------------------------------------------------------------------*/
/*   the new update routines:                                               */
/*--------------------------------------------------------------------------*/

void add_element_matrix(DOF_MATRIX *matrix, REAL sign, int n_row, int n_col,
			const DOF *row_dof, const DOF *col_dof, 
			const REAL **el_mat, const S_CHAR *bound)
{
  FUNCNAME("add_element_matrix");
  DOF        i, j, k, irow, jcol;
  int        free_col=0;
  MATRIX_ROW *row, *free_row;

  TEST_EXIT(matrix)("no matrix\n");
  if (!col_dof || n_col <= 0)
  {
    col_dof = row_dof;
    n_col = n_row;
  }
  if (row_dof != col_dof  && bound)
    ERROR_EXIT("no boundary treatment for row_dof != col_dof\n");
    
  for (i = 0; i < n_row; i++) 
  {
    irow = row_dof[i];

    if (matrix->matrix_row[irow] == nil)
    {
      row = matrix->matrix_row[irow] = get_matrix_row(matrix->fe_space);
      if (row_dof == col_dof)
      {
	row->col[0] = irow;           /* first entry is diagonal element */
	if (bound && bound[i] >= DIRICHLET)
	{
	  row->entry[0] = 1.0;
	  continue;
	}
	else
	  row->entry[0] = 0.0;
      }
      else
      {
	row->col[0] = UNUSED_ENTRY;
      }
    }
    else if (bound && bound[i] >= DIRICHLET)
      continue;

    for (j = 0; j < n_col; j++)
    {
      jcol = col_dof[j];
      row = matrix->matrix_row[irow];
      free_row = nil;
      do 
      {
	for (k=0; k<ROW_LENGTH; k++)
	{
	  if (row->col[k] == jcol)
	  {
	    row->entry[k] += sign*el_mat[i][j];
	    break;
	  }
	  if (ENTRY_NOT_USED(row->col[k])) 
	  {
	    free_col = k;
	    free_row = row;
	    if (row->col[k] == NO_MORE_ENTRIES) 
	    {
	      k = ROW_LENGTH;
	      break;
	    }
	  }
	}
	if (k < ROW_LENGTH) break;               /* done? */
	if (row->next || free_row)
	{
	  row = row->next;
	}
	else 
	{
	  free_row = row->next = get_matrix_row(matrix->fe_space);
	  free_col = 0;
	  row = nil;
	}
      } while (row);
      
      if (k >= ROW_LENGTH)                       /* not done? */
      {                     
	free_row->col[free_col]   = jcol;
	free_row->entry[free_col] = sign * el_mat[i][j];
      }
    }
  }
  return;
}

void add_element_vec(DOF_REAL_VEC *drv, REAL sign, int dim, const DOF *dof,
		     const REAL *vec, const S_CHAR *bound)
{
  FUNCNAME("add_element_vec");
  REAL   *drv_vec = nil;
  int    i;

  GET_DOF_VEC(drv_vec,drv);

  if (bound)
  {
    for (i = 0; i < dim; i++)
      if (bound[i] < DIRICHLET)
	drv_vec[dof[i]] += sign*vec[i];
  }
  else
  {
    for (i = 0; i < dim; i++)
      drv_vec[dof[i]] += sign*vec[i];

  }
  return;
}

void add_element_d_vec(DOF_REAL_D_VEC *drdv, REAL sign, int dim, 
		       const DOF  *dof, const REAL_D *vec, const S_CHAR *bound)
{
  FUNCNAME("add_element_d_vec");
  REAL_D *drdv_vec = nil;
  int    i, n;

  GET_DOF_VEC(drdv_vec, drdv);

  if (bound)
  {
    for (i = 0; i < dim; i++)
      if (bound[i] < DIRICHLET)
	for (n = 0; n < DIM_OF_WORLD; n++)
	  drdv_vec[dof[i]][n] += sign*vec[i][n];
  }
  else
  {
    for (i = 0; i < dim; i++)
      for (n = 0; n < DIM_OF_WORLD; n++)
	drdv_vec[dof[i]][n] += sign*vec[i][n];
  }
  return;
}


#define EL_BOUND(fct) !fct ? nil : (*fct)

static const EL_MATRIX_INFO *el_mat_info;
static DOF_MATRIX            *matrix;

static void update_matrix_fct(const EL_INFO *el_info)
{
  const REAL           **mat;
  const DOF            *row_dof, *col_dof;
  const S_CHAR         *bound;
  int                  n_row, n_col;
  const EL             *el = el_info->el;

  mat     = (*el_mat_info->el_matrix_fct)(el_info, el_mat_info->fill_info);
  row_dof = el_mat_info->get_row_dof(el, el_mat_info->row_admin, nil);
  n_row   = el_mat_info->n_row;
  n_col   = el_mat_info->n_col;
  if (n_col > 0 && el_mat_info->get_col_dof && el_mat_info->col_admin)
  {
    col_dof = el_mat_info->get_col_dof(el, el_mat_info->col_admin, nil);
  }
  else
  {
    col_dof = row_dof;
    n_col   = n_row;
  }
  if (col_dof == row_dof)
    bound = EL_BOUND(el_mat_info->get_bound)(el_info, nil);
  else
    bound = nil;

  add_element_matrix(matrix, el_mat_info->factor, n_row, n_col, row_dof,
		     col_dof, mat, bound);
  return;
}

void update_matrix(DOF_MATRIX *dof_matrix, const EL_MATRIX_INFO *minfo)
{
  FUNCNAME("update_matrix");
  FLAGS     fill_flag;

  TEST_EXIT(minfo)("no EL_MATRIX_INFO\n");
  TEST_EXIT(minfo->el_matrix_fct)("no el_matrix_fct in EL_MATRIX_INFO\n");
  TEST_EXIT(dof_matrix)("no DOF_MATRIX\n");

  matrix = dof_matrix;
  el_mat_info = minfo;
  if (minfo->get_bound)
    fill_flag = minfo->fill_flag|FILL_BOUND;
  else
    fill_flag = minfo->fill_flag;

  mesh_traverse(matrix->fe_space->mesh, -1, fill_flag, update_matrix_fct);
  return;
}

static const EL_VEC_INFO *el_vec_info;
static DOF_REAL_VEC      *dof_real_vec;

static void update_vec_fct(const EL_INFO *el_info)
{
  const REAL        *vec;
  const DOF         *dof;
  const S_CHAR      *bound;

  vec = (*el_vec_info->el_vec_fct)(el_info, el_vec_info->fill_info);
  dof = (*el_vec_info->get_dof)(el_info->el, el_vec_info->admin, nil);
  bound = EL_BOUND(el_vec_info->get_bound)(el_info, nil);

  add_element_vec(dof_real_vec, el_vec_info->factor, el_vec_info->n_dof, 
		  dof, vec, bound);
  return;
}

void update_real_vec(DOF_REAL_VEC *drv, const EL_VEC_INFO *vec_info)
{
  FUNCNAME("update_real_vec");
  FLAGS    fill_flag;

  TEST_EXIT(vec_info)("no EL_VEC_INFO\n");
  TEST_EXIT(vec_info->el_vec_fct)("no el_vec_fct in EL_VEC_INFO\n");
  TEST_EXIT(drv)("no DOF_REAL_VEC\n");

  dof_real_vec = drv;
  el_vec_info = vec_info;

  if (vec_info->get_bound)
    fill_flag = vec_info->fill_flag|FILL_BOUND;
  else
    fill_flag = vec_info->fill_flag;

  mesh_traverse(drv->fe_space->mesh, -1, fill_flag, update_vec_fct);
  return;
}

static const EL_VEC_D_INFO *el_vec_d_info;
static DOF_REAL_D_VEC      *dof_real_d_vec;

static void update_vec_d_fct(const EL_INFO *el_info)
{
  const REAL_D   *vec;
  const DOF      *dof;
  const S_CHAR   *bound;

  vec   = (*el_vec_d_info->el_vec_fct)(el_info, el_vec_d_info->fill_info);
  dof   = (*el_vec_d_info->get_dof)(el_info->el, el_vec_d_info->admin, nil);
  bound = EL_BOUND(el_vec_d_info->get_bound)(el_info, nil);

  add_element_d_vec(dof_real_d_vec, el_vec_d_info->factor,
		    el_vec_d_info->n_dof, dof, vec, bound);
  return;
}

void update_real_d_vec(DOF_REAL_D_VEC *drdv, const EL_VEC_D_INFO *vecd_info)
{
  FUNCNAME("update_real_d_vec");
  FLAGS    fill_flag;

  TEST_EXIT(vecd_info)("no EL_VEC_D_INFO\n");
  TEST_EXIT(vecd_info->el_vec_fct)("no el_vec_fct in EL_VEC_D_INFO\n");
  TEST_EXIT(drdv)("no DOF_REAL_D_VEC\n");

  el_vec_d_info = vecd_info;
  dof_real_d_vec = drdv;
  if (vecd_info->get_bound)
    fill_flag = vecd_info->fill_flag|FILL_BOUND;
  else
    fill_flag = vecd_info->fill_flag;

  mesh_traverse(drdv->fe_space->mesh, -1, fill_flag, update_vec_d_fct);
  return;
}


