/*--------------------------------------------------------------------------*/
/* ALBERTA:  an Adaptive multi Level finite element toolbox using           */
/*           Bisectioning refinement and Error control by Residual          */
/*           Techniques for scientific Applications                         */
/*                                                                          */
/* file:     estimator.c                                                    */
/*                                                                          */
/* description:  residual error estimator for elliptic and parabolic        */
/*               problems                                                   */
/*                                                                          */
/*--------------------------------------------------------------------------*/
/*                                                                          */
/*  authors:   Alfred Schmidt                                               */
/*             Zentrum fuer Technomathematik                                */
/*             Fachbereich 3 Mathematik/Informatik                          */
/*             Universitaet Bremen                                          */
/*             Bibliothekstr. 2                                             */
/*             D-28359 Bremen, Germany                                      */
/*                                                                          */
/*             Kunibert G. Siebert                                          */
/*             Institut fuer Mathematik                                     */
/*             Universitaet Augsburg                                        */
/*             Universitaetsstr. 14                                         */
/*             D-86159 Augsburg, Germany                                    */
/*                                                                          */
/*  http://www.alberta-fem.de/                                              */
/*                                                                          */
/*--------------------------------------------------------------------------*/
/*  (c) by A. Schmidt and K.G. Siebert (1996-2004)                          */
/*                                                                          */
/*     This program is free software; you can redistribute it and/or modify */
/*     it under the terms of the GNU General Public License as published by */
/*     the Free Software Foundation; either version 2 of the License, or    */
/*     any later version.                                                   */
/*                                                                          */
/*     This program is distributed in the hope that it will be useful,      */
/*     but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/*     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/*     GNU General Public License for more details.                         */
/*--------------------------------------------------------------------------*/

#include "alberta.h"

/*--------------------------------------------------------------------------*/
/*  residual type estimator for quasi linear elliptic problem:              */
/*   -\div A \nabla u + f(.,u,\nabla u) = 0                                 */
/*--------------------------------------------------------------------------*/

struct ellipt_est_data
{
  const DOF_REAL_VEC *uh;
  const BAS_FCTS     *bas_fcts;

  const REAL_D       (*A);
  int                is_diag;
  REAL               (*f)(const EL_INFO *, const QUAD *, int, REAL,
			  const REAL_D);

  const QUAD_FAST  *quad_fast;      /*--  element integration  -------------*/
#if DIM > 1
  const QUAD       *quad;           /*--  face integration     -------------*/
#endif
  REAL             *(*rw_est)(EL *);
  REAL             *(*rw_estc)(EL *);

  REAL             *uh_el;          /*--  vector for storing uh on el  -----*/

  FLAGS            f_flag;
  int              norm;            /*--  estimated norm: H1_NORM/L2_NORM --*/
#if DIM == 1
  REAL             C0;
#else
  REAL             C0, C1, C2;
#endif
  REAL             est_sum;
  REAL             est_max;
};

#if DIM == 1
#define h2_from_det(det) ((det)*(det))
#endif
#if DIM == 2
#define h2_from_det(det) (det)
#endif
#if DIM == 3
#define h2_from_det(det)  pow((det), 2.0/3.0)
#endif

/*--------------------------------------------------------------------------*/
/* element residual:  C0*h_S^2*||-div A nabla u^h + r||_L^2(S)  (H^1)       */
/*                    C0*h_S^4*||-div A nabla u^h + r||_L^2(S)  (L^2)       */
/*--------------------------------------------------------------------------*/

static REAL el_res2(const EL_INFO *el_info, const REAL_D *Lambda, REAL det, 
		    const struct ellipt_est_data *data)
{
  const REAL_DD   *D2uhqp;
  REAL            val, riq = 0.0, h2 = h2_from_det(det), uh_qpi = 0.0;
  int             iq, i, j;
  const REAL      *uh_qp = nil, *grd_uh_qpi = nil;
  const REAL_D    *grd_uh_qp = nil;
  const QUAD_FAST *quad_fast = data->quad_fast;
  const QUAD      *quad = quad_fast->quad;

  if (data->quad_fast->bas_fcts->degree > 1)
    D2uhqp = D2_uh_at_qp(data->quad_fast, Lambda, data->uh_el, nil);
  else
    D2uhqp = nil;

  if (data->f)
  {
    if (data->f_flag & INIT_UH)
      uh_qp = uh_at_qp(quad_fast, data->uh_el, nil);
    else
      uh_qpi = 0.0;

    if (data->f_flag & INIT_GRD_UH)
      grd_uh_qp = grd_uh_at_qp(quad_fast, Lambda, data->uh_el, nil);
    else
      grd_uh_qpi = nil;
  }

  for (val = iq = 0; iq < quad->n_points; iq++)
  {
    if (data->f)
    {
      if (data->f_flag & INIT_UH)     uh_qpi     = uh_qp[iq];
      if (data->f_flag & INIT_GRD_UH) grd_uh_qpi = grd_uh_qp[iq];
      riq = (*data->f)(el_info, quad, iq, uh_qpi, (void *) grd_uh_qpi);
    }
    else
      riq = 0.0;

    if (D2uhqp)
    {
      if (data->is_diag)
      {
	for (i = 0; i < DIM_OF_WORLD; i++)
	  riq -= data->A[i][i]*D2uhqp[iq][i][i];
      }
      else
      {
	for (i = 0; i < DIM_OF_WORLD; i++)
	  for (j = 0; j < DIM_OF_WORLD; j++)
	    riq -= data->A[i][j]*D2uhqp[iq][i][j];
      }
    }
    val += quad->w[iq]*SQR(riq);
  }

  if (data->norm == L2_NORM)
    val = data->C0*h2*h2*det*val;
  else
    val = data->C0*h2*det*val;

  return(val);
}

#if DIM > 1
/*--------------------------------------------------------------------------*/
/*  face residuals:  C1*h_Gamma*||[A(u_h)]||_L^2(Gamma)^2          (H^1)    */
/*                   C1*h_S^2*h_Gamma*||[A(u_h)]||_L^2(Gamma)^2    (L^2)    */
/*  Since det_S = det_Gamma*h_Gamma we use for det_Gamma*h_Gamma the term   */
/*  0.5(det_S + det_S')                                                     */
/*--------------------------------------------------------------------------*/

static REAL jump_res2(const EL_INFO *el_info, int face, const REAL_D *Lambda,
		      REAL det, const struct ellipt_est_data *data)
{
  EL_INFO        neigh_info[1];
  int            face_ind_el[DIM], face_ind_neigh[DIM];
  EL             *neigh = NEIGH(el_info->el,el_info)[face];
  int            opp_v  = OPP_VERTEX(el_info->el,el_info)[face];
  int            i, j, i1, i2, iq;
  REAL_D         jump, grd_uh_el, grd_uh_neigh, Lambda_neigh[DIM+1];
  const REAL     *uh_neigh;
  REAL           det_neigh, lambda[DIM+1], val = 0.0;
  const BAS_FCTS *bas_fcts = data->uh->fe_space->bas_fcts;
  const QUAD     *quad = data->quad;

/*--------------------------------------------------------------------------*/
/* orient the edge/face => same quadrature nodes from both sides!           */
/*--------------------------------------------------------------------------*/

  sort_face_indices(el_info->el, face, face_ind_el);
  sort_face_indices(neigh, opp_v, face_ind_neigh);

  neigh_info->mesh = el_info->mesh;
  neigh_info->el = neigh;
  neigh_info->fill_flag = FILL_COORDS;

  for (j = 0; j < DIM_OF_WORLD; j++)
    neigh_info->coord[opp_v][j] = el_info->opp_coord[face][j];

  for (i = 0; i < DIM; i++)
  {
    i1 = face_ind_el[i];
    i2 = face_ind_neigh[i];
    for (j = 0; j < DIM_OF_WORLD; j++)
      neigh_info->coord[i2][j] = el_info->coord[i1][j];
  }

  det_neigh = el_grd_lambda(neigh_info, Lambda_neigh);
  uh_neigh = bas_fcts->get_real_vec(neigh, data->uh, nil);

/*--------------------------------------------------------------------------*/
/*  now eval the jump at all quadrature nodes                               */
/*--------------------------------------------------------------------------*/

  for (val = iq = 0; iq < quad->n_points; iq++)
  {
    lambda[face] = 0.0;
    for (i = 0; i < DIM; i++)
      lambda[face_ind_el[i]] = quad->lambda[iq][i];
    eval_grd_uh(lambda, (void *) Lambda, data->uh_el, bas_fcts, grd_uh_el);

    lambda[opp_v] = 0.0;
    for (i = 0; i < DIM; i++)
      lambda[face_ind_neigh[i]] = quad->lambda[iq][i];
    eval_grd_uh(lambda, (void *) Lambda_neigh, uh_neigh, bas_fcts, 
		grd_uh_neigh);

    if (data->is_diag) 
    {
      for (i = 0; i < DIM_OF_WORLD; i++)
	jump[i] = data->A[i][i]*(grd_uh_el[i] - grd_uh_neigh[i]);
    }
    else
    {
      for (i = 0; i < DIM_OF_WORLD; i++)
	for (jump[i] = j = 0; j < DIM_OF_WORLD; j++)
	  jump[i] += data->A[i][j]*(grd_uh_el[j] - grd_uh_neigh[j]);
    }
    val += quad->w[iq]*SCP_DOW(jump,jump);
  }

  det = 0.5*(det + det_neigh);
  if (data->norm == L2_NORM)
    return(data->C1*h2_from_det(det)*det*val);
  else
    return(data->C1*det*val);
}

/*--------------------------------------------------------------------------*/
/*  neuman residual:  C1*h_Gamma*||A(u_h).normal||_L^2(Gamma)^2    (H^1)    */
/*                    C1*h_S^2*h_Gamma*||A(u_h).normal]||_L^2(Gamma)^2 (L^2)*/
/*  Since det_S = det_Gamma*h_Gamma we use for det_Gamma*h_Gamma the term   */
/*  det_S                                                                   */
/*--------------------------------------------------------------------------*/

static REAL neumann_res2(const EL_INFO *el_info, int face, 
			 const REAL_D *Lambda, REAL det, 
			 const struct ellipt_est_data *data)
{
  int            i, j, iq;
  REAL           lambda[DIM+1], n_A_grd_uh, val;
  REAL_D         normal, grd_uh, A_grd_uh;
  const QUAD     *quad = data->quad;

  get_face_normal(el_info, face, normal);

  for (val = iq = 0; iq < quad->n_points; iq++)
  {
    for (i = 0; i < face; i++)
      lambda[i] = quad->lambda[iq][i];
    lambda[face] = 0.0;
    for (i = face+1; i < DIM+1; i++)
      lambda[i] = quad->lambda[iq][i-1];

    eval_grd_uh(lambda, Lambda, data->uh_el, data->bas_fcts, grd_uh);

    if (data->is_diag) 
    {
      for (i = 0; i < DIM_OF_WORLD; i++)
	A_grd_uh[i] = data->A[i][i]*grd_uh[i];
    }
    else
    {
      for (i = 0; i < DIM_OF_WORLD; i++)
	for (A_grd_uh[i] = j = 0; j < DIM_OF_WORLD; j++)
	  A_grd_uh[i] += data->A[i][j]*grd_uh[j];
    }
    n_A_grd_uh = SCP_DOW(normal, A_grd_uh);
    val += quad->w[iq]*SQR(n_A_grd_uh);
  }

  if (data->norm == L2_NORM)
    return(data->C1*h2_from_det(det)*det*val);
  else
    return(data->C1*det*val);
}

static int is_diag_matrix(const REAL_DD A)
{
  int i,j ;

  for (i = 0; i < DIM_OF_WORLD; i++)
    for (j = i+1; j < DIM_OF_WORLD; j++)
      if (ABS(A[i][j]) > 1.e-25 || ABS(A[j][i]) > 1.e-25) return(false);
  return(true);
}
#endif

static struct ellipt_est_data ell_data[1];

static void clear_indicator_fct(const EL_INFO *el_info)
{
  el_info->el->mark = 1;
  if (ell_data->rw_est)  *(*ell_data->rw_est)(el_info->el) = 0.0;
  if (ell_data->rw_estc) *(*ell_data->rw_estc)(el_info->el) = 0.0;
  return;
}

static void ellipt_est_fct(const EL_INFO *el_info)
{
  EL       *el = el_info->el;
  REAL     det, est_el;
  REAL_D   Lambda[DIM+1];
#if DIM > 1
  int      face;
  EL       *neigh;
#endif

/*--- if rw_est, then there might already be contributions from jumps ------*/
  est_el = ell_data->rw_est ? *(*ell_data->rw_est)(el) : 0.0;

  ell_data->bas_fcts->get_real_vec(el, ell_data->uh, ell_data->uh_el);
  det = el_grd_lambda(el_info, Lambda);

/*---  element residual  ---------------------------------------------------*/
  if (ell_data->C0)
    est_el += el_res2(el_info, (void *) Lambda, det, ell_data);

#if DIM > 1
/*---  face residuals  -----------------------------------------------------*/

  if (ell_data->C1)
  {
    for (face = 0; face < N_NEIGH; face++)
    {
      if ((neigh = NEIGH(el,el_info)[face]))
      {
/*--------------------------------------------------------------------------*/
/*  if rw_est is nil, compute jump for both neighbouring elements           */
/*                    only this allows correct computation of est_max       */
/*  if rw_est is not nil, compute jump only once for each edge/face         */
/*                    if neigh->mark: estimate not computed on neighbour!   */
/*  contribution for the element and for neighbour: 0.5*jump!               */
/*--------------------------------------------------------------------------*/
	if (!ell_data->rw_est  ||  neigh->mark)
	{
	  REAL est = jump_res2(el_info, face, (void *) Lambda, det, ell_data);

	  est_el += est;
/*--  if rw_est, add neighbour contribution to neigbour indicator  ---------*/
	  if (ell_data->rw_est) *(*ell_data->rw_est)(neigh) += est;
	}
      }
      else if (IS_NEUMANN(el_info->boundary[face]))
      {
	est_el += neumann_res2(el_info, face, (void *) Lambda, det, ell_data);
      }
    }
  }
#endif

/*--  if rw_est, write indicator to element  -------------------------------*/
  if (ell_data->rw_est) *(*ell_data->rw_est)(el) = est_el;

  ell_data->est_sum += est_el;
  ell_data->est_max = MAX(ell_data->est_max, est_el);

  el_info->el->mark = 0; /*--- all contributions are computed!  ------------*/
  return;
}

/*--------------------------------------------------------------------------*/

REAL ellipt_est(const DOF_REAL_VEC *uh, ADAPT_STAT *adapt,
		REAL *(rw_est)(EL *), REAL *(rw_estc)(EL *),
                int degree, int norm, REAL C[3], const REAL_DD A,
                REAL (*f)(const EL_INFO *,const QUAD *,int,REAL,const REAL_D),
		FLAGS f_flag)
{
  FUNCNAME("ellipt_est");
  static WORKSPACE   ws = {0, nil};
  FLAGS              fill_flag;
  const QUAD         *quad;

  if (!(ell_data->uh = uh))
  {
    MSG("no discrete solution; doing nothing\n");
    return(0.0);
  }

  ell_data->bas_fcts = uh->fe_space->bas_fcts;
  ell_data->A        = A;
#if DIM == 1
  ell_data->is_diag  = true;
#else
  ell_data->is_diag  = is_diag_matrix(A);
#endif
  ell_data->f        = f;
  ell_data->f_flag   = f_flag;

  if (degree < 0) degree = 2*ell_data->bas_fcts->degree;
  quad = get_quadrature(DIM, degree);
  if (ell_data->bas_fcts->degree > 1)
    fill_flag = INIT_PHI|INIT_GRD_PHI|INIT_D2_PHI;
  else
    fill_flag = INIT_PHI|INIT_GRD_PHI;

  ell_data->quad_fast = get_quad_fast(ell_data->bas_fcts, quad, fill_flag);
#if DIM > 1
  ell_data->quad      = get_quadrature(DIM-1, degree);
#endif
  ell_data->rw_est    = rw_est;
  ell_data->rw_estc   = rw_estc;

  REALLOC_WORKSPACE(&ws, ell_data->bas_fcts->n_bas_fcts*sizeof(REAL));
  ell_data->uh_el = ws.work;
  
  ell_data->norm  = norm;
  if (C)
  {
    ell_data->C0    = C[0] > 1.e-25 ? SQR(C[0]) : 0.0;
#if DIM > 1
    ell_data->C1    = C[1] > 1.e-25 ? SQR(C[1]) : 0.0;
    ell_data->C2    = C[2] > 1.e-25 ? SQR(C[2]) : 0.0;
#endif
  }
  else
  {
#if DIM == 1
    ell_data->C0 = 1.0;
#else
    ell_data->C0 = ell_data->C1 = ell_data->C2 = 1.0;
#endif
  }

  if (rw_est) /*---  clear error indicators   -----------------------------*/
    mesh_traverse(uh->fe_space->mesh, -1, CALL_LEAF_EL, clear_indicator_fct);

  ell_data->est_sum = ell_data->est_max = 0.0;
#if DIM == 1
  fill_flag = FILL_COORDS|CALL_LEAF_EL;
#else
  fill_flag = FILL_NEIGH|FILL_COORDS|FILL_OPP_COORDS|FILL_BOUND|CALL_LEAF_EL;
#endif
  mesh_traverse(uh->fe_space->mesh, -1, fill_flag, ellipt_est_fct);

  ell_data->est_sum = sqrt(ell_data->est_sum);
  if (adapt)
  {
    adapt->err_sum = ell_data->est_sum;
    adapt->err_max = ell_data->est_max;
  }

  return(ell_data->est_sum);
}

/*--------------------------------------------------------------------------*/
/*                                                                          */
/* error estimator for (nonlinear) heat equation:                           */
/*              u_t - div A grad u + f(x,t,u,grad u) = 0                    */
/*                                                                          */
/* eta_h = C[0]*||h^2 ((U - Uold)/tau - div A grad U + f(x,t,U,grad U))||   */
/*           + C[1]*|||h^1.5 [A grad U]|||                                  */
/* eta_c = C[2]*||(Uold - Uold_coarse)/tau||                                */
/* eta_t = C[3]*||U - Uold||                                                */
/*                                                                          */
/* heat_est() return value is the TIME DISCRETIZATION ESTIMATE, eta_t       */
/*                                                                          */
/*--------------------------------------------------------------------------*/

struct heat_est_data
{
  const DOF_REAL_VEC *uh, *uh_old;
  const BAS_FCTS     *bas_fcts;

  const REAL_D       (*A);
  int                is_diag;
  REAL               (*f)(const EL_INFO *, const QUAD *, int, REAL,
			  REAL, const REAL_D);

  const QUAD_FAST  *quad_fast;      /*--  element integration  -------------*/
#if DIM > 1
  const QUAD       *quad;           /*--  face integration     -------------*/
#endif
  REAL             *(*rw_est)(EL *);
  REAL             *(*rw_estc)(EL *);

  REAL             *uh_el;          /*--  vector for storing uh on el  -----*/
  REAL             *uh_old_el;      /*--  vector for storing uh on el  -----*/
  REAL             *uh_qp;          /*--  vector for storing uh at quadpts -*/
  REAL             *uh_old_qp;      /*--  vector for storing uh at quadpts -*/

  REAL             time, timestep;

  REAL             C0, C1, C2, C3;  /*--  interior,jump,coarsen,time coefs -*/
  REAL             est_sum;
  REAL             est_max;
  REAL             est_t_sum;
  FLAGS            f_flag;
};

static struct heat_est_data heat_data[1];

static void heat_clear_indicator_fct(const EL_INFO *el_info)
{
  el_info->el->mark = 1;
  if (heat_data->rw_est)  *(*heat_data->rw_est)(el_info->el) = 0.0;
  if (heat_data->rw_estc) *(*heat_data->rw_estc)(el_info->el) = 0.0;
  return;
}

/*--------------------------------------------------------------------------*/
/* element residual:  C0*h_S^2*|| U_t -div A nabla u^h + r ||_L^2(S)        */
/*--------------------------------------------------------------------------*/
/* time residual:  C3*|| U - Uold ||_L^2(S)                                 */
/*--------------------------------------------------------------------------*/

static REAL heat_el_res2(const EL_INFO *el_info, const REAL_D *Lambda,
			 REAL det, struct heat_est_data *data)
{
  const REAL_DD   *D2uhqp;
  REAL            val, riq = 0.0, h2 = h2_from_det(det), uh_qpi = 0.0;
  int             iq, i, j;
  const REAL      *uh_qp = nil, *uh_old_qp = nil, *grd_uh_qpi = nil;
  const REAL_D    *grd_uh_qp = nil;
  const QUAD_FAST *quad_fast = data->quad_fast;
  const QUAD      *quad = quad_fast->quad;

  uh_old_qp = uh_at_qp(quad_fast, data->uh_old_el, data->uh_old_qp);
  uh_qp     = uh_at_qp(quad_fast, data->uh_el, data->uh_qp);

  if (data->C3) {
    for (val = iq = 0; iq < quad->n_points; iq++)
    {
      riq = (uh_qp[iq] - uh_old_qp[iq]);
      val += quad->w[iq]*SQR(riq);
    }
    data->est_t_sum += data->C3*det*val;
  }

  if (!(data->C0)) return(0.0);


  if (data->quad_fast->bas_fcts->degree > 1)
    D2uhqp = D2_uh_at_qp(data->quad_fast, Lambda, data->uh_el, nil);
  else
    D2uhqp = nil;

  if (data->f)
  {
    if (!(data->f_flag & INIT_UH))
      uh_qpi = 0.0;

    if (data->f_flag & INIT_GRD_UH)
      grd_uh_qp = grd_uh_at_qp(quad_fast, Lambda, data->uh_el, nil);
    else
      grd_uh_qpi = nil;
  }

  for (val = iq = 0; iq < quad->n_points; iq++)
  {
    riq = (uh_qp[iq] - uh_old_qp[iq]) / data->timestep;

    if (data->f)
    {
      if (data->f_flag & INIT_UH)     uh_qpi     = uh_qp[iq];
      if (data->f_flag & INIT_GRD_UH) grd_uh_qpi = grd_uh_qp[iq];
      riq += (*data->f)(el_info, quad, iq, data->time, uh_qpi, 
			(void *) grd_uh_qpi);
    }

    if (D2uhqp)
    {
      if (data->is_diag)
      {
	for (i = 0; i < DIM_OF_WORLD; i++)
	  riq -= data->A[i][i]*D2uhqp[iq][i][i];
      }
      else
      {
	for (i = 0; i < DIM_OF_WORLD; i++)
	  for (j = 0; j < DIM_OF_WORLD; j++)
	    riq -= data->A[i][j]*D2uhqp[iq][i][j];
      }
    }
    val += quad->w[iq]*SQR(riq);
  }

  val = data->C0*h2*h2*det*val;

  return(val);
}

#if DIM > 1
/*--------------------------------------------------------------------------*/
/*  face residuals:  C1*h_Gamma^3*||[A * nabla u_h]||_L^2(Gamma)^2          */
/*  Since det_S = det_Gamma*h_Gamma we use for det_Gamma*h_Gamma the term   */
/*    0.5(det_S + det_S')                                                   */
/*  coarsening est:  C2*h_Gamma^3*||[nabla u_h]||_L^2(Gamma)^2              */
/*--------------------------------------------------------------------------*/

static REAL heat_jump_res2(const EL_INFO *el_info, int face,
			   const REAL_D *Lambda,
			   REAL det, const struct heat_est_data *data,
			   REAL *estc)
{
  EL_INFO        neigh_info[1];
  int            face_ind_el[DIM], face_ind_neigh[DIM];
  EL             *neigh = NEIGH(el_info->el,el_info)[face];
  int            opp_v  = OPP_VERTEX(el_info->el,el_info)[face];
  int            i, j, i1, i2, iq;
  REAL_D         jump, jumpg, grd_uh_el, grd_uh_neigh, Lambda_neigh[DIM+1];
  const REAL     *uh_neigh;
  REAL           det_neigh, lambda[DIM+1], val = 0.0, valg = 0.0, h3det;
  const BAS_FCTS *bas_fcts = data->uh->fe_space->bas_fcts;
  const QUAD     *quad = data->quad;

/*--------------------------------------------------------------------------*/
/* orient the edge/face => same quadrature nodes from both sides!           */
/*--------------------------------------------------------------------------*/

  sort_face_indices(el_info->el, face, face_ind_el);
  sort_face_indices(neigh, opp_v, face_ind_neigh);

  neigh_info->mesh = el_info->mesh;
  neigh_info->el = neigh;
  neigh_info->fill_flag = FILL_COORDS;

  for (j = 0; j < DIM_OF_WORLD; j++)
    neigh_info->coord[opp_v][j] = el_info->opp_coord[face][j];

  for (i = 0; i < DIM; i++)
  {
    i1 = face_ind_el[i];
    i2 = face_ind_neigh[i];
    for (j = 0; j < DIM_OF_WORLD; j++)
      neigh_info->coord[i2][j] = el_info->coord[i1][j];
  }

  det_neigh = el_grd_lambda(neigh_info, Lambda_neigh);
  uh_neigh = bas_fcts->get_real_vec(neigh, data->uh, nil);

/*--------------------------------------------------------------------------*/
/*  now eval the jump at all quadrature nodes                               */
/*--------------------------------------------------------------------------*/

  for (val = iq = 0; iq < quad->n_points; iq++)
  {
    lambda[face] = 0.0;
    for (i = 0; i < DIM; i++)
      lambda[face_ind_el[i]] = quad->lambda[iq][i];
    eval_grd_uh(lambda, (void *) Lambda, data->uh_el, bas_fcts, grd_uh_el);

    lambda[opp_v] = 0.0;
    for (i = 0; i < DIM; i++)
      lambda[face_ind_neigh[i]] = quad->lambda[iq][i];
    eval_grd_uh(lambda, (void *) Lambda_neigh, uh_neigh, bas_fcts, 
		grd_uh_neigh);

    for (i = 0; i < DIM_OF_WORLD; i++) {
	jumpg[i] = grd_uh_el[i] - grd_uh_neigh[i];
    }
    if (data->is_diag) 
    {
	for (i = 0; i < DIM_OF_WORLD; i++)
	    jump[i] = data->A[i][i]*jumpg[i];
    }
    else
    {
      for (i = 0; i < DIM_OF_WORLD; i++)
	  for (jump[i] = j = 0; j < DIM_OF_WORLD; j++)
	      jump[i] += data->A[i][j] * jumpg[j];
    }
    val += quad->w[iq]*SCP_DOW(jump,jump);
    valg += quad->w[iq]*SCP_DOW(jumpg,jumpg);
  }

  det = 0.5*(det + det_neigh);
  h3det = h2_from_det(det)*det;

  *estc = data->C2 * h3det * valg;

  return(data->C1 * h3det * val);
}

/*--------------------------------------------------------------------------*/
/*  Neumann residual:  C1*h_Gamma*||A(u_h).normal||_L^2(Gamma)^2            */
/*  Since det_S = det_Gamma*h_Gamma we use for det_Gamma*h_Gamma the term   */
/*  det_S                                                                   */
/*--------------------------------------------------------------------------*/

static REAL heat_neumann_res2(const EL_INFO *el_info, int face, 
			      const REAL_D *Lambda, REAL det, 
			      const struct heat_est_data *data)
{
  int            i, j, iq;
  REAL           lambda[DIM+1], n_A_grd_uh, val;
  REAL_D         normal, grd_uh, A_grd_uh;
  const QUAD     *quad = data->quad;

  get_face_normal(el_info, face, normal);

  for (val = iq = 0; iq < quad->n_points; iq++)
  {
    for (i = 0; i < face; i++)
      lambda[i] = quad->lambda[iq][i];
    lambda[face] = 0.0;
    for (i = face+1; i < DIM+1; i++)
      lambda[i] = quad->lambda[iq][i-1];

    eval_grd_uh(lambda, Lambda, data->uh_el, data->bas_fcts, grd_uh);

    if (data->is_diag) 
    {
      for (i = 0; i < DIM_OF_WORLD; i++)
	A_grd_uh[i] = data->A[i][i]*grd_uh[i];
    }
    else
    {
      for (i = 0; i < DIM_OF_WORLD; i++)
	for (A_grd_uh[i] = j = 0; j < DIM_OF_WORLD; j++)
	  A_grd_uh[i] += data->A[i][j]*grd_uh[j];
    }
    n_A_grd_uh = SCP_DOW(normal, A_grd_uh);
    val += quad->w[iq]*SQR(n_A_grd_uh);
  }

  return(data->C1*h2_from_det(det)*det*val);
}
#endif


#if DIM == 1
/* coarsening estimate in 1D is  h^3 * [u']^2                               */

static REAL heat_estc_el(const EL_INFO *el_info, const REAL_D *Lambda,
			 REAL det, const struct heat_est_data *data)
{
    EL_INFO    neigh_info[1];
    EL         *neigh;
    int        face, opp_v, j;
    REAL_D     grd_uh_el, grd_uh_neigh, jumpg, Lambda_neigh[DIM+1];
    REAL       det_neigh, valg;
    static REAL lambdas[2][DIM+1]={{0.0,1.0},{1.0,0.0}};
    const REAL     *uh_neigh;
    const BAS_FCTS *bas_fcts = data->uh->fe_space->bas_fcts;

    neigh_info->mesh = el_info->mesh;
    neigh_info->fill_flag = FILL_COORDS;
    valg = 0.0;

    for (face = 0; face < N_NEIGH; face++) {
	opp_v = 1 - face;
	neigh = NEIGH(el_info->el,el_info)[face];
	if (neigh) {
	    neigh_info->el = neigh;
	    for (j = 0; j < DIM_OF_WORLD; j++) {
		neigh_info->coord[opp_v][j] = el_info->opp_coord[face][j];
		neigh_info->coord[face][j] = el_info->coord[opp_v][j];
	    }

	    det_neigh = el_grd_lambda(neigh_info, Lambda_neigh);
	    uh_neigh = bas_fcts->get_real_vec(neigh, data->uh, nil);

	    eval_grd_uh(lambdas[face], (void *) Lambda, 
			data->uh_el, bas_fcts, grd_uh_el);
	
	    eval_grd_uh(lambdas[opp_v], (void *) Lambda_neigh, 
			uh_neigh, bas_fcts, grd_uh_neigh);
	    for (j = 0; j < DIM_OF_WORLD; j++) {
		jumpg[j] = grd_uh_el[j] - grd_uh_neigh[j];
	    }
	    valg += SCP_DOW(jumpg,jumpg);
	}
    }

    valg *= data->C2 * det * det * det;

    return(valg);
}

#endif

/*--------------------------------------------------------------------------*/

static void heat_est_fct(const EL_INFO *el_info)
{
  EL       *el = el_info->el;
  REAL     det, est_el, estc_el;
  REAL_D   Lambda[DIM+1];
#if DIM > 1
  int      face;
  EL       *neigh;
#endif

/*--- if rw_est, then there might already be contributions from jumps ------*/
  est_el = heat_data->rw_est ? *(*heat_data->rw_est)(el) : 0.0;
  estc_el = heat_data->rw_estc ? *(*heat_data->rw_estc)(el) : 0.0;

  heat_data->bas_fcts->get_real_vec(el, heat_data->uh,
				    heat_data->uh_el);
  heat_data->bas_fcts->get_real_vec(el, heat_data->uh_old,
				    heat_data->uh_old_el);
  det = el_grd_lambda(el_info, Lambda);

/*---  element and time residual  ------------------------------------------*/
  if (heat_data->C0 || heat_data->C3)
    est_el += heat_el_res2(el_info, (void *) Lambda, det, heat_data);

#if DIM > 1
/*---  face residuals  -----------------------------------------------------*/

  if (heat_data->C1 || heat_data->C2)
  {
    for (face = 0; face < N_NEIGH; face++)
    {
      if ((neigh = NEIGH(el,el_info)[face]))
      {
/*--------------------------------------------------------------------------*/
/*  if rw_est is nil, compute jump for both neighbouring elements           */
/*                    only this allows correct computation of est_max       */
/*  if rw_est is not nil, compute jump only once for each edge/face         */
/*                    if neigh->mark: estimate not computed on neighbour!   */
/*  contribution for the element and for neighbour: 0.5*jump!               */
/*--------------------------------------------------------------------------*/
	if (!heat_data->rw_est  ||  neigh->mark)
	{
	  REAL estc;
	  REAL est = heat_jump_res2(el_info, face, (void *) Lambda, det, 
				    heat_data, &estc);

	  est_el += est;
	  estc_el += estc;
/*--  if rw_est, add neighbour contribution to neigbour indicator  ---------*/
	  if (heat_data->rw_est) *(*heat_data->rw_est)(neigh) += est;
	  if (heat_data->rw_estc) *(*heat_data->rw_estc)(neigh) += estc;
	}
      }
      else if (IS_NEUMANN(el_info->boundary[face]))
      {
	est_el += heat_neumann_res2(el_info, face, (void *) Lambda, det, 
				    heat_data);
      }
    }
  }
#endif

#if DIM==1
/*--  if rw_estc, calculate coarsening error estimate  ---------------------*/
  if (heat_data->rw_estc && heat_data->C2)
  {
    *(*heat_data->rw_estc)(el) = heat_estc_el(el_info, (void *) Lambda,
					      det, heat_data);
  }
#endif

/*--  if rw_est, write indicator to element  -------------------------------*/
  if (heat_data->rw_est) *(*heat_data->rw_est)(el) = est_el;

  heat_data->est_sum += est_el;
  heat_data->est_max = MAX(heat_data->est_max, est_el);

  el_info->el->mark = 0; /*--- all contributions are computed!  ------------*/

  return;
}

/*--------------------------------------------------------------------------*/

REAL heat_est(const DOF_REAL_VEC *uh, ADAPT_INSTAT *adapt,
	      REAL *(rw_est)(EL *), REAL *(rw_estc)(EL *),
	      int degree, REAL C[4], 
	      const DOF_REAL_VEC *uh_old, const REAL_DD A,
	      REAL (*f)(const EL_INFO *, const QUAD *, int iq, REAL t,
			REAL u, const REAL_D grd_u),
	      FLAGS f_flag)
{
  FUNCNAME("heat_est");
  static WORKSPACE   ws = {0, nil};
  FLAGS              fill_flag;
  const QUAD         *quad;

  if (!(heat_data->uh = uh))
  {
    MSG("no discrete solution; doing nothing\n");
    return(0.0);
  }

  if (!(heat_data->uh_old = uh_old))
  {
    MSG("no discrete solution from previous timestep; doing nothing\n");
    /* here, initial error could be calculated... */
    return(0.0);
  }

  heat_data->bas_fcts = uh->fe_space->bas_fcts;
  heat_data->A        = A;
#if DIM == 1
  heat_data->is_diag  = true;
#else
  heat_data->is_diag  = is_diag_matrix(A);
#endif
  heat_data->f        = f;
  heat_data->f_flag   = f_flag;

  if (degree < 0) degree = 2*heat_data->bas_fcts->degree;
  quad = get_quadrature(DIM, degree);
  if (heat_data->bas_fcts->degree > 1)
    fill_flag = INIT_PHI|INIT_GRD_PHI|INIT_D2_PHI;
  else
    fill_flag = INIT_PHI|INIT_GRD_PHI;

  heat_data->quad_fast = get_quad_fast(heat_data->bas_fcts, quad, fill_flag);
#if DIM > 1
  heat_data->quad      = get_quadrature(DIM-1, degree);
#endif
  heat_data->rw_est    = rw_est;
  heat_data->rw_estc   = rw_estc;

  REALLOC_WORKSPACE(&ws, 2*(heat_data->bas_fcts->n_bas_fcts
			    + heat_data->quad_fast->n_points)*sizeof(REAL));
  heat_data->uh_el     = ws.work;
  heat_data->uh_old_el = heat_data->uh_el + heat_data->bas_fcts->n_bas_fcts;
  heat_data->uh_qp     =
    heat_data->uh_old_el + heat_data->bas_fcts->n_bas_fcts;
  heat_data->uh_old_qp = heat_data->uh_qp + heat_data->quad_fast->n_points;

#if 0
  heat_data->uh_qp     = MEM_ALLOC(6, REAL);
  heat_data->uh_old_qp = MEM_ALLOC(6, REAL);
#endif

  
  if (C)
  {
    heat_data->C0    = C[0] > 1.e-25 ? SQR(C[0]) : 0.0;
    heat_data->C1    = C[1] > 1.e-25 ? SQR(C[1]) : 0.0;
    heat_data->C2    = C[2] > 1.e-25 ? SQR(C[2]) : 0.0;
    heat_data->C3    = C[3] > 1.e-25 ? SQR(C[3]) : 0.0;
  }
  else
  {
    heat_data->C0 = heat_data->C1 = heat_data->C2 = heat_data->C3 = 1.0;
  }

  heat_data->time = adapt->time;
  heat_data->timestep = adapt->timestep;


  if (rw_est)  /*---  clear error indicators   -----------------------------*/
    mesh_traverse(uh->fe_space->mesh, -1, CALL_LEAF_EL,
		  heat_clear_indicator_fct);

  heat_data->est_sum = heat_data->est_max = heat_data->est_t_sum = 0.0;
  fill_flag = FILL_NEIGH|FILL_COORDS|FILL_OPP_COORDS|FILL_BOUND|CALL_LEAF_EL;

  mesh_traverse(uh->fe_space->mesh, -1, fill_flag, heat_est_fct);

  heat_data->est_sum   = sqrt(heat_data->est_sum);
  heat_data->est_t_sum = sqrt(heat_data->est_t_sum);
  if (adapt)
  {
    adapt->adapt_space->err_sum = heat_data->est_sum;
    adapt->adapt_space->err_max = heat_data->est_max;
/*     adapt->err_t = heat_data->est_t_sum; */
  }


  return(heat_data->est_t_sum);
}

