/*--------------------------------------------------------------------------*/
/* ALBERTA:  an Adaptive multi Level finite element toolbox using           */
/*           Bisectioning refinement and Error control by Residual          */
/*           Techniques for scientific Applications                         */
/*                                                                          */
/* file:     eval.c                                                         */
/*                                                                          */
/* description:  routines for evaluation fe-functions and derivatives       */
/*                                                                          */
/*--------------------------------------------------------------------------*/
/*                                                                          */
/*  authors:   Alfred Schmidt                                               */
/*             Zentrum fuer Technomathematik                                */
/*             Fachbereich 3 Mathematik/Informatik                          */
/*             Universitaet Bremen                                          */
/*             Bibliothekstr. 2                                             */
/*             D-28359 Bremen, Germany                                      */
/*                                                                          */
/*             Kunibert G. Siebert                                          */
/*             Institut fuer Mathematik                                     */
/*             Universitaet Augsburg                                        */
/*             Universitaetsstr. 14                                         */
/*             D-86159 Augsburg, Germany                                    */
/*                                                                          */
/*  http://www.alberta-fem.de/                                              */
/*                                                                          */
/*--------------------------------------------------------------------------*/
/*  (c) by A. Schmidt and K.G. Siebert (1996-2004)                          */
/*                                                                          */
/*     This program is free software; you can redistribute it and/or modify */
/*     it under the terms of the GNU General Public License as published by */
/*     the Free Software Foundation; either version 2 of the License, or    */
/*     any later version.                                                   */
/*                                                                          */
/*     This program is distributed in the hope that it will be useful,      */
/*     but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/*     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/*     GNU General Public License for more details.                         */
/*--------------------------------------------------------------------------*/

#include "alberta.h"

/*--------------------------------------------------------------------------*/
/*  some routines for evaluation of a finite element function, its gradient */
/*  and second derivatives; all those with _fast use the preevaluated       */
/*  basis functions at that point.                                          */
/*--------------------------------------------------------------------------*/

REAL eval_uh(const REAL lambda[DIM+1], const REAL *uh_loc, const BAS_FCTS *b)
{
  int       i;
  BAS_FCT   **phi = b->phi;
  REAL      val = 0.0;

  for (i = 0; i < b->n_bas_fcts; i++)
    val += uh_loc[i]*phi[i](lambda);

  return(val);
}

const REAL *eval_uh_d(const REAL lambda[DIM+1], const REAL_D *uh_loc,
		      const BAS_FCTS *b, REAL_D values)
{
  static REAL_D Values;
  int       i, n;
  BAS_FCT   **phi = b->phi;
  REAL      phil, *val = values ? values : Values;

  for (n = 0; n < DIM_OF_WORLD; n++)
    val[n] = 0;

  for (i = 0; i < b->n_bas_fcts; i++)
  {
    phil = phi[i](lambda);
    for (n = 0; n < DIM_OF_WORLD; n++)
      val[n] += uh_loc[i][n]*phil;
  }

  return((const REAL *) val);
}


REAL eval_uh_fast(const REAL *uh_loc, const REAL *phi_val, int n_bas_fcts)
{
  int       i;
  REAL      val = 0.0;

  for (i = 0; i < n_bas_fcts; i++)
    val += uh_loc[i]*phi_val[i];

  return(val);
}

const REAL *eval_uh_d_fast(const REAL_D *uh_loc, const REAL *phi_val,
			   int n_bas_fcts, REAL_D values)
{
  static REAL_D  Values;
  int            i, n;
  REAL           *val = values ? values : Values;

  for (n = 0; n < DIM_OF_WORLD; n++)
    val[n] = 0.0;

  for (i = 0; i < n_bas_fcts; i++) 
    for (n = 0; n < DIM_OF_WORLD; n++)
      val[n] += uh_loc[i][n]*phi_val[i];

  return((const REAL *) val);
}

const REAL *eval_grd_uh(const REAL lambda[DIM+1], 
			const REAL_D Lambda[DIM+1],
			const REAL *uh_loc, const BAS_FCTS *b, REAL_D grd_uh)
{
  static REAL_D  grd;
  const REAL     *grd_b;
  int            i, j;
  GRD_BAS_FCT    **grd_phi = b->grd_phi;
  REAL           *val, grd1[DIM+1];
  
  val = grd_uh ? grd_uh : grd;
 
  for (j = 0; j < DIM+1; j++)
    grd1[j] = 0.0;

  for (i = 0; i < b->n_bas_fcts; i++)
  {
    grd_b = grd_phi[i](lambda);
    for (j = 0; j < DIM+1; j++)
      grd1[j] += uh_loc[i]*grd_b[j];
  }

  for (i = 0; i < DIM_OF_WORLD; i++)
  {
    for (val[i] = j = 0; j < DIM+1; j++)
      val[i] += Lambda[j][i]*grd1[j];
  }
  
  return((const REAL *) val);
}

const REAL_D *eval_grd_uh_d(const REAL lambda[DIM+1], 
			    const REAL_D Lambda[DIM+1],
			    const REAL_D *uh_loc, const BAS_FCTS *b,
			    REAL_DD grd_uh)
{
  static REAL_DD  grd;
  const REAL      *grd_b;
  int             i, j, n;
  GRD_BAS_FCT     **grd_phi = b->grd_phi;
  REAL_D          *val;
  REAL            grd1[DIM_OF_WORLD][DIM+1];
  
  val = grd_uh ? grd_uh : grd;
 
  for (n = 0; n < DIM_OF_WORLD; n++)
    for (j = 0; j < DIM+1; j++)
      grd1[n][j] = 0.0;

  for (i = 0; i < b->n_bas_fcts; i++)
  {
    grd_b = grd_phi[i](lambda);
    for (j = 0; j < DIM+1; j++)
      for (n = 0; n < DIM_OF_WORLD; n++)
	grd1[n][j] += uh_loc[i][n]*grd_b[j];
  }

  for (i = 0; i < DIM_OF_WORLD; i++)
  {
      for (n = 0; n < DIM_OF_WORLD; n++)
	for (val[n][i] = j = 0; j < DIM+1; j++)
	  val[n][i] += Lambda[j][i]*grd1[n][j];
  }
  
  return((const REAL_D *) val);
}


const REAL *eval_grd_uh_fast(const REAL_D Lambda[DIM+1], 
			     const REAL *uh_loc, const REAL (*grd_phi)[DIM+1],
			     int n_bas_fcts, REAL_D grd_uh)
{
  static REAL_D  grd;
  int            i, j;
  REAL           *val, grd1[DIM+1];
  
  val = grd_uh ? grd_uh : grd;

  for (j = 0; j < DIM+1; j++)
    grd1[j] = 0.0;

  for (i = 0; i < n_bas_fcts; i++)
  {
    for (j = 0; j < DIM+1; j++)
      grd1[j] += uh_loc[i]*grd_phi[i][j];
  }

  for (i = 0; i < DIM_OF_WORLD; i++)
  {
    for (val[i] = j = 0; j < DIM+1; j++)
      val[i] += Lambda[j][i]*grd1[j];
  }
  
  return((const REAL *) val);
}

const REAL_D *eval_grd_uh_d_fast(const REAL_D Lambda[DIM+1], 
				 const REAL_D *uh_loc, 
				 const REAL (*grd_phi)[DIM+1],
				 int n_bas_fcts, REAL_DD grd_uh)
{
  static REAL_DD  grd;
  int             i, j, n;
  REAL_D          *val;
  REAL            grd1[DIM+1];
  
  val = grd_uh ? grd_uh : grd;

  for (n = 0; n < DIM_OF_WORLD; n++)
  {
    for (j = 0; j < DIM+1; j++)
      grd1[j] = 0.0;

    for (i = 0; i < n_bas_fcts; i++)
    {
      for (j = 0; j < DIM+1; j++)
	grd1[j] += uh_loc[i][n]*grd_phi[i][j];
    }

    for (i = 0; i < DIM_OF_WORLD; i++)
    {
      for (val[n][i] = j = 0; j < DIM+1; j++)
	val[n][i] += Lambda[j][i]*grd1[j];
    }
  }
  
  return((const REAL_D *) val);
}

const REAL_D *eval_D2_uh(const REAL lambda[DIM+1],
			 const REAL_D Lambda[DIM+1],
			 const REAL *uh_loc, const BAS_FCTS *b, REAL_DD D2_uh)
{
  static REAL_DD D2;
  const REAL     (*D2_b)[DIM+1];
  int            i, j, k, l;
  D2_BAS_FCT     **D2_phi = b->D2_phi;
  REAL           (*val)[DIM_OF_WORLD], D2_tmp[DIM+1][DIM+1];
  
  val = D2_uh ? D2_uh : D2;
  
  for (k = 0; k < DIM+1; k++)
    for (l = 0; l < DIM+1; l++)
      D2_tmp[k][l] = 0.0;

  for (i = 0; i < b->n_bas_fcts; i++)
  {
    D2_b = D2_phi[i](lambda);
    for (k = 0; k < DIM+1; k++)
      for (l = 0; l < DIM+1; l++)
      D2_tmp[k][l] += uh_loc[i]*D2_b[k][l];
  }

  for (i = 0; i < DIM_OF_WORLD; i++)
    for (j = 0; j < DIM_OF_WORLD; j++)
    {
      val[i][j] = 0.0;
      for (k = 0; k < DIM+1; k++)
	for (l = 0; l < DIM+1; l++)
	  val[i][j] += Lambda[k][i]*Lambda[l][j]*D2_tmp[k][l];
  }
  
  return((const REAL_D *) val);
}

const REAL_DD *eval_D2_uh_d(const REAL lambda[DIM+1],
			    const REAL_D Lambda[DIM+1],
			    const REAL_D *uh_loc, const BAS_FCTS *b, 
			    REAL_DD *D2_uh)
{
  static REAL_DD  D2[DIM_OF_WORLD];
  const REAL      (*D2_b)[DIM+1];
  int             i, j, k, l, n;
  D2_BAS_FCT      **D2_phi = b->D2_phi;
  REAL            D2_tmp[DIM+1][DIM+1];
  REAL_DD         *val;
  
  val = D2_uh ? D2_uh : D2;
  
  for (n = 0; n < DIM_OF_WORLD; n++)
  {
    for (k = 0; k < DIM+1; k++)
      for (l = 0; l < DIM+1; l++)
	D2_tmp[k][l] = 0.0;

    for (i = 0; i < b->n_bas_fcts; i++)
    {
      D2_b = D2_phi[i](lambda);
      for (k = 0; k < DIM+1; k++)
	for (l = 0; l < DIM+1; l++)
	  D2_tmp[k][l] += uh_loc[i][n]*D2_b[k][l];
    }

    for (i = 0; i < DIM_OF_WORLD; i++)
      for (j = 0; j < DIM_OF_WORLD; j++)
      {
	val[n][i][j] = 0.0;
	for (k = 0; k < DIM+1; k++)
	  for (l = 0; l < DIM+1; l++)
	    val[n][i][j] += Lambda[k][i]*Lambda[l][j]*D2_tmp[k][l];
      }
  }
  
  return((const REAL_DD *) val);
}

const REAL_D *eval_D2_uh_fast(const REAL_D Lambda[DIM+1], 
			      const REAL *uh_loc, 
			      const REAL (*D2_phi)[DIM+1][DIM+1],
			      int n_bas_fcts, REAL_DD D2_uh)
{
  static REAL_DD   D2;
  int              i, j, k, l;
  REAL             (*val)[DIM_OF_WORLD], D2_tmp[DIM+1][DIM+1];
  
  val = D2_uh ? D2_uh : D2;
  
  for (k = 0; k < DIM+1; k++)
    for (l = 0; l < DIM+1; l++)
      D2_tmp[k][l] = 0.0;

  for (i = 0; i < n_bas_fcts; i++)
  {
    for (k = 0; k < DIM+1; k++)
      for (l = 0; l < DIM+1; l++)
      D2_tmp[k][l] += uh_loc[i]*D2_phi[i][k][l];
  }

  for (i = 0; i < DIM_OF_WORLD; i++)
    for (j = 0; j < DIM_OF_WORLD; j++)
    {
      val[i][j] = 0.0;
      for (k = 0; k < DIM+1; k++)
	for (l = 0; l < DIM+1; l++)
	  val[i][j] += Lambda[k][i]*Lambda[l][j]*D2_tmp[k][l];
  }
  
  return((const REAL_D *) val);
}

const REAL_DD *eval_D2_uh_d_fast(const REAL_D Lambda[DIM+1], 
				 const REAL_D *uh_loc, 
				 const REAL (*D2_phi)[DIM+1][DIM+1],
				 int n_bas_fcts, REAL_DD *D2_uh)
{
  static REAL_DD   D2[DIM_OF_WORLD];
  int              i, j, k, l, n;
  REAL             D2_tmp[DIM+1][DIM+1];
  REAL_DD          *val;

  val = D2_uh ? D2_uh : D2;
 
  for (n = 0; n < DIM_OF_WORLD; n++)
  {
    for (k = 0; k < DIM+1; k++)
      for (l = 0; l < DIM+1; l++)
	D2_tmp[k][l] = 0.0;

    for (i = 0; i < n_bas_fcts; i++)
    {
      for (k = 0; k < DIM+1; k++)
	for (l = 0; l < DIM+1; l++)
	  D2_tmp[k][l] += uh_loc[i][n]*D2_phi[i][k][l];
    }

    for (i = 0; i < DIM_OF_WORLD; i++)
      for (j = 0; j < DIM_OF_WORLD; j++)
      {
	val[n][i][j] = 0.0;
	for (k = 0; k < DIM+1; k++)
	  for (l = 0; l < DIM+1; l++)
	    val[n][i][j] += Lambda[k][i]*Lambda[l][j]*D2_tmp[k][l];
      }
  }
  return((const REAL_DD *) val);
}

/*--------------------------------------------------------------------------*/
/*  standard routines for getting values of a finite element function and   */
/*  its first (and second) derivatives at the quadrature points             */
/*--------------------------------------------------------------------------*/

const REAL *uh_at_qp(const QUAD_FAST *fast, const REAL *uh_loc, REAL *vec)
{
  FUNCNAME("uh_at_qp");
  static REAL    *quad_vec = nil;
  static size_t  size = 0;
  REAL           *val, **phi;
  int            i, j;

  if (vec)
  {
    val = vec;
  }
  else
  {
    if (size < (size_t) fast->quad->n_points) 
    {
      size_t  new_size = MAX(MAX_N_QUAD_POINTS, fast->quad->n_points);
      quad_vec = MEM_REALLOC(quad_vec, size, new_size, REAL);
      size = new_size;
    }
    val = quad_vec;
  }

  phi = fast->phi;
  for (i = 0; i < fast->n_points; i++)
    for (val[i] = j = 0; j < fast->n_bas_fcts; j++)
      val[i] += uh_loc[j]*phi[i][j];
    
  return((const REAL *) val);
}

const REAL_D *grd_uh_at_qp(const QUAD_FAST *fast,
			   const REAL_D Lambda[DIM+1],
			   const REAL *uh_loc, REAL_D *vec)
{
  FUNCNAME("grd_uh_at_qp");
  static REAL_D  *quad_vec = nil;
  static size_t  size = 0;
  REAL_D         *val;
  REAL           (*grd_phi)[DIM+1], grd1[DIM+1];
  int            i, j, k, l;

  if (vec)
  {
    val = vec;
  }
  else
  {
    if (size < (size_t) fast->quad->n_points) 
    {
      size_t  new_size = MAX(MAX_N_QUAD_POINTS, fast->quad->n_points);
      quad_vec = MEM_REALLOC(quad_vec, size, new_size, REAL_D);
      size = new_size;
    }
    val = quad_vec;
  }

  for (i = 0; i < fast->n_points; i++)
  {
    for (k = 0; k < DIM+1; k++)
      grd1[k] = 0.0;
    grd_phi = fast->grd_phi[i];
    for (j = 0; j < fast->n_bas_fcts; j++)
    {
      for (k = 0; k < DIM+1; k++)
	grd1[k] += uh_loc[j]*grd_phi[j][k];
    }

    for (l = 0; l < DIM_OF_WORLD; l++)
    {
      for (val[i][l] = k = 0; k < DIM+1; k++)
	val[i][l] += Lambda[k][l]*grd1[k];
    }
  }

  return((const REAL_D *) val);
}

const REAL_DD *D2_uh_at_qp(const QUAD_FAST *fast,
			   const REAL_D Lambda[DIM+1],
			   const REAL *uh_loc, REAL_DD *vec)
{
  FUNCNAME("D2_uh_at_qp");
  static REAL_DD  *quad_vec = nil;
  static size_t   size = 0;
  REAL_DD         *val;
  REAL            (*D2_phi)[DIM+1][DIM+1], D2_tmp[DIM+1][DIM+1];
  int             i, j, k, l, iq;

  if (vec)
  {
    val = vec;
  }
  else
  {
    if (size < (size_t) fast->n_points) 
    {
      size_t  new_size = MAX(MAX_N_QUAD_POINTS, fast->n_points);
      quad_vec = MEM_REALLOC(quad_vec, size, new_size, REAL_DD);
      size = new_size;
    }
    val = quad_vec;
  }

  for (iq = 0; iq < fast->n_points; iq++)
  {
    for (k = 0; k < DIM+1; k++)
      for (l = 0; l < DIM+1; l++)
	D2_tmp[k][l] = 0.0;

    for (i = 0; i < fast->n_bas_fcts; i++)
    {
      D2_phi = fast->D2_phi[iq];
      for (k = 0; k < DIM+1; k++)
	for (l = 0; l < DIM+1; l++)
	  D2_tmp[k][l] += uh_loc[i]*D2_phi[i][k][l];
    }

    for (i = 0; i < DIM_OF_WORLD; i++)
      for (j = 0; j < DIM_OF_WORLD; j++)
      {
	val[iq][i][j] = 0.0;
	for (k = 0; k < DIM+1; k++)
	  for (l = 0; l < DIM+1; l++)
	    val[iq][i][j] += Lambda[k][i]*Lambda[l][j]*D2_tmp[k][l];
      }
  }

  return((const REAL_DD *) val);
}

const REAL_D *uh_d_at_qp(const QUAD_FAST *fast, const REAL_D *uh_loc,
			 REAL_D *vec)
{
  FUNCNAME("uh_d_at_qp");
  static REAL_D  *quad_vec = nil;
  static size_t  size = 0;
  REAL           **phi;
  REAL_D         *val;
  int            i, j, k;

  if (vec)
  {
    val = vec;
  }
  else
  {
    if (size < (size_t) fast->quad->n_points) 
    {
      size_t  new_size = MAX(MAX_N_QUAD_POINTS, fast->quad->n_points);
      quad_vec = MEM_REALLOC(quad_vec, size, new_size, REAL_D);
      size = new_size;
    }
    val = quad_vec;
  }

  phi = fast->phi;
  for (i = 0; i < fast->n_points; i++)
    for (k = 0; k < DIM_OF_WORLD; k++)
      for (val[i][k] = j = 0; j < fast->n_bas_fcts; j++)
	val[i][k] += uh_loc[j][k]*phi[i][j];
    
  return((const REAL_D *) val);
}

const REAL_DD *grd_uh_d_at_qp(const QUAD_FAST *fast,
			      const REAL_D Lambda[DIM+1], 
			      const REAL_D *uh_loc, REAL_DD *vec)
{
  FUNCNAME("grd_d_uh_at_qp");
  static REAL_DD  *quad_vec = nil;
  static size_t   size = 0;
  REAL_DD         *val;
  REAL            (*grd_phi)[DIM+1], grd1[DIM+1];
  int             i, j, k, l, m;

  if (vec)
  {
    val = vec;
  }
  else
  {
    if (size < (size_t) fast->quad->n_points) 
    {
      size_t  new_size = MAX(MAX_N_QUAD_POINTS, fast->quad->n_points);
      
      quad_vec = MEM_REALLOC(quad_vec, size, new_size, REAL_DD);
      size = new_size;
    }
    val = quad_vec;
  }

  for (i = 0; i < fast->n_points; i++)
  {
    grd_phi = fast->grd_phi[i];

    for (k = 0; k < DIM_OF_WORLD; k++)
    {
      for (m = 0; m < DIM+1; m++)
	grd1[m] = 0.0;

      for (j = 0; j < fast->n_bas_fcts; j++)
      {
	for (m = 0; m < DIM+1; m++)
	  grd1[m] += uh_loc[j][k]*grd_phi[j][m];
      }

      for (l = 0; l < DIM_OF_WORLD; l++)
      {
	for (val[i][k][l] = m = 0; m < DIM+1; m++)
	  val[i][k][l] += Lambda[m][l]*grd1[m];
      }
    }
  }
  return((const REAL_DD *) val);
}

const REAL_DD (*D2_uh_d_at_qp(const QUAD_FAST *fast,
			      const REAL_D grd_lam[DIM+1],
			      const REAL_D *uh_loc, 
			      REAL_DD (*vec)[DIM_OF_WORLD]))[DIM_OF_WORLD]
{
  FUNCNAME("D2_uh_d_at_qp");
  static REAL_DD  (*quad_vec)[DIM_OF_WORLD] = nil;
  static size_t   size = 0;
  REAL_DD         (*val)[DIM_OF_WORLD];
  REAL            (*D2_phi)[DIM+1][DIM+1], D2_tmp[DIM+1][DIM+1];
  int             i, j, k, l, n, iq;

  if (vec)
  {
    val = vec;
  }
  else
  {
    if (size < (size_t) fast->n_points) 
    {
      size_t  new_size = MAX(MAX_N_QUAD_POINTS, fast->n_points);
      quad_vec = (REAL_DD (*) [DIM_OF_WORLD])
	MEM_REALLOC(quad_vec,size*DIM_OF_WORLD,new_size*DIM_OF_WORLD,REAL_DD);
      size = new_size;
    }
    val = quad_vec;
  }

  for (iq = 0; iq < fast->n_points; iq++)
  {
    for (n = 0; n < DIM_OF_WORLD; n++)
    {

      for (k = 0; k < DIM+1; k++)
	for (l = 0; l < DIM+1; l++)
	  D2_tmp[k][l] = 0.0;

      for (i = 0; i < fast->n_bas_fcts; i++)
      {
	D2_phi = fast->D2_phi[iq];
	for (k = 0; k < DIM+1; k++)
	  for (l = 0; l < DIM+1; l++)
	    D2_tmp[k][l] += uh_loc[i][n]*D2_phi[i][k][l];
      }

      for (i = 0; i < DIM_OF_WORLD; i++)
	for (j = 0; j < DIM_OF_WORLD; j++)
	{
	  val[iq][n][i][j] = 0.0;
	  for (k = 0; k < DIM+1; k++)
	    for (l = 0; l < DIM+1; l++)
	      val[iq][n][i][j] += grd_lam[k][i]*grd_lam[l][j]*D2_tmp[k][l];
	}
    }
  }

  return((const REAL_DD (*) [DIM_OF_WORLD]) val);
}

REAL eval_div_uh_d(const REAL lambda[DIM+1], const REAL_D Lambda[DIM+1],
		   const REAL_D *uh_loc, const BAS_FCTS *b)
{
  REAL            div;
  const REAL      *grd_b;
  int             i, j, n;
  GRD_BAS_FCT     **grd_phi = b->grd_phi;
  REAL            grd1[DIM_OF_WORLD][DIM+1];
  
  for (n = 0; n < DIM_OF_WORLD; n++)
    for (j = 0; j < DIM+1; j++)
      grd1[n][j] = 0.0;

  for (i = 0; i < b->n_bas_fcts; i++)
  {
    grd_b = grd_phi[i](lambda);
    for (j = 0; j < DIM+1; j++)
      for (n = 0; n < DIM_OF_WORLD; n++)
	grd1[n][j] += uh_loc[i][n]*grd_b[j];
  }

  for (div = n = 0; n < DIM_OF_WORLD; n++)
    for (j = 0; j < DIM+1; j++)
      div += Lambda[j][n]*grd1[n][j];
  
  return(div);
}

REAL eval_div_uh_d_fast(const REAL_D Lambda[DIM+1], const REAL_D *uh_loc, 
			const REAL (*grd_phi)[DIM+1], int n_bas_fcts)
{
  REAL      div;
  int       i, j, n;
  REAL      grd1[DIM+1];
  
  for (div = n = 0; n < DIM_OF_WORLD; n++)
  {
    for (j = 0; j < DIM+1; j++)
      grd1[j] = 0.0;

    for (i = 0; i < n_bas_fcts; i++)
    {
      for (j = 0; j < DIM+1; j++)
	grd1[j] += uh_loc[i][n]*grd_phi[i][j];
    }

    for (j = 0; j < DIM+1; j++)
      div += Lambda[j][n]*grd1[j];
  }
  
  return(div);
}

const REAL *div_uh_d_at_qp(const QUAD_FAST *fast, const REAL_D Lambda[DIM+1], 
			   const REAL_D *uh_loc, REAL *vec)
{
  FUNCNAME("div_uh_d_at_qp");
  static REAL     *quad_vec = nil;
  static size_t   size = 0;
  REAL            *val, div;
  REAL            (*grd_phi)[DIM+1], grd1[DIM+1];
  int             i, j, n, m;

  if (vec)
  {
    val = vec;
  }
  else
  {
    if (size < (size_t) fast->quad->n_points) 
    {
      size_t  new_size = MAX(MAX_N_QUAD_POINTS, fast->quad->n_points);
      
      quad_vec = MEM_REALLOC(quad_vec, size, new_size, REAL);
      size = new_size;
    }
    val = quad_vec;
  }

  for (m = 0; m < fast->n_points; m++)
  {
    grd_phi = fast->grd_phi[m];

    for (div = n = 0; n < DIM_OF_WORLD; n++)
    {
      for (j = 0; j < DIM+1; j++)
	grd1[j] = 0.0;

      for (i = 0; i < fast->n_bas_fcts; i++)
      {
	for (j = 0; j < DIM+1; j++)
	  grd1[j] += uh_loc[i][n]*grd_phi[i][j];
      }

      for (j = 0; j < DIM+1; j++)
	div += Lambda[j][n]*grd1[j];
    }
    val[m] = div;
  }

  return((const REAL *) val);
}

/*--------------------------------------------------------------------------*/
/* calculation H1 or L2 Norm of a finite element function                   */
/*--------------------------------------------------------------------------*/

static const REAL *(*get_real_vec)(const EL *, const DOF_REAL_VEC *, REAL *);
static const DOF_REAL_VEC  *uh;
static REAL                norm;
static const QUAD_FAST     *quad_fast;

static void H1_norm_uh_fct(const EL_INFO *el_info)
{
  REAL          det, norm2, normT;
  REAL_D        Lambda[DIM+1];
  const REAL    *uh_loc;
  const REAL_D  *grduh_vec;
  int         iq, j;

  det = el_grd_lambda(el_info, Lambda);
  uh_loc = (*get_real_vec)(el_info->el, uh, nil);
  grduh_vec = grd_uh_at_qp(quad_fast, (const REAL_D*) Lambda, uh_loc, nil);

  for (normT = iq = 0; iq < quad_fast->n_points; iq++)
  {
    for (norm2 = j = 0; j < DIM_OF_WORLD; j++)
      norm2 += SQR(grduh_vec[iq][j]);

    normT += quad_fast->w[iq]*norm2;
  }
  norm += det*normT;

  return;
}


REAL H1_norm_uh(const QUAD *quad, const DOF_REAL_VEC *u_h)
{
  FUNCNAME("H1_norm_uh");
  int      deg;

  if (!(uh = u_h))
  {
    ERROR("no DOF vector u_h; returning 0.0\n");
    return(0.0);
  }
  if (!quad)
  {
    deg = 2*uh->fe_space->bas_fcts->degree-2;
    quad = get_quadrature(DIM, deg);
  }
  quad_fast = get_quad_fast(uh->fe_space->bas_fcts, quad, INIT_GRD_PHI);
  get_real_vec = uh->fe_space->bas_fcts->get_real_vec;

  norm = 0.0;
  mesh_traverse(uh->fe_space->mesh, -1, CALL_LEAF_EL|FILL_COORDS,
		H1_norm_uh_fct);

  return(sqrt(norm));
}

static void L2_norm_uh_fct(const EL_INFO *el_info)
{
  REAL        det, normT;
  const REAL  *uh_loc, *uh_vec;
  int         iq;

  det = el_det(el_info);
  uh_loc = (*get_real_vec)(el_info->el, uh, nil);
  uh_vec = uh_at_qp(quad_fast, uh_loc, nil);

  for (normT = iq = 0; iq < quad_fast->n_points; iq++)
  {
    normT += quad_fast->w[iq]*SQR(uh_vec[iq]);
  }
  norm += det*normT;

  return;
}

REAL L2_norm_uh(const QUAD *quad, const DOF_REAL_VEC *u_h)
{
  FUNCNAME("L2_norm_uh");
  int      deg;

  if (!(uh = u_h))
  {
    ERROR("no DOF vector u_h; returning 0.0\n");
    return(0.0);
  }
  if (!quad)
  {
    deg = 2*uh->fe_space->bas_fcts->degree;
    quad = get_quadrature(DIM, deg);
  }
  quad_fast = get_quad_fast(uh->fe_space->bas_fcts, quad, INIT_PHI);
  get_real_vec = uh->fe_space->bas_fcts->get_real_vec;

  norm = 0.0;
  mesh_traverse(uh->fe_space->mesh, -1, CALL_LEAF_EL|FILL_COORDS,
		L2_norm_uh_fct);

  return(sqrt(norm));
}

/*--------------------------------------------------------------------------*/
/* calculation H1 or L2 Norm of a vector valued finite element function     */
/*--------------------------------------------------------------------------*/

static const REAL_D *(*get_real_d_vec)(const EL *, const DOF_REAL_D_VEC *,
				       REAL_D *);
static const DOF_REAL_D_VEC  *uh_d;

static void H1_norm_uh_d_fct(const EL_INFO *el_info)
{
  REAL          det, norm2, normT;
  REAL_D        Lambda[DIM+1];
  const REAL_D  *uh_loc;
  const REAL_DD *grduh_vec;
  int           iq, j, n;

  det = el_grd_lambda(el_info, Lambda);
  uh_loc = (*get_real_d_vec)(el_info->el, uh_d, nil);
  grduh_vec = grd_uh_d_at_qp(quad_fast, (void *) Lambda, uh_loc, nil);

  for (normT = iq = 0; iq < quad_fast->n_points; iq++)
  {
    for (norm2 = j = 0; j < DIM_OF_WORLD; j++)
      for (n = 0; n < DIM_OF_WORLD; n++)
	norm2 += SQR(grduh_vec[iq][j][n]);

    normT += quad_fast->w[iq]*norm2;
  }
  norm += det*normT;

  return;
}


REAL H1_norm_uh_d(const QUAD *quad, const DOF_REAL_D_VEC *u_h)
{
  FUNCNAME("H1_norm_uh_d");
  int      deg;

  if (!(uh_d = u_h))
  {
    ERROR("no DOF vector u_h; returning 0.0\n");
    return(0.0);
  }
  if (!quad)
  {
    deg = 2*uh_d->fe_space->bas_fcts->degree-2;
    quad = get_quadrature(DIM, deg);
  }
  quad_fast = get_quad_fast(uh_d->fe_space->bas_fcts, quad, INIT_GRD_PHI);
  get_real_d_vec = uh_d->fe_space->bas_fcts->get_real_d_vec;

  norm = 0.0;
  mesh_traverse(uh_d->fe_space->mesh, -1, CALL_LEAF_EL|FILL_COORDS, 
		H1_norm_uh_d_fct);

  return(sqrt(norm));
}

static void L2_norm_uh_d_fct(const EL_INFO *el_info)
{
  REAL         det, normT, norm2;
  const REAL_D *uh_loc, *uh_vec;
  int          iq, n;

  det = el_det(el_info);
  uh_loc = (*get_real_d_vec)(el_info->el, uh_d, nil);
  uh_vec = uh_d_at_qp(quad_fast, uh_loc, nil);

  for (normT = iq = 0; iq < quad_fast->n_points; iq++)
  {
    for (norm2 = n = 0; n < DIM_OF_WORLD; n++)
      norm2 += SQR(uh_vec[iq][n]);
    normT += quad_fast->w[iq]*norm2;
  }
  norm += det*normT;

  return;
}

REAL L2_norm_uh_d(const QUAD *quad, const DOF_REAL_D_VEC *u_h)
{
  FUNCNAME("L2_norm_uh_d");
  int      deg;

  if (!(uh_d = u_h))
  {
    ERROR("no DOF vector u_h; returning 0.0\n");
    return(0.0);
  }
  if (!quad)
  {
    deg = 2*uh_d->fe_space->bas_fcts->degree;
    quad = get_quadrature(DIM, deg);
  }
  quad_fast = get_quad_fast(uh_d->fe_space->bas_fcts, quad, INIT_PHI);
  get_real_d_vec = uh_d->fe_space->bas_fcts->get_real_d_vec;

  norm = 0.0;
  mesh_traverse(uh_d->fe_space->mesh, -1, CALL_LEAF_EL|FILL_COORDS, 
		L2_norm_uh_d_fct);

  return(sqrt(norm));
}
