/*--------------------------------------------------------------------------*/
/* ALBERTA:  an Adaptive multi Level finite element toolbox using           */
/*           Bisectioning refinement and Error control by Residual          */
/*           Techniques for scientific Applications                         */
/*                                                                          */
/* file:     memory.c                                                       */
/*                                                                          */
/* description:  routines for memory handling                               */
/*                                                                          */
/*--------------------------------------------------------------------------*/
/*                                                                          */
/*  authors:   Alfred Schmidt                                               */
/*             Zentrum fuer Technomathematik                                */
/*             Fachbereich 3 Mathematik/Informatik                          */
/*             Universitaet Bremen                                          */
/*             Bibliothekstr. 2                                             */
/*             D-28359 Bremen, Germany                                      */
/*                                                                          */
/*             Kunibert G. Siebert                                          */
/*             Institut fuer Mathematik                                     */
/*             Universitaet Augsburg                                        */
/*             Universitaetsstr. 14                                         */
/*             D-86159 Augsburg, Germany                                    */
/*                                                                          */
/*             Daniel Koester                                               */
/*             Institut fuer Mathematik                                     */
/*             Albert-Ludwigs-Universitaet Freiburg                         */
/*             Hermann-Herder-Str. 10                                       */
/*             D-79104 Freiburg                                             */
/*                                                                          */
/*  http://www.alberta-fem.de/                                              */
/*                                                                          */
/*--------------------------------------------------------------------------*/
/*  (c) by A. Schmidt and K.G. Siebert (1996-2004)                          */
/*                                                                          */
/*     This program is free software; you can redistribute it and/or modify */
/*     it under the terms of the GNU General Public License as published by */
/*     the Free Software Foundation; either version 2 of the License, or    */
/*     any later version.                                                   */
/*                                                                          */
/*     This program is distributed in the hope that it will be useful,      */
/*     but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/*     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/*     GNU General Public License for more details.                         */
/*--------------------------------------------------------------------------*/

/*--------------------------------------------------------------------------*/
/* to_do:                                                                   */
/*   free memory of blocks                               		    */
/*   be careful if mesh->max_edge_neigh changes (rc_lists...)	            */
/*                                                      		    */
/*--------------------------------------------------------------------------*/

#include "alberta.h"

const DOF_ADMIN *get_dof_admin(MESH *mesh, const char *name,
			       const int n_dof[DIM+1]);

/*--------------------------------------------------------------------------*/
/*  some routines for allocation and deallocation                           */
/*--------------------------------------------------------------------------*/

/*
 * default value for increasing size of memory in an memoryadmin.
 *
 * see also setCapacityIncrement(...)
 */
#define DEFAULTCAPACITYINCREMENT 1000


#define ALIGNSIZE(size) ((size) + ((size) % sizeof(void*) == 0 ? 0 : sizeof(void*) - (size) % sizeof(void*)))

typedef struct FreeMemory FREEMEMORY;
struct FreeMemory {
    FREEMEMORY *next;
};

typedef struct Block BLOCK;
struct Block {
    BLOCK *next;
    void *end;
};

typedef struct MemoryAdmin MEMORYADMIN;
struct MemoryAdmin {
    unsigned int capacity;
    unsigned int capacityIncrement;
    size_t objectSize;
    BLOCK *blocks;
    FREEMEMORY *freeMem;
};

static int newBlock(MEMORYADMIN* ma, unsigned int capacityIncrement) {
    FUNCNAME("newBlock");
    BLOCK* block;
    FREEMEMORY *freeMem;
    void **i;
    size_t ma_objectSize_div_sizeofvoid;
    size_t sizeofBlock = ALIGNSIZE(sizeof(BLOCK));

    TEST_EXIT(ma)("ma=nil\n");
    ma_objectSize_div_sizeofvoid = ma->objectSize / sizeof(void*);

    if(!capacityIncrement) {
	WARNING("capacityIncrement == 0\n");
	return 1;
    }

    block = alberta_alloc(sizeofBlock + capacityIncrement * ma->objectSize, funcName, __FILE__, __LINE__);

    ma->capacity += capacityIncrement;
    block->next = ma->blocks;
    ma->blocks = block;
    freeMem = (FREEMEMORY*)((void**)block + sizeofBlock / sizeof(void*));
    block->end = ((void**)freeMem) + capacityIncrement * ma_objectSize_div_sizeofvoid;

    for(i = (void**)freeMem; i < (void**)(block->end) - ma_objectSize_div_sizeofvoid; i += ma_objectSize_div_sizeofvoid) {
	((FREEMEMORY*)i)->next = (FREEMEMORY*)(i + ma_objectSize_div_sizeofvoid);
    }
    ((FREEMEMORY*)i)->next = ma->freeMem;
    ma->freeMem = freeMem;

    return 0;
}

/*
 * newObject(..) returns a void-pointer to a memoryadmin for objects of size
 * objectSize. Use the returned value as parameter for getMemory.
 */
static void* newObject(size_t objectSize, unsigned int initialCapacity) {
    FUNCNAME("newObject");
    MEMORYADMIN *ma;

    ma = alberta_alloc(sizeof(MEMORYADMIN), funcName, __FILE__, __LINE__);

    ma->capacity = 0;
    ma->capacityIncrement = (initialCapacity ? initialCapacity : DEFAULTCAPACITYINCREMENT);
    if(!ma->capacityIncrement)
	WARNING("capacityIncrement == %u\n", ma->capacityIncrement);

    ma->objectSize = ALIGNSIZE(objectSize);
    ma->blocks = nil;
    ma->freeMem = nil;
    if(initialCapacity)
	newBlock(ma, initialCapacity);

    return ma;
}

/*
 * A memoryadmin allocates new memory, whenever getMemory is called and there
 * is no more free memory already allocated. The size of the new memory is
 * given by objectSize * capacityIncrement. By default capacityIncrement ==
 * DEFAULTCAPACITYINCREMENT
 */
static void setCapacityIncrement(void* memoryAdmin, unsigned int capacityIncrement) {
    FUNCNAME("setCapacityIncrement");

    TEST_EXIT(memoryAdmin)("memoryAdmin == nil\n");

    ((MEMORYADMIN*)memoryAdmin)->capacityIncrement = capacityIncrement;

    if(!((MEMORYADMIN*)memoryAdmin)->capacityIncrement) {
	WARNING("capacityIncrement == 0\n");
    }
    return;
}

/*
 * returns a pointer to an object of size objectsize given by creating the
 * memoryadmin.
 */
static void* getMemory(void* memoryAdmin) {
    FUNCNAME("getMemory");
    void* tmp;

    TEST_EXIT(memoryAdmin)("memoryAdmin==nil\n");

    if(((MEMORYADMIN*)memoryAdmin)->capacityIncrement == 0)
	WARNING("capacityIncrement == 0 => MAY BE AN ENDLESS LOOP\n");

    while(((MEMORYADMIN*)memoryAdmin)->capacity <= 0)
	if(newBlock(memoryAdmin, ((MEMORYADMIN*)memoryAdmin)->capacityIncrement) != 0) {
	    WARNING("newBlock failed\n");
	    return nil;
	}

    tmp = (void*)(((MEMORYADMIN*)memoryAdmin)->freeMem);

    ((MEMORYADMIN*)memoryAdmin)->freeMem = ((FREEMEMORY*)tmp)->next;

    ((MEMORYADMIN*)memoryAdmin)->capacity--;

    return tmp;
}

/*
 * returns memory at *object to memoryadmin. !!! object must be the result of
 * an call of getMemory(...) with the same memoryadmin. !!!
 */
static void freeMemory(void* object, void* memoryAdmin) {
    FUNCNAME("freeMemory");

    TEST_EXIT(memoryAdmin)("memoryAdmin=nil\n");
    ((FREEMEMORY*)object)->next = ((MEMORYADMIN*)memoryAdmin)->freeMem;
    ((MEMORYADMIN*)memoryAdmin)->freeMem = object;
    ((MEMORYADMIN*)memoryAdmin)->capacity++;
    return;
}

/*
 * the counterpart to newObject, frees all memory allocated with the given
 * memoryadmin.
 */
static void deleteObject(void* memoryAdmin) {
    FUNCNAME("deleteObject");
    BLOCK *b, *tmp;
    size_t sizeofBlock;

    TEST_EXIT(memoryAdmin)("memoryAdmin=nil\n");

    b = ((MEMORYADMIN*)memoryAdmin)->blocks;
    while(b != nil) {
	tmp = b->next;
        sizeofBlock = (unsigned int)b->end - (unsigned int)b;

	alberta_free(b, sizeofBlock);
	b = tmp;
    }
    alberta_free(memoryAdmin, sizeof(MEMORYADMIN));
    return;
}

typedef struct dof_admin_mem_info DOF_ADMIN_MEM_INFO;
struct dof_admin_mem_info
{
    void *dof_matrix;
    void *matrix_row;
    void *dof_int_vec;
    void *dof_dof_vec;
    void *int_dof_vec;
    void *dof_uchar_vec;
    void *dof_schar_vec;
    void *dof_real_vec;
    void *dof_real_d_vec;
};


/*--------------------------------------------------------------------------*/
/*  memory management for DOF admin structures                              */
/*--------------------------------------------------------------------------*/

const DOF_ADMIN *get_dof_admin(MESH *mesh, const char *name, 
	const int n_dof[DIM+1])
{
    FUNCNAME("get_dof_admin");
    DOF_ADMIN         *admin;
    int               i;
    DOF_ADMIN_MEM_INFO *mem_info;
    extern void add_dof_admin_to_mesh(DOF_ADMIN *, MESH *);


    admin = MEM_CALLOC(1, DOF_ADMIN);
    admin->mesh = mesh;
    admin->name = name ? strdup(name) : nil;

    admin->dof_free         = nil;
    admin->dof_free_size    = admin->first_hole     = 0;

    for (i = 0; i < DIM+1; i++)
	admin->n_dof[i] = n_dof[i];

    mem_info = admin->mem_info = (void*)(MEM_ALLOC(1, DOF_ADMIN_MEM_INFO));

    mem_info->dof_matrix = newObject(sizeof(DOF_MATRIX), 0); 
    setCapacityIncrement(mem_info->dof_matrix, 10);
    mem_info->matrix_row = newObject(sizeof(MATRIX_ROW), 0);
    setCapacityIncrement(mem_info->matrix_row, 100);
    mem_info->dof_int_vec = newObject(sizeof(DOF_INT_VEC), 0); 
    setCapacityIncrement(mem_info->dof_int_vec, 10);
    mem_info->dof_dof_vec = newObject(sizeof(DOF_DOF_VEC), 0); 
    setCapacityIncrement(mem_info->dof_dof_vec, 10);
    mem_info->int_dof_vec = newObject(sizeof(DOF_DOF_VEC), 0);
    setCapacityIncrement(mem_info->int_dof_vec, 10); 
    mem_info->dof_uchar_vec = newObject(sizeof(DOF_UCHAR_VEC), 0); 
    setCapacityIncrement(mem_info->dof_uchar_vec, 10);
    mem_info->dof_schar_vec = newObject(sizeof(DOF_SCHAR_VEC), 0); 
    setCapacityIncrement(mem_info->dof_schar_vec, 10);
    mem_info->dof_real_vec = newObject(sizeof(DOF_REAL_VEC), 0); 
    setCapacityIncrement(mem_info->dof_real_vec, 10);
    mem_info->dof_real_d_vec = newObject(sizeof(DOF_REAL_D_VEC), 0); 
    setCapacityIncrement(mem_info->dof_real_d_vec, 10);

    add_dof_admin_to_mesh(admin, mesh);

    return(admin);
}


/*--------------------------------------------------------------------------*/
/*  Free all DOF_ADMINs in a rather brutal way, only makes sense when       */
/*  freeing an entire mesh.                                                 */
/*--------------------------------------------------------------------------*/

static void free_dof_admins(MESH *mesh)
{
  FUNCNAME("free_dof_admins");

  DOF_ADMIN **admin = mesh->dof_admin;
  int i, n;
  DOF_MATRIX     *dof_matrix,    *dof_matrix_next;
  DOF_INT_VEC    *dof_int_vec,   *dof_int_vec_next;
  DOF_DOF_VEC    *dof_dof_vec,   *dof_dof_vec_next;
  DOF_DOF_VEC    *int_dof_vec,   *int_dof_vec_next;
  DOF_UCHAR_VEC  *dof_uchar_vec, *dof_uchar_vec_next;
  DOF_SCHAR_VEC  *dof_schar_vec, *dof_schar_vec_next;
  DOF_REAL_VEC   *dof_real_vec,  *dof_real_vec_next;
  DOF_REAL_D_VEC *dof_real_d_vec,*dof_real_d_vec_next;

  n = mesh->n_dof_admin;
  if ((n > 0) && !admin)
    ERROR_EXIT("no mesh->dof_admin but n_dof_admin=%d\n", n);
  if ((n <= 0) && admin)
    ERROR_EXIT("found mesh->dof_admin but n_dof_admin=%d\n", n);

  for(i = 0; i < n; i++) {
    for(dof_matrix = admin[i]->dof_matrix; dof_matrix; dof_matrix = dof_matrix_next) {
      dof_matrix_next = dof_matrix->next;
      free_dof_matrix(dof_matrix);
    }
    for(dof_int_vec = admin[i]->dof_int_vec; dof_int_vec; dof_int_vec = dof_int_vec_next) {
      dof_int_vec_next = dof_int_vec->next;
      free_dof_int_vec(dof_int_vec);
    }
    for(dof_dof_vec = admin[i]->dof_dof_vec; dof_dof_vec; dof_dof_vec = dof_dof_vec_next) {
      dof_dof_vec_next = dof_dof_vec->next;
      free_dof_dof_vec(dof_dof_vec);
    }
    for(int_dof_vec = admin[i]->int_dof_vec; int_dof_vec; int_dof_vec = int_dof_vec_next) {
      int_dof_vec_next = int_dof_vec->next;
      free_int_dof_vec(int_dof_vec);
    }
    for(dof_uchar_vec = admin[i]->dof_uchar_vec; dof_uchar_vec; dof_uchar_vec = dof_uchar_vec_next) {
      dof_uchar_vec_next = dof_uchar_vec->next;
      free_dof_uchar_vec(dof_uchar_vec);
    }
    for(dof_schar_vec = admin[i]->dof_schar_vec; dof_schar_vec; dof_schar_vec = dof_schar_vec_next) {
      dof_schar_vec_next = dof_schar_vec->next;
      free_dof_schar_vec(dof_schar_vec);
    }
    for(dof_real_vec = admin[i]->dof_real_vec; dof_real_vec; dof_real_vec = dof_real_vec_next) {
      dof_real_vec_next = dof_real_vec->next;
      free_dof_real_vec(dof_real_vec);
    }
    for(dof_real_d_vec = admin[i]->dof_real_d_vec; dof_real_d_vec; dof_real_d_vec = dof_real_d_vec_next) {
      dof_real_d_vec_next = dof_real_d_vec->next;
      free_dof_real_d_vec(dof_real_d_vec);
    }

    deleteObject(((DOF_ADMIN_MEM_INFO *)admin[i]->mem_info)->dof_matrix);
    deleteObject(((DOF_ADMIN_MEM_INFO *)admin[i]->mem_info)->matrix_row);
    deleteObject(((DOF_ADMIN_MEM_INFO *)admin[i]->mem_info)->dof_int_vec);
    deleteObject(((DOF_ADMIN_MEM_INFO *)admin[i]->mem_info)->dof_dof_vec);
    deleteObject(((DOF_ADMIN_MEM_INFO *)admin[i]->mem_info)->int_dof_vec);
    deleteObject(((DOF_ADMIN_MEM_INFO *)admin[i]->mem_info)->dof_uchar_vec);
    deleteObject(((DOF_ADMIN_MEM_INFO *)admin[i]->mem_info)->dof_schar_vec);
    deleteObject(((DOF_ADMIN_MEM_INFO *)admin[i]->mem_info)->dof_real_vec);
    deleteObject(((DOF_ADMIN_MEM_INFO *)admin[i]->mem_info)->dof_real_d_vec);

    MEM_FREE(admin[i]->mem_info, 1, DOF_ADMIN_MEM_INFO);
    MEM_FREE(admin[i]->dof_free, admin[i]->dof_free_size, DOF_FREE_UNIT);
  }

  return;
}


/*--------------------------------------------------------------------------*/
/*  allocate memory for a new mesh, and build a simply linked list of these */
/*--------------------------------------------------------------------------*/

typedef struct mesh_info MESH_INFO;

struct mesh_info
{
    MESH       *mesh;
    MESH_INFO  *next;
};

static MESH_INFO  *first_mesh = nil;


MESH *get_mesh(const char *name, void (*init_dof_admins)(MESH *),
	       void (*init_leaf_data)(LEAF_DATA_INFO *))
{
  FUNCNAME("get_mesh");
  MESH_INFO       *new_mesh;
  MESH            *mesh;
  int i;
  MESH_MEM_INFO   *mem_info;

  new_mesh = MEM_CALLOC(1,MESH_INFO);
  mesh =     MEM_CALLOC(1,MESH);
  
  new_mesh->mesh = mesh;
  new_mesh->next = first_mesh;
  first_mesh = new_mesh;
  
  mesh->name = name ? strdup(name) : nil;
  
  if (init_dof_admins)
    init_dof_admins(mesh);

  if (mesh->n_dof[VERTEX] == 0)
  {
    int n_dof[DIM+1] = { 0 };

    WARNING("Mesh '%s': no DOF_ADMIN using vertices given; adding one.\n", mesh->name);

    n_dof[VERTEX] = 1;
    get_dof_admin(mesh, "vertex dofs", n_dof);
  }

  mem_info = mesh->mem_info = (void*)(MEM_ALLOC(1, MESH_MEM_INFO));
  
  mem_info->dof_ptrs = newObject(mesh->n_node_el * sizeof(DOF*), 1000); 
  
  for(i = 0; i < DIM + 1; i++)
    mem_info->dofs[i] = nil;
  
  mem_info->element = newObject(sizeof(EL), 0); 
#if DIM == 3
  mem_info->rc_list = nil; 
#endif
  mem_info->real_d = newObject(sizeof(REAL_D), 0);

  mem_info->leaf_data = nil;
  if (init_leaf_data)
  {
    unsigned int  n;
    
    (*init_leaf_data)(mesh->leaf_data_info);
    n = mesh->leaf_data_info->leaf_data_size;
    
    if (n > 0)
    {
      if (n != ALIGNSIZE(n))
	WARNING("installing leafdata of size %d with aligned size %d\n", 
		n, ALIGNSIZE(n));
      mem_info->leaf_data = newObject(ALIGNSIZE(n), 0);
    }
  }
  
  return(mesh);
}

void free_mesh(MESH *mesh)
{
    FUNCNAME("free_mesh");
    MESH_INFO     *mesh_info, **prev;
    MESH_MEM_INFO *mem_info = mesh->mem_info;
    int i;

    prev = &first_mesh;
    for (mesh_info = first_mesh; mesh_info; mesh_info = mesh_info->next)
    {
      if (mesh_info->mesh == mesh)
      {
	*prev = mesh_info->next;
	MEM_FREE(mesh_info, 1, MESH_INFO);

	/* free all elements/dofs/... */
	deleteObject(mem_info->dof_ptrs);

	for(i = 0; i < DIM+1; i++)
	{
	  if(mem_info->dofs[i])  deleteObject(mem_info->dofs[i]);
	}

	deleteObject(mem_info->element);
#if DIM == 3
	deleteObject(mem_info->rc_list);
#endif
	deleteObject(mem_info->real_d);

	if (mem_info->leaf_data)
	  deleteObject(mem_info->leaf_data);

	MEM_FREE(mem_info->coords, mem_info->count, REAL_D);
	MEM_FREE(mem_info, 1, MESH_MEM_INFO);

	free_dof_admins(mesh);

	MEM_FREE(mesh->dof_admin, mesh->n_dof_admin, DOF_ADMIN);
	MEM_FREE(mesh, 1, MESH);
	break;
      }
      prev = &(mesh_info->next);
    }
    TEST(mesh_info)("mesh not found; mesh not freed\n");

    return;
}


/*--------------------------------------------------------------------------*/
/*  memory management for FE_SPACE structures                               */
/*--------------------------------------------------------------------------*/

const FE_SPACE *get_fe_space(MESH *mesh, const char *name,
	const int n_dof[DIM+1],
	const BAS_FCTS *bas_fcts)
{
    FUNCNAME("get_fe_space");
    const DOF_ADMIN   *admin = nil;
    FE_SPACE          *fe_space;
    int               i, j;
    const int         *ndof;

    extern void add_dof_admin_to_mesh(DOF_ADMIN *, MESH *);

    fe_space = MEM_ALLOC(1, FE_SPACE);
    fe_space->name = name ? strdup(name) : nil;

    if (bas_fcts)  ndof = bas_fcts->n_dof;
    else           ndof = n_dof;
    TEST_EXIT(ndof)("no n_dof or bas_fcts->n_dof\n");

    for (i = 0; i < mesh->n_dof_admin; i++) {
	admin = mesh->dof_admin[i];
	for (j = 0; j <= DIM; j++) {
	    if (admin->n_dof[j] != ndof[j]) break;
	}
	if (j > DIM) break;
	admin = nil;
    }

    if (!admin) {
	admin = get_dof_admin(mesh, name, ndof);
    }

    fe_space->admin = admin;
    fe_space->bas_fcts = bas_fcts;
    fe_space->mesh = mesh;

    TEST_EXIT((bas_fcts==nil) || (bas_fcts->init_element == 0))
	("Use of bas_fcts->init_element is NOT IMPLEMENTED YET !!!\n");

    return(fe_space);
}


/*--------------------------------------------------------------------------*/
/*  memory management for DOF pointers                                      */
/*--------------------------------------------------------------------------*/

/* #define DOF_PTRS_BLOCK  1000 */

#if DIM == 2
#define MAX_DOF_PTRS       7
#else
#define MAX_DOF_PTRS      15
#endif

static DOF **get_dof_ptrs(MESH *mesh)
{
    FUNCNAME("get_dof_ptrs");
    int i, n;
    DOF **ptrs;
    MESH_MEM_INFO *mem_info;

    TEST_EXIT(mesh)("mesh=nil\n");
    TEST_EXIT(mesh->mem_info)("mesh \"%s\": mesh->mem_info=nil\n", mesh->name);
    mem_info = (MESH_MEM_INFO*)(mesh->mem_info);
    n = mesh->n_node_el;
    if (n <= 0)
	return(nil);

    TEST_EXIT(n <= MAX_DOF_PTRS)("mesh \"%s\": too many nodes: %d > %d\n", mesh->name, n, MAX_DOF_PTRS);
    TEST_EXIT(mem_info->dof_ptrs)("mesh \"%s\": mesh->mem_info->dof_ptrs=nil\n", mesh->name);

    ptrs = (DOF **)getMemory(mem_info->dof_ptrs);
    for (i = 0; i < n; i++)
	ptrs[i] = nil;

    return(ptrs);
}


static void free_dof_ptrs(DOF **ptrs, MESH *mesh)
{
    FUNCNAME("free_dof_ptrs");
    int      n;
    MESH_MEM_INFO *mem_info;

    TEST_EXIT(ptrs)("ptrs=nil\n");
    TEST_EXIT(mesh)("mesh=nil\n");
    TEST_EXIT(mesh->mem_info)("mesh \"%s\": mesh->mem_info=nil\n", mesh->name);
    n = mesh->n_node_el;
    if (n <= 0)
	return;

    TEST_EXIT(n <= MAX_DOF_PTRS)("mesh \"%s\": too many nodes: %d > %d\n", mesh->name, n, MAX_DOF_PTRS);

    mem_info = (MESH_MEM_INFO*)(mesh->mem_info);

    TEST_EXIT(mem_info->dof_ptrs)("mesh \"%s\": mesh->mem_info->dof_ptrs=nil\n", mesh->name);

    freeMemory((void *)ptrs, mem_info->dof_ptrs);
    return;
}


/*--------------------------------------------------------------------------*/
/*  memory management for DOFs                                              */
/*  WARNING: This routine will not tolerate any changes to mesh->ndof!      */
/*  Any change in the number of DOF_ADMINS should therefore be              */
/*  completed before accessing the mesh.                                    */
/*--------------------------------------------------------------------------*/

#define DOF_BLOCK 1000

/* #define MAX_DOF    100 */          /* This limit is no longer necessary! */

DOF *get_dof(MESH *mesh, int position)
{
    FUNCNAME("get_dof");
    DOF_ADMIN *admin;
    DOF     *dof;
    int     i, j, n, n0, ndof;
    MESH_MEM_INFO* mem_info;


    TEST_EXIT(mesh)("mesh=nil\n");
    mem_info = mesh->mem_info;
    TEST_EXIT(mem_info)("mesh \"%s\": mesh->mem_info=nil\n", mesh->name);

    TEST_EXIT(position >= VERTEX && position <= CENTER)
	("mesh \"%s\": unknown position %d\n", mesh->name, position);

    ndof = mesh->n_dof[position];
    if (ndof <= 0) return(nil);

/*     TEST_EXIT(ndof <= MAX_DOF) */
/* 	("ndof too big: ndof=%d, MAX_DOF=%d\n",ndof,MAX_DOF); */

    if (!(mem_info->dofs[position])) {
    /*XXX*/
      mem_info->dofs[position] = newObject(sizeof(DOF)*ndof, DOF_BLOCK);

      /*
   	dof_block_infos[ndof] = MEM_ALLOC(1, MEM_BLOCK_INFO);
	dof_block_infos[ndof]->free       = nil;
	dof_block_infos[ndof]->info       = nil;
	dof_block_infos[ndof]->free_count = 0;
	dof_block_infos[ndof]->used_count = 0;
	dof_block_infos[ndof]->unit_size  = sizeof(DOF)*ndof;
	dof_block_infos[ndof]->block_size = DOF_BLOCK;
	*/
    }
    else
      TEST_EXIT(((MEMORYADMIN *)mem_info->dofs[position])->objectSize >= sizeof(DOF)*ndof)
("mesh \"%s\": ndofs has changed for position %d\n", mesh->name, position);

    dof = (DOF *)getMemory(mem_info->dofs[position]);

    for (i = 0; i < mesh->n_dof_admin; i++)
    {
	admin = mesh->dof_admin[i];
	TEST_EXIT(admin)("mesh \"%s\": no dof_admin[%d]\n", mesh->name, i);

	n  = admin->n_dof[position];
	n0 = admin->n0_dof[position];

	TEST_EXIT(n+n0 <= ndof)("dof_admin \"%s\": n=%d, n0=%d too large: ndof=%d\n", admin->name, n, n0, ndof);

	for (j = 0; j < n; j++)
	{
	    dof[n0+j] = get_dof_index(admin);
	}
    }

    return(dof);
}

void free_dof(DOF *dof, MESH *mesh, int position)
{
    FUNCNAME("free_dof");

    DOF_ADMIN *admin;
    int     i, j, n, n0, ndof;
    MESH_MEM_INFO* mem_info;

    TEST_EXIT(mesh)("mesh=nil\n");
    mem_info = (MESH_MEM_INFO*)mesh->mem_info;
    TEST_EXIT(mem_info)("mesh \"%s\": mesh->mem_info=nil\n", mesh->name);

    TEST_EXIT(position >= VERTEX && position <= CENTER)
	("mesh \"%s\": unknown position %d\n",mesh->name, position);

    ndof = mesh->n_dof[position];
    if (ndof) 
    {
	if (!dof)
	{
	    MSG("dof = nil, but ndof=%d\n", ndof);
	    return;
	}
    }
    else
    {
	if (dof)
	{
	    MSG("dof != nil, but ndof=0\n");
	}
	return;
    }

/*     TEST_EXIT(ndof <= MAX_DOF) */
/* 	("ndof too big: ndof=%d, MAX_DOF=%d\n",ndof,MAX_DOF); */

/*     TEST_EXIT(mem_info->dofs[ndof]) */
/*       ("no dof_block found for ndof=%d\n",ndof); */

    TEST_EXIT(mem_info->dofs[position]) 
      ("mesh \"%s\": no memory management present for %d DOFs.", mesh->name, position);

    for (i = 0; i < mesh->n_dof_admin; i++)
    {
	admin = mesh->dof_admin[i];
	TEST_EXIT(admin)("mesh \"%s\": no dof_admin[%d]\n", mesh->name, i);

	n  = admin->n_dof[position];
	n0 = admin->n0_dof[position];

	TEST_EXIT(n+n0 <= ndof)("dof_admin \"%s\": n=%d, n0=%d too large: ndof=%d\n", admin, n, n0, ndof);

	for (j = 0; j < n; j++)
	{
	    free_dof_index(admin, dof[n0+j]);
	}
    }

    freeMemory((void *)dof, mem_info->dofs[position]);
    return;
}

/*--------------------------------------------------------------------------*/
/*  memory management for leaf_data                                         */
/*--------------------------------------------------------------------------*/

#define LEAF_DATA_BLOCK  1000

#if 0
void *get_leaf_data(MESH *mesh)
{
    FUNCNAME("get_leaf_data");
    int      n;
    void     *leaf_data;
    MEMORYADMIN  *ldbi;

    TEST_EXIT(mesh)("mesh=nil\n");

    /* Adjust leaf_data_size to a multiple of sizeof(void*). */
    /* Note that this change is visible to users!            */

    mesh->leaf_data_info->leaf_data_size = 
      ALIGNSIZE(mesh->leaf_data_info->leaf_data_size);

    n = mesh->leaf_data_info->leaf_data_size;
    if (n <= 0)
	return(nil);

/*  ldbi = (MEMORYADMIN *) mesh->leaf_data_info->leaf_data_block_info;*/
    ldbi = (MEMORYADMIN *) ((MESH_MEM_INFO*)(mesh->mem_info))->leaf_data;

    if (!ldbi)
    {
	ldbi = newObject(n, 0);
/* 	mesh->leaf_data_info->leaf_data_block_info = (const void *) ldbi; */
	((MESH_MEM_INFO*)(mesh->mem_info))->leaf_data = ldbi;
    }

    leaf_data = getMemory(ldbi);

    return(leaf_data);
}

void free_leaf_data(void *leaf_data, MESH *mesh)
{
    FUNCNAME("free_leaf_data");
    unsigned int n;
    MEMORYADMIN  *ldbi;

    if (!leaf_data)  return;

    TEST_EXIT(mesh)("mesh = nil\n");
    n = mesh->leaf_data_info->leaf_data_size;
/*     ldbi = (MEMORYADMIN *)(mesh->leaf_data_info->leaf_data_block_info); */
    ldbi = (MEMORYADMIN *) ((MESH_MEM_INFO*)(mesh->mem_info))->leaf_data;
    if (!ldbi  || n <= 0) return;

    TEST_EXIT(n == ((MEMORYADMIN*)ldbi)->objectSize)
	("mesh \"%s\": size %d of leaf data has changed; old size = %d\n", mesh->name, n, (MEMORYADMIN*)ldbi->objectSize);

    freeMemory((void *)leaf_data, ldbi);
    return;
}
#else
void *get_leaf_data(MESH *mesh)
{
  FUNCNAME("get_leaf_data");
  MEMORYADMIN  *ldbi;

  TEST_EXIT(mesh)("pointer to mesh = nil\n");

  ldbi = ((MESH_MEM_INFO*)(mesh->mem_info))->leaf_data;

  if (ldbi)
    return(getMemory(ldbi));
  else
    return(nil);
}

void free_leaf_data(void *leaf_data, MESH *mesh)
{
    FUNCNAME("free_leaf_data");
    MEMORYADMIN  *ldbi;

    if (!leaf_data)  return;

    TEST_EXIT(mesh)("pointer to mesh = nil\n");
    ldbi = (MEMORYADMIN *) ((MESH_MEM_INFO*)(mesh->mem_info))->leaf_data;

    if (!ldbi) return;

    freeMemory((void *)leaf_data, ldbi);
    return;
}
#endif

#if 0  /*  not used at the moment              */
/*--------------------------------------------------------------------------*/
/*  memory management for leaf_data                                         */
/*--------------------------------------------------------------------------*/

#define QUAD_VEC_BLOCK  100
#define MAX_QUAD_INFO     3

static MEM_BLOCK_INFO *quad_vec_block_infos[MAX_QUAD_INFO+1];

REAL *get_quad_vec(int dim)
{
    FUNCNAME("get_quad_vec");
    int             i, n;
    REAL            *quad_vec;
    MEM_BLOCK_INFO  *gvbi;

    TEST_EXIT(dim >= 1  &&  dim  <= 3)("dim = %d not allowed\n");

    gvbi = quad_vec_block_infos[dim];

    if (!gvbi)
    {
	gvbi = MEM_ALLOC(1, MEM_BLOCK_INFO);
	gvbi->free       = nil;
	gvbi->info       = nil;
	gvbi->free_count = 0;
	gvbi->used_count = 0;
	gvbi->unit_size  = get_max_no_quad_point(dim)*sizeof(REAL);
	gvbi->block_size = QUAD_VEC_BLOCK;

	quad_vec_block_infos[dim] = gvbi;
    }

    quad_vec = (REAL *) get_mem(gvbi);

    return(quad_vec);
}

void free_quad_vec(REAL *quad_vec, int dim)
{
    FUNCNAME("free_quad_vec");
    int             n;
    MEM_BLOCK_INFO  *gvbi;

    if (!quad_vec)  return;

    TEST_EXIT(dim >= 1  &&  dim  <= 3)("dim = %d not allowed\n");

    gvbi = quad_vec_block_infos[dim];

    if (!gvbi) return;

    free_mem((void *) quad_vec, gvbi);
    return;
}

static MEM_BLOCK_INFO *quad_vec_d_block_infos[MAX_QUAD_INFO+1];

REAL_D *get_quad_vec_d(int dim)
{
    FUNCNAME("get_quad_vec_d");
    int      i, n;
    REAL_D          *quad_vec_d;
    MEM_BLOCK_INFO  *gvdbi;

    TEST_EXIT(dim >= 1  &&  dim  <= 3)("dim = %d not allowed\n");

    gvdbi = quad_vec_d_block_infos[dim];

    if (!gvdbi)
    {
	gvdbi = MEM_ALLOC(1, MEM_BLOCK_INFO);
	gvdbi->free       = nil;
	gvdbi->info       = nil;
	gvdbi->free_count = 0;
	gvdbi->used_count = 0;
	gvdbi->unit_size  = get_max_no_quad_point(dim)*sizeof(REAL_D);
	gvdbi->block_size = QUAD_VEC_BLOCK;

	quad_vec_d_block_infos[dim] = gvdbi;
    }

    quad_vec_d = (REAL_D *) get_mem(gvdbi);

    return(quad_vec_d);
}

void free_quad_vec_d(REAL_D *quad_vec_d, int dim)
{
    FUNCNAME("free_quad_vec_d");
    int             n;
    MEM_BLOCK_INFO  *gvdbi;

    if (!quad_vec_d)  return;

    TEST_EXIT(dim >= 1  &&  dim  <= 3)("dim = %d not allowed\n");

    gvdbi = quad_vec_d_block_infos[dim];

    if (!gvdbi) return;

    free_mem((void *) quad_vec_d, gvdbi);
    return;
}
#endif

/*--------------------------------------------------------------------------*/
/*  memory management for elements                                          */
/*--------------------------------------------------------------------------*/

#define EL_BLOCK   1000

#if EL_INDEX
static int el_index = 0;
#endif

EL *get_element(MESH *mesh) 
{
    FUNCNAME("get_element");
    EL *el;

    TEST_EXIT(mesh)("mesh == nil\n");
    TEST_EXIT(mesh->mem_info)("mesh \"%s\": no memory management present.\n", mesh->name);
    el = (EL *)getMemory(((MESH_MEM_INFO*)(mesh->mem_info))->element);
    el->child[0]  = nil;
    el->child[1]  = (EL *) get_leaf_data(mesh);
    el->dof       = get_dof_ptrs(mesh);
#if EL_INDEX
    el->index = el_index++;
#endif

#if DIM > 1
    el->new_coord = nil;
#endif
    return el;
}

void free_element(EL *el, MESH *mesh)
{
    free_dof_ptrs(el->dof, mesh);
#if DIM > 1
    if (el->new_coord) {
	free_real_d(mesh, el->new_coord);
	el->new_coord = nil;
    }
#endif
    if (el->child[1]) free_leaf_data((void *) el->child[1], mesh);

    freeMemory(el, ((MESH_MEM_INFO*)(mesh->mem_info))->element);
}


#if DIM == 3
/*--------------------------------------------------------------------------*/
/*  some routines for allocation and deallocation of list for collecting    */
/*  neighbours at the refinement edge in 3 dimensions                       */
/*--------------------------------------------------------------------------*/

#define LIST_BLOCK 20

/* static MEM_BLOCK_INFO rc_list_block_info = {nil, nil, 0, 0,  0, LIST_BLOCK}; */

RC_LIST_EL *get_rc_list(MESH *mesh)
{
    FUNCNAME("get_rc_list");

    if (!((MESH_MEM_INFO*)mesh->mem_info)->rc_list)
    {
        ((MESH_MEM_INFO*)mesh->mem_info)->rc_list = 
        newObject(mesh->max_edge_neigh*sizeof(RC_LIST_EL), LIST_BLOCK);
    }
    else
    {
	TEST_EXIT(((MEMORYADMIN *)((MESH_MEM_INFO*)mesh->mem_info)->rc_list)->objectSize == (mesh->max_edge_neigh * sizeof(RC_LIST_EL)))
("mesh \"%s\": mesh->max_edge_neigh changed\n", mesh->name);
    }

    return (RC_LIST_EL *)getMemory(((MESH_MEM_INFO*)mesh->mem_info)->rc_list);
}


void free_rc_list(MESH *mesh, RC_LIST_EL *list)
{
    freeMemory((void *)list, ((MESH_MEM_INFO *)(mesh->mem_info))->rc_list);
}

#endif

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

REAL *get_real_d(MESH *mesh)
{
    FUNCNAME("get_real_d");

    TEST_EXIT(mesh)("mesh==nil\n");
    return (REAL *)getMemory(((MESH_MEM_INFO*)(mesh->mem_info))->real_d);
}

void free_real_d(MESH *mesh, REAL *ptr)
{
    FUNCNAME("free_real_d");

    TEST_EXIT(mesh)("mesh==nil\n");
    freeMemory((void *)ptr, ((MESH_MEM_INFO*)(mesh->mem_info))->real_d);
    return;
}

/*--------------------------------------------------------------------------*/
/*  memory management for matrix rows                                       */
/*  matrix rows not connected to a FE_SPACE, e.g. in multigrid.c are        */
/*  treated specially. Others are fixed to a FE_SPACE and thus to a unique  */
/*  DOF_ADMIN and MESH.                                                     */
/*--------------------------------------------------------------------------*/

static void *unconnected_rows = nil;

MATRIX_ROW  *get_matrix_row(const FE_SPACE *fe_space)
{
    MATRIX_ROW *row;
    static void *matrix_row_info = nil;
    int        j;

    if(!fe_space || !fe_space->admin) {
      if(!unconnected_rows)
	unconnected_rows = newObject(sizeof(MATRIX_ROW), 100); 
      matrix_row_info = unconnected_rows;
    }
    else
      matrix_row_info = ((DOF_ADMIN_MEM_INFO *)fe_space->admin->mem_info)->matrix_row;

    row = (MATRIX_ROW *)getMemory(matrix_row_info);
    row->next = nil;
    for (j=0; j<ROW_LENGTH; j++) 
	row->col[j] = NO_MORE_ENTRIES;

    return(row);
}

void free_matrix_row(const FE_SPACE *fe_space, MATRIX_ROW *row)
{
  static void *matrix_row_info = nil;

  if(!fe_space || !fe_space->admin)
    matrix_row_info = unconnected_rows;
  else
    matrix_row_info = ((DOF_ADMIN_MEM_INFO *)fe_space->admin->mem_info)->matrix_row;

  freeMemory((void *)row, matrix_row_info);
}

ALBERTA_DEFUNUSED(static void delete_unconnected_rows(void))
{
  if(unconnected_rows) deleteObject(unconnected_rows);
  unconnected_rows = nil;
}

/*--------------------------------------------------------------------------*/
/*  memory management for matrices                                          */
/*  unconnected_matrices: see unconnected rows above...                     */
/*--------------------------------------------------------------------------*/

static void *unconnected_matrices = nil;

DOF_MATRIX  *get_dof_matrix(const char *name, const FE_SPACE *fe_space)
{
    DOF_MATRIX *mat;
    static void *matrix_info = nil;
    void  add_dof_matrix_to_admin(DOF_MATRIX *, DOF_ADMIN *);

    if(!fe_space || !fe_space->admin) {
      if(!unconnected_matrices)
	unconnected_matrices = newObject(sizeof(DOF_MATRIX), 10); 
      matrix_info = unconnected_matrices;
    }
    else
      matrix_info = ((DOF_ADMIN_MEM_INFO *)fe_space->admin->mem_info)->dof_matrix;

    mat = (DOF_MATRIX *)getMemory(matrix_info);

    mat->next   = nil;
    mat->fe_space = fe_space;
    mat->name   = name ? strdup(name) : nil;
    mat->matrix_row = nil;
    mat->size   = 0;

    mat->refine_interpol = mat->coarse_restrict = nil;

    if (fe_space && fe_space->admin)
	add_dof_matrix_to_admin(mat, (DOF_ADMIN *)fe_space->admin);

    return(mat);
}

void free_dof_matrix(DOF_MATRIX *mat)
{
  void remove_dof_matrix_from_admin(DOF_MATRIX *);

  if (mat->fe_space && mat->fe_space->admin)
    remove_dof_matrix_from_admin(mat);
  
  clear_dof_matrix(mat);  /* free all matrix_rows */
  

  MEM_FREE(mat->matrix_row, mat->size, MATRIX_ROW *);
  mat->matrix_row = nil;
  mat->size = 0;

  if(mat->fe_space && mat->fe_space->admin)
    freeMemory((void *)mat, ((DOF_ADMIN_MEM_INFO*)(mat->fe_space->admin->mem_info))->dof_matrix);
  else
    freeMemory((void *)mat, unconnected_matrices);
}

ALBERTA_DEFUNUSED(static void delete_unconnected_matrices(void))
{
  if(unconnected_matrices) deleteObject(unconnected_matrices);
  unconnected_matrices = nil;
}


/*--------------------------------------------------------------------------*/
/*  memory management for dof_vectors                                       */
/*  unconnected dof_vectors are handled as above                            */
/*--------------------------------------------------------------------------*/

static void *unconnected_dof_int_vecs = nil;

DOF_INT_VEC  *get_dof_int_vec(const char *name, const FE_SPACE *fe_space)
{
    void add_dof_int_vec_to_admin(DOF_INT_VEC *, DOF_ADMIN *);
    DOF_INT_VEC *vec;
    static void *vec_info = nil;


    if(!fe_space || !fe_space->admin) {
      if(!unconnected_dof_int_vecs)
	unconnected_dof_int_vecs = newObject(sizeof(DOF_INT_VEC), 10);
      vec_info = unconnected_dof_int_vecs;
    }
    else
      vec_info = ((DOF_ADMIN_MEM_INFO*)(fe_space->admin->mem_info))->dof_int_vec;

    vec = (DOF_INT_VEC *)getMemory(vec_info);

    vec->next  = nil;
    vec->fe_space = fe_space;
    vec->name  = name ? strdup(name): nil;
    vec->size  = 0;
    vec->vec   = nil;
    vec->refine_interpol = vec->coarse_restrict = nil;

    if (fe_space && fe_space->admin)
	add_dof_int_vec_to_admin(vec, (DOF_ADMIN *)fe_space->admin);

    return(vec);
}

void free_dof_int_vec(DOF_INT_VEC *vec)
{
  void remove_dof_int_vec_from_admin(DOF_INT_VEC *);
  
  if (vec->fe_space && vec->fe_space->admin)
    remove_dof_int_vec_from_admin(vec);
  
  MEM_FREE(vec->vec, vec->size, int);
  vec->vec = nil;
  vec->size = 0;
  
  if(vec->fe_space && vec->fe_space->admin)
    freeMemory((void *)vec, ((DOF_ADMIN_MEM_INFO*)(vec->fe_space->admin->mem_info))->dof_int_vec);
  else
    freeMemory((void *)vec, unconnected_dof_int_vecs);
}

ALBERTA_DEFUNUSED(static void delete_unconnected_dof_int_vecs(void))
{
  if(unconnected_dof_int_vecs) deleteObject(unconnected_dof_int_vecs);
  unconnected_dof_int_vecs = nil;
}

static void *unconnected_dof_dof_vecs = nil;

DOF_DOF_VEC  *get_dof_dof_vec(const char *name, const FE_SPACE *fe_space)
{
    void add_dof_dof_vec_to_admin(DOF_DOF_VEC *, DOF_ADMIN *);
    DOF_DOF_VEC *vec;
    static void *vec_info = nil;


    if(!fe_space || !fe_space->admin) {
      if(!unconnected_dof_dof_vecs)
	unconnected_dof_dof_vecs = newObject(sizeof(DOF_DOF_VEC), 10);
      vec_info = unconnected_dof_dof_vecs;
    }
    else
      vec_info = ((DOF_ADMIN_MEM_INFO*)(fe_space->admin->mem_info))->dof_dof_vec;

    vec = (DOF_DOF_VEC *)getMemory(vec_info);

    vec->next  = nil;
    vec->fe_space = fe_space;
    vec->name  = name ? strdup(name) : nil;
    vec->size  = 0;
    vec->vec   = nil;
    vec->refine_interpol = vec->coarse_restrict = nil;

    if (fe_space && fe_space->admin)
	add_dof_dof_vec_to_admin(vec, (DOF_ADMIN *)fe_space->admin);

    return(vec);
}

void free_dof_dof_vec(DOF_DOF_VEC *vec)
{
  void remove_dof_dof_vec_from_admin(DOF_DOF_VEC *);
  
  if (vec->fe_space && vec->fe_space->admin)
    remove_dof_dof_vec_from_admin(vec);
  
  MEM_FREE(vec->vec, vec->size, int);
  vec->vec = nil;
  vec->size = 0;
  
  if(vec->fe_space && vec->fe_space->admin)
    freeMemory((void *)vec, ((DOF_ADMIN_MEM_INFO*)(vec->fe_space->admin->mem_info))->dof_dof_vec);
  else
    freeMemory((void *)vec, unconnected_dof_dof_vecs);
}

ALBERTA_DEFUNUSED(static void delete_unconnected_dof_dof_vecs(void))
{
  if(unconnected_dof_dof_vecs) deleteObject(unconnected_dof_dof_vecs);
  unconnected_dof_dof_vecs = nil;
}


static void *unconnected_int_dof_vecs = nil;

DOF_DOF_VEC  *get_int_dof_vec(const char *name, const FE_SPACE *fe_space)
{
  DOF_DOF_VEC *vec;
  void add_int_dof_vec_to_admin(DOF_DOF_VEC *, DOF_ADMIN *);
  static void *vec_info = nil;
  
  
  if(!fe_space || !fe_space->admin) {
    if(!unconnected_int_dof_vecs)
      unconnected_int_dof_vecs = newObject(sizeof(DOF_DOF_VEC), 10);
    vec_info = unconnected_int_dof_vecs;
  }
  else
    vec_info = ((DOF_ADMIN_MEM_INFO*)(fe_space->admin->mem_info))->int_dof_vec;
  
  vec = (DOF_DOF_VEC *)getMemory(vec_info);
  
  vec->next  = nil;
  vec->fe_space = fe_space;
  vec->name  = name ? strdup(name) : nil;
  vec->size  = 0;
  vec->vec   = nil;
  vec->refine_interpol = vec->coarse_restrict = nil;
  
  if (fe_space && fe_space->admin)
    add_int_dof_vec_to_admin(vec, (DOF_ADMIN *)fe_space->admin);

  return(vec);
}

void free_int_dof_vec(DOF_DOF_VEC *vec)
{
  void remove_int_dof_vec_from_admin(DOF_DOF_VEC *);
  
  if (vec->fe_space && vec->fe_space->admin)
    remove_int_dof_vec_from_admin(vec);
  
  MEM_FREE(vec->vec, vec->size, int);
  vec->vec = nil;
  vec->size = 0;

  if(vec->fe_space && vec->fe_space->admin)
    freeMemory((void *)vec, ((DOF_ADMIN_MEM_INFO*)(vec->fe_space->admin->mem_info))->int_dof_vec);
  else
    freeMemory((void *)vec, unconnected_int_dof_vecs);
}

ALBERTA_DEFUNUSED(static void delete_unconnected_int_dof_vecs(void))
{
  if(unconnected_int_dof_vecs) deleteObject(unconnected_int_dof_vecs);
  unconnected_int_dof_vecs = nil;
}


static void *unconnected_dof_uchar_vecs = nil;

DOF_UCHAR_VEC  *get_dof_uchar_vec(const char *name, const FE_SPACE *fe_space)
{
  DOF_UCHAR_VEC *vec;
  void add_dof_uchar_vec_to_admin(DOF_UCHAR_VEC *, DOF_ADMIN *);
  static void *vec_info = nil;


  if(!fe_space || !fe_space->admin) {
    if(!unconnected_dof_uchar_vecs)
      unconnected_dof_uchar_vecs = newObject(sizeof(DOF_UCHAR_VEC), 10);
    vec_info = unconnected_dof_uchar_vecs;
  }
  else
    vec_info = ((DOF_ADMIN_MEM_INFO*)(fe_space->admin->mem_info))->dof_uchar_vec;
  
  vec = (DOF_UCHAR_VEC *)getMemory(vec_info);
  
  vec->next = nil;
  vec->fe_space = fe_space;
  vec->name = name ? strdup(name) : nil;
  vec->size = 0;
  vec->vec = nil;
  vec->refine_interpol = vec->coarse_restrict = nil;
  
  if (fe_space && fe_space->admin)
    add_dof_uchar_vec_to_admin(vec, (DOF_ADMIN *)fe_space->admin);
  
  return(vec);
}

void free_dof_uchar_vec(DOF_UCHAR_VEC *vec)
{
  void remove_dof_uchar_vec_from_admin(DOF_UCHAR_VEC *);
  
  if (vec->fe_space && vec->fe_space->admin)
    remove_dof_uchar_vec_from_admin(vec);
  
  MEM_FREE(vec->vec, vec->size, U_CHAR);
  vec->vec = nil;
  vec->size = 0;

  if(vec->fe_space && vec->fe_space->admin)
    freeMemory((void *)vec, ((DOF_ADMIN_MEM_INFO*)(vec->fe_space->admin->mem_info))->dof_uchar_vec);
  else
    freeMemory((void *)vec, unconnected_dof_uchar_vecs);
}

ALBERTA_DEFUNUSED(static void delete_unconnected_dof_uchar_vecs(void))
{
  if(unconnected_dof_uchar_vecs) deleteObject(unconnected_dof_uchar_vecs);
  unconnected_dof_uchar_vecs = nil;
}


static void *unconnected_dof_schar_vecs = nil;

DOF_SCHAR_VEC  *get_dof_schar_vec(const char *name, const FE_SPACE *fe_space)
{
  DOF_SCHAR_VEC *vec;
  void add_dof_schar_vec_to_admin(DOF_SCHAR_VEC *, DOF_ADMIN *);
  static void *vec_info = nil;


  if(!fe_space || !fe_space->admin) {
    if(!unconnected_dof_schar_vecs)
      unconnected_dof_schar_vecs = newObject(sizeof(DOF_SCHAR_VEC), 10);
    vec_info = unconnected_dof_schar_vecs;
  }
  else
    vec_info = ((DOF_ADMIN_MEM_INFO*)(fe_space->admin->mem_info))->dof_schar_vec;
  
  vec = (DOF_SCHAR_VEC *)getMemory(vec_info);

  vec->next = nil;
  vec->fe_space = fe_space;
  vec->name = name ? strdup(name) : nil;
  vec->size = 0;
  vec->vec = nil;
  vec->refine_interpol = vec->coarse_restrict = nil;
  
  if (fe_space && fe_space->admin)
    add_dof_schar_vec_to_admin(vec, (DOF_ADMIN *)fe_space->admin);
  
  return(vec);
}

void free_dof_schar_vec(DOF_SCHAR_VEC *vec)
{
  void remove_dof_schar_vec_from_admin(DOF_SCHAR_VEC *);
  
  if (vec->fe_space && vec->fe_space->admin)
    remove_dof_schar_vec_from_admin(vec);
  
  MEM_FREE(vec->vec, vec->size, S_CHAR);
  vec->vec = nil;
  vec->size = 0;
  if(vec->fe_space && vec->fe_space->admin)
    freeMemory((void *)vec, ((DOF_ADMIN_MEM_INFO*)(vec->fe_space->admin->mem_info))->dof_schar_vec);
  else
    freeMemory((void *)vec, unconnected_dof_schar_vecs);
}

ALBERTA_DEFUNUSED(static void delete_unconnected_dof_schar_vecs(void))
{
  if(unconnected_dof_schar_vecs) deleteObject(unconnected_dof_schar_vecs);
  unconnected_dof_schar_vecs = nil;
}


static void *unconnected_dof_real_vecs = nil;

DOF_REAL_VEC  *get_dof_real_vec(const char *name, const FE_SPACE *fe_space)
{
  DOF_REAL_VEC *vec;
  void add_dof_real_vec_to_admin(DOF_REAL_VEC *, DOF_ADMIN *);
  static void *vec_info = nil;


  if(!fe_space || !fe_space->admin) {
    if(!unconnected_dof_real_vecs)
      unconnected_dof_real_vecs = newObject(sizeof(DOF_REAL_VEC), 10);
    vec_info = unconnected_dof_real_vecs;
  }
  else
    vec_info = ((DOF_ADMIN_MEM_INFO*)(fe_space->admin->mem_info))->dof_real_vec;
  
  vec = (DOF_REAL_VEC *)getMemory(vec_info);

  vec->next = nil;
  vec->fe_space = fe_space;
  vec->name = name ? strdup(name) : nil;
  vec->size = 0;
  vec->vec = nil;
  vec->refine_interpol = vec->coarse_restrict = nil;
  
  if (fe_space && fe_space->admin)
    add_dof_real_vec_to_admin(vec, (DOF_ADMIN *)fe_space->admin);
  
  return(vec);
}

void free_dof_real_vec(DOF_REAL_VEC *vec)
{
  void remove_dof_real_vec_from_admin(DOF_REAL_VEC *);
  
  if (vec->fe_space && vec->fe_space->admin)
    remove_dof_real_vec_from_admin(vec);
  
  MEM_FREE(vec->vec, vec->size, REAL);
  vec->vec = nil;
  vec->size = 0;
  if(vec->fe_space && vec->fe_space->admin)
    freeMemory((void *)vec, ((DOF_ADMIN_MEM_INFO*)(vec->fe_space->admin->mem_info))->dof_real_vec);
  else
    freeMemory((void *)vec, unconnected_dof_real_vecs);
}

ALBERTA_DEFUNUSED(static void delete_unconnected_dof_real_vecs(void))
{
  if(unconnected_dof_real_vecs) deleteObject(unconnected_dof_real_vecs);
  unconnected_dof_real_vecs = nil;
}


static void *unconnected_dof_real_d_vecs = nil;

DOF_REAL_D_VEC  *get_dof_real_d_vec(const char *name, const FE_SPACE *fe_space)
{
  DOF_REAL_D_VEC *vec;
  void add_dof_real_d_vec_to_admin(DOF_REAL_D_VEC *, DOF_ADMIN *);
  static void *vec_info = nil;


  if(!fe_space || !fe_space->admin) {
    if(!unconnected_dof_real_d_vecs)
      unconnected_dof_real_d_vecs = newObject(sizeof(DOF_REAL_D_VEC), 10);
    vec_info = unconnected_dof_real_d_vecs;
  }
  else
    vec_info = ((DOF_ADMIN_MEM_INFO*)(fe_space->admin->mem_info))->dof_real_d_vec;
  
  vec = (DOF_REAL_D_VEC *)getMemory(vec_info);

  vec->next = nil;
  vec->fe_space = fe_space;
  vec->name = name ? strdup(name) : nil;
  vec->size = 0;
  vec->vec = nil;
  vec->refine_interpol = vec->coarse_restrict = nil;

  if (fe_space && fe_space->admin)
    add_dof_real_d_vec_to_admin(vec, (DOF_ADMIN *)fe_space->admin);
  
  return(vec);
}

void free_dof_real_d_vec(DOF_REAL_D_VEC *vec)
{
  void remove_dof_real_d_vec_from_admin(DOF_REAL_D_VEC *);
  
  if (vec->fe_space && vec->fe_space->admin)
    remove_dof_real_d_vec_from_admin(vec);
  
  MEM_FREE(vec->vec, vec->size, REAL_D);
  vec->vec = nil;
  vec->size = 0;
  if(vec->fe_space && vec->fe_space->admin)
    freeMemory((void *)vec, ((DOF_ADMIN_MEM_INFO*)(vec->fe_space->admin->mem_info))->dof_real_d_vec);
  else
    freeMemory((void *)vec, unconnected_dof_real_d_vecs);
}

ALBERTA_DEFUNUSED(static void delete_unconnected_dof_real_d_vecs(void))
{
  if(unconnected_dof_real_d_vecs) deleteObject(unconnected_dof_real_d_vecs);
  unconnected_dof_real_d_vecs = nil;
}

